-- module --
module(...,package.seeall)

------------------------------------------------
-- Function implementing single inheritance.
-- The syntax is as follows:
--
-- LuaSubClass = class.inherit( LuaBaseClass )
-- {
--     put defaults here
-- }
--
-- If you don't want to specify any defaults, 
-- do one of the following:
--
-- LuaSubClass = class.inherit( LuaBaseClass )()
-- LuaSubClass = class.inherit( LuaBaseClass ){}
-- 
-- An instance of the class defines 'owner', 
-- which is a pointer to the owning C++ actor.
-- Create a new instance as follows (owner is a
-- C++ actor):
--
-- local instance = LuaSubClass:new( owner )
------------------------------------------------
function inherit( base_class )
	return function( defaults )
		-- Check input.
		defaults = defaults or {}
		
		-- Setup the metatable for the subclass and super.
		if base_class then
			setmetatable( defaults, { __index = base_class } )
		end
		
		-- Class.
		defaults.class = defaults
		
		-- Super.
		defaults.super = base_class
		
		-- Returns true if the callee is a subclass
		-- of the specified baseclass.
		function defaults:isa( base_class )
			local found   = false
			local current = defaults
			
			while not found and current do
				if current == base_class then
					found = true
				else
					current = current.super
				end
			end
			
			return found
		end
		
		-- Method to spawn a new instance.
		function defaults:new( owner )
			local instance = {}
			
			-- Setup the metatable for the instance.
			setmetatable( instance, { __index = defaults } )
			
			-- If we're attaching this instance to
			-- a C++ actor, attach it and return the
			-- actual C++ actor, not the lua instance.
			if owner ~= nil then
				-- Define owner.
				instance.owner = owner
				
				-- Define lua state.
				owner.lua = instance
				
				return owner
			end
			
			return instance
		end
		
		return defaults
	end
end

-- Create a base class with optional defaults.
-- The syntax is as follows:
--
-- LuaClass = class.new
-- {
--     put defaults here
-- }
--
-- If you don't want to specify any defaults, 
-- do the following:
--
-- LuaClass = class.new()
function new( defaults )
	return inherit()( defaults )
end

-- Returns a function that spawns a lua
-- class into a specified actor.
function partial_spawner( class )
	-- The spawn method.
	return function( owner )
		return class:new( owner )
	end	
end

-- Returns a function that spawns a lua
-- class into a new actor of type 'actor'.
function spawner( actor, class )
	-- The spawn method.
	return function()
		return class:new( actor() )
	end	
end

function gameObjectBinder(class)
	return function(object)
		local actual_object = object:get()
		class:new(actual_object) -- bind the script to the object	
		
		if actual_object.lua.onBindToObject then
			actual_object.lua.onBindToObject(actual_object.lua, object)
		end
	end
end
