-- module --
module(...,package.seeall)

-- includes --
local engine    = engine.instance
local chunks    = engine.chunks
local constants = engine.platform.constants
local assets    = chunks.assets
local pool      = chunks.pool
local util      = chunks.util
local music     = chunks.music
local credits   = chunks.credits
local load      = chunks.load

LuaCreditsMenuGame = chunks.class.new
{	
	m_scroll_speed = 30, --pixels per second
	
	m_asset_group = nil,
	m_scene_camera = nil,
}

-- The spawner.
spawn = chunks.class.spawner( SingleTouchWindowGame, LuaCreditsMenuGame )

function LuaCreditsMenuGame:created(game)
	print("LuaCreditsMenuGame:created")
	
	local game_window = game:getWindow()	
	
	-- Set clearing info.	
	-- Window camera.
	do
		local camera = game:getWindow():spawn( OrthographicCamera() )

		camera.viewport_offset = constants.main_menu_camera.window.viewport_offset
		camera.viewport_size   = constants.main_menu_camera.window.viewport_size
		camera.screen_size     = constants.main_menu_camera.window.screen_size
		camera.world_size      = constants.main_menu_camera.window.world_size
		camera.clipping_planes = constants.main_menu_camera.window.clipping_planes
		camera.clear_color     = constants.clear_color
		camera.clear_color_on  = constants.clear_color_on
		camera.clear_depth_on  = constants.clear_depth_on
		camera.scene           = game_window

		game_window.camera = camera
	end
	
	game_window.size = game_window.camera.world_size
	game_window.hidden = true
	
	self.m_scroll_speed = credits.scroll_speed or self.m_scroll_speed
	
	self.m_asset_group = ResourceGroup("texture")
	self:preloadAssets()
	
	Resources.get():addResourceGroup(self.m_asset_group)
	
	-- spawn credits
	local scroll_time = self:spawnCreditsWindow()
	
	-- spawn background units
	self:spawnBackgroundScene(scroll_time)
	
	Resources.get():removeResourceGroup(self.m_asset_group)
	
	music.checkOrPlaySong( assets.music.loop )
end

function LuaCreditsMenuGame:onInputEvents(game, events)
    -- Input handling for the android back button.  
    if engine.platform.device == "android" then
		for event in events:all() do
			local e = AndroidEvent.fast_cast( event )
			if e and e.type == AndroidEvent.AndroidEvent_BackButton then
				self:doUserWantsToQuitCredits()
			end
		end
	end

end

function LuaCreditsMenuGame:spawnBackgroundScene(move_time)
	local game_window = self.owner:getWindow()
	
	local background = game_window:spawn(Window())
	background.location = pool.get_v3f(0, 0, 0.0)
	background.size = game_window.size	
	background.color = pool.get_colorf(0.0, 0.0, 0.0, 1.0)
	
	background.on_release = function()
		self:doUserWantsToQuitCredits()
	end
	
	local window_half_size = game_window.size * 0.5
	local wave_action_actor = game_window:spawn(ActorBase())
	wave_action_actor.location = background.location + pool.get_v3f(0.0, 0.0, 200)
	wave_action_actor.hidden = true
	
	local options = {
						loop_count = 0,
						ease_out_current_animation_time = 0.3,
						ease_in_time = 0.1,
					}
	local unit_size = 70
	local start_x = -(game_window.size.x * 0.5) - unit_size
	local end_x = game_window.size.x + (unit_size * 2)
	local distance = math.abs(end_x - start_x)
	
	local screen_time = move_time / #credits.background_units
	for i,v in ipairs(credits.background_units) do
		local delay = v.delay
		local unit_name = v.unit
		local level = v.level or 1
		
		wave_action_actor:addAction(LuaAction(delay, function (inner)
			Resources.get():addResourceGroup(self.m_asset_group)			
			
			local y_position = v.y_position or math.random(-15, 7) * 10
			local container = inner:spawn(ActorBase())
			container.hidden = true
			container.location = pool.get_v3f(start_x, y_position, (y_position * -0.1))
			
			do -- Create the unit and squish on z
				local unit = util.loadObjectRenderDataFromStats(container, "units", unit_name, level)
				unit.scale = unit.scale * pool.get_v3f(1.0, 1.0, 0.0001)
				unit:playAnimation("walk", options)
			end
			
			local speed = load.getStat("units", unit_name, "speed", level)
			container:addAction(MoveOffsetAction(distance / speed, InterpAction.Lerp, pool.get_v3f(end_x, 0.0, 0.0)))
			container:addAction(DestroyAction(0.0))			
			
			Resources.get():removeResourceGroup(self.m_asset_group)
		end))
	end
end


function LuaCreditsMenuGame:spawnCreditsWindow()
	local game_window = self.owner:getWindow()
	
	local credits_window = game_window:spawn(Window())
	credits_window.hidden = true
	credits_window.location = pool.get_v3f(0.0, -game_window.size.y / 2, constants.main_menu_camera.window.clipping_planes.x - 1)
	
	local current_text_format = normal_text_format
	local last_window = credits_window
	local total_size = v2f(0.0, 0.0)
	for k,v in pairs(credits.data) do
		local text = v.text
		local texture = v.texture
		
		if text then
			if v.format then
				current_text_format = v.format
			end
		
			local label = last_window:spawn(Text())
			label.style = current_text_format[1]
			label.options = {height = current_text_format[2]}
			label.color = current_text_format[3]
			
			if #text == 0 then
				label.text = lstring("whitespace")
			else
				label.text = lstring(text)
			end
			label.anchor = {Anchor.Anchor_T, Anchor.Anchor_B, pool.get_v3f(0.0, 0.0, 0.0)}
			
			last_window = label
		elseif texture then
			local window = last_window:spawn(Window())
			assets.setTexture(window, texture)
			window.anchor = {Anchor.Anchor_T, Anchor.Anchor_B, pool.get_v3f(0.0, 0.0, 0.0)}
			
			last_window = window
		end
		
		total_size = total_size + last_window.size
	end
	
	local scroll_height = total_size.y + game_window.size.y
	local scroll_time = scroll_height/ self.m_scroll_speed
	
	local scroll_sequence = SequenceAction()
	scroll_sequence:addAction(MoveOffsetAction(scroll_time, InterpAction.Lerp, pool.get_v3f(0.0, scroll_height, 0.0)))
	scroll_sequence:addAction(LuaAction(0.0, function(inner)
		assets.playSound(assets.sounds.boomstick)
	end))
	scroll_sequence:addAction(LuaAction(4.0, function(inner)
		inner.hidden_recursive = true
		self:doUserWantsToQuitCredits()
	end))
	credits_window:addAction(scroll_sequence)
	
	return scroll_time
end

function LuaCreditsMenuGame:doUserWantsToQuitCredits()
	engine:goToPreviousGame()
end


function LuaCreditsMenuGame:destroy()
	self:unloadAssets()
end


function LuaCreditsMenuGame:preloadAssets()
	--TODO: FIXME local sound_manager = Audio.get().sound
	
	-- sounds
	--sound_manager:precache(URI(assets.getSound(assets.sounds.button_click)))

end

function LuaCreditsMenuGame:unloadAssets()
	Resources.get():unload(self.m_asset_group)
end
