-- module --
module(...,package.seeall)
local class = require( "common.class" )

-- globals --
instance = nil

-- LuaEngine --
LuaEngine = class.new
{
    platform = nil,
    chunks   = nil,
    m_loading_game = nil,
    m_root_game = nil,
    --g_god_mode = true,
    g_castle_units = true,
    m_game_spawn_function_stack = nil,
    pause_callback = nil,
    m_default_language_used = true,
    m_language = "en",
    m_upgrade_menu_window = nil
}

-- The created method.
function LuaEngine:created( owner )
    -- Seed random number generator.
    math.randomseed( os.time() )

    -- Uncomment to use pvr atlas of units
    --Resources.get():load("?://texture/objects/objects_atlas")

    -- Create the tables.
    self.platform = {}
    self.chunks   = {}
    self.m_game_spawn_function_stack = {}

    -- Figure out the device type.
    do
        local info     = DeviceInfo.get()
        local platform = info:query( "platform.type" )

        -- ios
        if platform == "ios" then
            local display_scale = info:query( "display.scale" )

            self.platform.device = "iphone"

            if info:is( "ios.ipad_ui_idiom" ) then
                    self.platform.device = "ipad"
            elseif display_scale and display_scale > 1.0 then
                    self.platform.device = "retina"
            end
        elseif platform == "android" then
            self.platform.device = "android"
        end

        print( "device: " .. self.platform.device )
    end

    -- Get device language
    do
        local info  = DeviceInfo.get()
        local language = info:query( "platform.language" )

        self.m_language = language

        if language == "en" then
            self.m_default_language_used = true
        else
            self.m_default_language_used = false
        end
    end

    -- Load device-specific constants.
    do
        self.platform.constants = require( "platform." .. self.platform.device .. ".constants" )

        owner.delta_time_cap = self.platform.constants.delta_time_cap
    end

    -- Make ourself global.
    instance = self

    require("chunks")

    self.m_loading_game = self.owner:getRoot():spawn(self.chunks.loading_game.spawn())
    self.m_root_game = self.m_loading_game:spawn(Game())

    self:setupDefaultUserdata()

    self:spawnMainMenu()

    -- self.chunks.events.onGameReady()

    -- Cancel all notifications whnever the game is started.
    local notifService = NotificationService.get()
    if notifService then
        notifService:cancelScheduledLocalNotifications()
    end
end

function LuaEngine:resetToMainMenu()
    do --Kill All games
        local root_game = self.m_root_game
        local num_games = root_game:getNumOwned()
        if num_games > 0 then
            for i = num_games - 1, 0, -1 do
                local game = root_game:getOwned(i)
                if game ~= nil then
                    game:destroy()
                end
            end
        end
    end

    self.g_god_mode = false
    self.m_game_spawn_function_stack = {}
    self:setupDefaultUserdata()
    self:spawnMainMenu()
end

function LuaEngine:resetData(owner)
    do --Kill All games
        local root_game = self.m_root_game
        local num_games = root_game:getNumOwned()
        if num_games > 0 then
            for i = num_games - 1, 0, -1 do
                local game = root_game:getOwned(i)
                if game ~= nil then
                    game:destroy()
                end
            end
        end
    end

    --UserData.instance():resetData()

    UserData.instance():reset()

    self.g_god_mode = false
    self.m_game_spawn_function_stack = {}
    self:setupDefaultUserdata()
    self:spawnMainMenu()
end

-- Spawn the main menu.
function LuaEngine:spawnMainMenu()
    self:spawnGame(function ()
        -- Overrides.
        if self.platform.constants.main_menu_overrides ~= nil then
            self.platform.constants.main_menu_overrides()
        end

        return self.chunks.main_menu.spawn()
    end)
end

-- Spawn the upgrade menu.
function LuaEngine:spawnUpgradeMenu()
    -- Overrides.
    --[[if self.platform.constants.main_menu_overrides ~= nil then
        self.platform.constants.main_menu_overrides()
    end]]

    self:spawnGame(self.chunks.upgrade_menu.spawn)
end

function LuaEngine:setUpgradeUnitMenuWindow(window)
   self.m_upgrade_menu_window = window
end

function LuaEngine:getUpgradeUnitMenuWindow()
    return self.m_upgrade_menu_window
end

-- Spawn the REAL upgrade Unit menu.
function LuaEngine:spawnUpgradeUnitMenu()
    -- Overrides.
    --[[if self.platform.constants.main_menu_overrides ~= nil then
        self.platform.constants.main_menu_overrides()
    end]]

    if UserData.instance():getGold() > 0 then
        self:spawnGame(self.chunks.upgrade_unit_menu.spawn)
    else
        self:spawnSelectUnitMenu()
    end
end



function LuaEngine:spawnSelectUnitMenu()
    -- Overrides.
    --[[if self.platform.constants.main_menu_overrides ~= nil then
        self.platform.constants.main_menu_overrides()
    end]]

    if UserData.instance():getCurrentWave() >= self.chunks.globals.min_wave_to_show_select_unit_menu then
        if self.chunks.util.hasAnyUnitsUnlocked() == true then
            self:spawnGame(self.chunks.select_unit_menu.spawn)
            return
        end
    end

    self:spawnMainGame()
end

function LuaEngine:spawnDebugOptionsMenu()
    -- Overrides.
    --[[if self.platform.constants.main_menu_overrides ~= nil then
        self.platform.constants.main_menu_overrides()
    end]]

    self:spawnGame(self.chunks.debug_options_menu.spawn)
end

function LuaEngine:spawnOptionsMenu()
    -- Overrides.
    if self.platform.constants.options_menu_overrides ~= nil then
        self.platform.constants.options_menu_overrides()
    end

    self:spawnGame(self.chunks.options_menu.spawn)
end

function LuaEngine:spawnUnlockedMenu()
    -- Overrides.
    --[[if self.platform.constants.main_menu_overrides ~= nil then
        self.platform.constants.main_menu_overrides()
    end]]

    local set = self.chunks.unlocked_menu_window.getUnlockedSet()
    if #set > 0 then
        self:spawnGame(self.chunks.unlocked_menu.spawn)
    else
        self:spawnUpgradeUnitMenu()
    end
end

-- Spawn credits.
function LuaEngine:spawnCreditsMenu()
    -- Spawn the game.
    self:spawnGame(function()
        -- Overrides.
        --[[if self.platform.constants.game_overrides ~= nil then
            self.platform.constants.game_overrides()
        end]]

        return self.chunks.credits_menu.spawn()
    end)
end

-- Spawn a game.
function LuaEngine:spawnMainGame()
    -- Spawn the game.
    self:spawnGame(function()
        -- Overrides.
        if self.platform.constants.game_overrides ~= nil then
            self.platform.constants.game_overrides()
        end

        return self.chunks.main_game.spawn()
    end)
end

function LuaEngine:setupDefaultUserdata()
    local defaults = { --selected_spells = "super_boomstick_blast",
                             }

    for key,default_value in pairs(defaults) do
        local current_value = UserData.instance():getValue(key)
        if #current_value == 0 then
            UserData.instance():setValue(key, default_value)
        end
    end

    self.chunks.util.tryToAutoSelectObjects()
end

function LuaEngine:unloadResources()
    --Resources.get():unloadAll()
    --Resources.get():unloadUnreferencedResources()
    Resources.get():unload("texture", ".*")
    --TODO: FIXME Audio.get().sound:unloadSounds("sound", ".*")
end

function LuaEngine:goToPreviousGame()
    print("LuaEngine:goToPreviousGame")

    self.m_loading_game.lua:fadeIn(
        function()
            self:killTopGame()
            self:removeTopGameFromSpawnStack()
        end,
        function()
            self.chunks.analytics.flush()
            collectgarbage("collect")
            self:unloadResources()

            local game_spawn_function = self.m_game_spawn_function_stack[#self.m_game_spawn_function_stack]
            if game_spawn_function == nil then
                self:spawnMainMenu()  -- Default to the main menu if nothing else exists in the stack
            else
                self.m_root_game:spawn(game_spawn_function())
            end
        end,
        function()
            self.m_loading_game.lua:fadeOut()
        end
    )
end

function LuaEngine:spawnGame(game_spawn_function)
    self.m_loading_game.lua:fadeIn(
        function()
            self:killTopGame()
        end,
        function()
            self.chunks.analytics.flush()
            collectgarbage("collect")
            self:unloadResources()

            table.insert(self.m_game_spawn_function_stack, game_spawn_function)
            self.m_root_game:spawn(game_spawn_function())
        end,
        function()
            self.m_loading_game.lua:fadeOut()
        end
    )
end

function LuaEngine:removeTopGameFromSpawnStack()
    table.remove(self.m_game_spawn_function_stack, #self.m_game_spawn_function_stack)
end

function LuaEngine:clearGameSpawnStack()
    self.m_game_spawn_function_stack = {}
end

function LuaEngine:killTopGame()
    local root_game = self.m_root_game
    local top_game_index = root_game:getNumOwned() - 1
    if top_game_index < 0 then
        return false
    end

    local top_game = root_game:getOwned(0)
    top_game:destroy()

    return true
end

function LuaEngine:willResignActive(owner)
    if self.pause_callback then
        self.pause_callback()
    end
end

function LuaEngine:setPauseCallback(callback)
    self.pause_callback = callback
end

function LuaEngine:suspend(owner)
    if self.pause_callback and DeviceInfo.get():query("platform.type") == "iphone" then
        self.pause_callback()
    end

    local game_service = GameService.get()
    local notifService = NotificationService.get()
    if notifService and game_service then

        local seconds_to_ad = game_service:getTimeToIncentAd()

        if( seconds_to_ad > 0 ) then
            local title = lstring("play")
            local body = lstring("ash_needs_upgrades")
            local params = {}
            notifService:scheduleLocalNotificationIn(seconds_to_ad, title, body, params)
            print("Scheduled incent ad notification.")
        end

        -- 24 Hour Notification
        local time_day = 60 * 60 * 24

        notifService:scheduleLocalNotificationIn(time_day, -- Time to Display
            lstring("play"), -- Title
            lstring("come_back_1"), -- Body
            {}) -- Parameters

        -- 72 Hour Notification
        local time_3days = time_day * 3

        notifService:scheduleLocalNotificationIn(time_3days, -- Time to Display
            lstring("play"), -- Title
            lstring("come_back_3"), -- Body
            {}) -- Parameters

        -- 1 Week Notification
        local time_1week = time_day * 7

        notifService:scheduleLocalNotificationIn(time_1week, -- Time to Display
            lstring("play"), -- Title
            lstring("come_back_7"), -- Body
            {}) -- Parameters

        print("Scheduled long term notifications.")
    end
end

function LuaEngine:resume(owner)
    if self.pause_callback and DeviceInfo.get():query("platform.type") == "android"then
        self.pause_callback()
    end

    local notifService = NotificationService.get()
    if notifService then
        notifService:cancelScheduledLocalNotifications()
    end

    local upgradeMenu = self:getUpgradeUnitMenuWindow();
    if( upgradeMenu ) then
        upgradeMenu:updateAdDisplay()
    end

    -- self.chunks.events.onGameReady()
end

function LuaEngine:_queueTask(task)
    local count = #self.m_tasks
    table.insert(self.m_tasks, count + 1, task)
end

function LuaEngine:_runNextTask()

    local count = #self.m_tasks
    if count > 0 then
        func = self.m_tasks[1]
        table.remove(self.m_tasks, 1)
        func()
    end

end

function LuaEngine:doStartupTasks()
    print("LuaEngine:doStartupTasks")
    
    self.m_tasks = {}

    -- GameCenter Login
    do
        local gc = GameCenterService.get()
        if gc then
            local gc_task = function()
                print("Running GameCenter login task")
                local callback = function(handle, responseCode, response)
                    print("GameCenter login task completed: " .. tostring(responseCode))

                    self.m_async_handle = nil
                    self:_runNextTask()

                end

                self.m_async_handle = gc:login(false, callback)
            end
            self:_queueTask(gc_task)
        end
    end

    self:_runNextTask()
end

-- Bind the engine.
function bindEngine( owner )
    LuaEngine:new( owner )
end



