-- module --
module(...,package.seeall)

-- includes --
local engine    = engine.instance
local chunks    = engine.chunks
local constants = engine.platform.constants
local assets    = chunks.assets
local pool      = chunks.pool
local util      = chunks.util
local load      = chunks.load
local globals	= chunks.globals

local white = colorf(1, 1, 1, 1)
local active_color = colorf(1.0, 0.89, 0.28, 1.0)
local inactive_color = colorf(0.75, 0.75, 0.75, 1.0)
local current_wave = 0
local scroll_view_inset = v2f(1, 4)

local level_string = lstring("level")
local stat_string = lstring("stat")

LuaInfoWindow = chunks.class.new
{
	m_selection_box = nil,
	m_popup_time = 0.2,
	m_coin_store = nil
}

-- The spawner.
spawn = chunks.class.spawner( Window, LuaInfoWindow )

function LuaInfoWindow:created(window)
    --For the Android Back Button
	window:getGame().lua.m_dismissable = window

	window.size = constants.main_menu_camera.window.world_size
	window.location = window.location + pool.get_v3f(0, 0, 400)
	window.color = pool.get_colorf(0, 0, 0, 0)
	window:addAction(AlphaAction(self.m_popup_time, InterpAction.Smooth, 0.75))

	--The background window
	local upgrade_window = util.createWindowWithTexture(window, assets.textures.selection_box)
	upgrade_window.force_color = true
	upgrade_window.anchor = {Anchor.Anchor_C, Anchor.Anchor_C, pool.get_v3f(0, 24, 1)}
	upgrade_window.scale = pool.get_v3f(0, 0, 1)
	upgrade_window:addAction(ScaleAction(self.m_popup_time, InterpAction.Smooth, pool.get_v3f(1, 1, 1)))
	self.m_upgrade_window = upgrade_window

	--Back button
	local text = lstring("options_back")
	local format =  {"Georgia Bold", 15.0, pool.get_colorf(1.0, 1.0, 1.0, 1.0), 45.0, TextAlign.Center, 0.0, 3.0, 0, false}
	local back_button = util.createButtonWithDefaultOrLocale(upgrade_window, text, format, assets.textures.blank_button_locale, 
		assets.textures.back_button)

	back_button.anchor = {Anchor.Anchor_TR, Anchor.Anchor_TR, pool.get_v3f(16, 6, 1)}
	back_button.on_release = function(owner)
		assets.playSound(assets.sounds.button_click_wav)
		self.m_upgrade_window:addAction(ScaleAction(self.m_popup_time, InterpAction.Smooth, pool.get_v3f(0, 0, 1)))
		self.owner:addAction(AlphaAction(self.m_popup_time, InterpAction.Smooth, 0))
		self.owner:addAction(DestroyAction(0))
	end
end

function LuaInfoWindow:setNameAndChunkName(name, chunk_name, purchase_fn, unpurchase_fn, must_be_bought)
	local ud = UserData.instance()
	local level = ud:getTypeLevel(name)

	--Settings
	local padding_size = 7
	local border_height = 21
	local border_width = 7
	local iron_cost_font = "Helvetica-Bold"
	local iron_cost_font_size = 14
	local title_font = "Georgia Bold"
	local title_font_size = 28
	local subtitle_font = "Georgia Bold"
	local subtitle_font_size = 16
	local subtitle_max_width = 160
	local description_font = "Georgia Bold"
	local description_font_size = 18
	local description_font_color = colorf(1, 0.92, 0.55, 1)
	local flavor_font = "Helvetica Bold Oblique"
	local flavor_font_size = 14
	local current_lvl_font = "Georgia Bold"
	local current_lvl_font_size = 24
	local lvl_offset = 50
	local current_stat_font = "Georgia Bold"
	local current_stat_font_size = 20
	local upgrade_font = "Helvetica-Bold"
	local upgrade_font_size = 16
	local upgrade_scale = 1.25
	local upgrade_cost_font = "Helvetica-Bold"
	local upgrade_cost_font_size = 14

	--Icon
	local icon_name = load.getStat(chunk_name, name, "upgrade_icon_texture") or
		load.getStat(chunk_name, name, "icon_texture") or
		"UI/game/units/icon_generic"
	local icon = util.createWindowWithTexture(self.m_upgrade_window, assets.findTextureByPathName(icon_name))
	icon.anchor = {Anchor.Anchor_TL, Anchor.Anchor_TL, pool.get_v3f(padding_size + border_width, -(padding_size + border_height), 1)}

	--Iron cost on icon
	local iron_cost = load.getStat("units", name, "iron_cost")
	if iron_cost then
		print("IRON", iron_cost)
		local iron_cost_text = icon:spawn(Text())
		iron_cost_text.style = iron_cost_font
		iron_cost_text.options = {height = iron_cost_font_size}
		iron_cost_text.color = pool.get_colorf(1, 1, 1, 1)
		iron_cost_text.anchor = {Anchor.Anchor_C, Anchor.Anchor_C, pool.get_v3f(13, -15, 1)}
		stat_string:arg(1, iron_cost)
		iron_cost_text.text = stat_string
	end

	--Title
	local title_label = icon:spawn(Text())
	title_label.style = title_font
	title_label.options = {height = title_font_size, max_width = 206}
	title_label.color = pool.get_colorf(1, 1, 1, 1)
	title_label.anchor = {Anchor.Anchor_TL, Anchor.Anchor_TR, pool.get_v3f(padding_size, 0+7, 1)}
	title_label.text = lstring(name .. "_title")
	
	do
		local window_half_size = (self.m_upgrade_window.size.x - (border_width + padding_size) * 2) * 0.5
		local text_half_size = title_label.size.x * 0.5
		if window_half_size - (icon.size.x + padding_size) >= text_half_size then
			title_label:destroy()
			title_label = self.m_upgrade_window:spawn(Text())
			title_label.style = title_font
			title_label.options = {height = title_font_size}
			title_label.color = pool.get_colorf(1, 1, 1, 1)
			title_label.anchor = {Anchor.Anchor_T, Anchor.Anchor_T, pool.get_v3f(icon.size.x / 2, -(border_height + padding_size)+7, 1)}
			title_label.text = lstring(name .. "_title")
		end
	end

	--Subtitle
	local subtitle_label = title_label:spawn(Text())
	subtitle_label.style = subtitle_font
	subtitle_label.options = {height = subtitle_font_size, max_width = subtitle_max_width, align = TextAlign.Center}
	subtitle_label.color = pool.get_colorf(1, 1, 1, 1)
	subtitle_label.anchor = {Anchor.Anchor_T, Anchor.Anchor_B, pool.get_v3f(0, 0, 1)}
	local subtitle_label_text = (chunk_name == "spells") and "skill" or (load.getStat(chunk_name, name, "unit_type") or "TODO")
	subtitle_label.text = lstring(subtitle_label_text)

	--Item has upgrade
	local current_wave = ud:getCurrentWave()
	local unlock_wave = load.getStat(chunk_name, name, "unlock_wave")
	local locked = unlock_wave and unlock_wave > current_wave
	local unbought = (must_be_bought and ud:getLocked(name))
	local has_next_level = load.hasNextLevel(chunk_name, name)
	if (not locked) and (not unbought) and has_next_level and chunk_name ~= "weapons" and name ~= "smithy" then
		--Current Lvl label
		local current_lvl_label = self.m_upgrade_window:spawn(Text())
		current_lvl_label.style = current_lvl_font
		current_lvl_label.options = {height = current_lvl_font_size, align = TextAlign.Center}
		current_lvl_label.color = pool.get_colorf(1, 1, 1, 1)
		current_lvl_label.anchor = {Anchor.Anchor_T, Anchor.Anchor_T, pool.get_v3f(-lvl_offset, -(icon.size.y + border_height + padding_size)+6, 1)}
		level_string:arg(1, level)
		current_lvl_label.text = level_string

		local function getDamage(lvl, weapon_stat_name)
			local damage = load.getStat(chunk_name, name, "damage", lvl)
			if not damage then
				local weapon = load.getStat(chunk_name, name, weapon_stat_name or "weapon", lvl)
				if weapon then
					damage = load.getStat("weapons", weapon, "dps", lvl) or load.getStat("weapons", weapon, "damage", lvl)
					if not damage then
						local ammo = load.getStat("weapons", weapon, "ammo", lvl)
						damage = load.getStat("ammo", ammo, "damage", lvl)
					end
				else
					local ammo = load.getStat(chunk_name, name, "ammo", lvl)
					damage = load.getStat("ammo", ammo, "damage", lvl)
				end
			end
			return damage
		end

		--Current info
		local is_ash = name == "ash"
		local ash_melee_weapon
		if is_ash and (not ud:getLocked("iron_fist")) then
			if not ud:getLocked("chainsaw") then
				ash_melee_weapon = "chainsaw"
			else
				ash_melee_weapon = "iron_fist"
			end
		end
		local current_parent = current_lvl_label
		local current_icon_parent = current_lvl_label
		local current_icon_anchor = {Anchor.Anchor_R, Anchor.Anchor_L, v3f(0, -28, 0)}
		local current_info = load.getStat(chunk_name, name, "info_text", level)
		if current_info then
			local info_icon_file = load.getStat(chunk_name, name, "info_icon")
			local info_icon_asset = info_icon_file and assets.findTextureByPathName(info_icon_file) or assets.textures.attack_icon
			local info_icon = util.createWindowWithTexture(current_icon_parent, info_icon_asset)
			info_icon.anchor = current_icon_anchor

			local current_info_label = current_parent:spawn(Text())
			current_info_label.style = current_stat_font
			current_info_label.options = {height = current_stat_font_size}
			current_info_label.color = pool.get_colorf(1, 1, 1, 1)
			current_info_label.anchor = {Anchor.Anchor_T, Anchor.Anchor_B, pool.get_v3f(0, 0-1, 0)}
			stat_string:arg(1, current_info)
			current_info_label.text = stat_string

		else
			--Current damage
			local current_damage = getDamage(level, is_ash and "secondary_weapon" or "weapon")
			if current_damage then
				local info_icon_file = load.getStat(chunk_name, name, "info_icon")
				local damage_icon_asset
				if info_icon_file then
					damage_icon_asset = assets.findTextureByPathName(info_icon_file)
				elseif current_damage < 0 then
					damage_icon_asset = assets.textures.heal_icon
				else
					damage_icon_asset = assets.textures.attack_icon
				end
				local damage_icon = util.createWindowWithTexture(current_icon_parent, damage_icon_asset)
				damage_icon.anchor = current_icon_anchor
				current_icon_parent = damage_icon
				current_icon_anchor = {Anchor.Anchor_T, Anchor.Anchor_B, v3f(0, -8, 0)}

				local current_damage_label = current_parent:spawn(Text())
				current_damage_label.style = current_stat_font
				current_damage_label.options = {height = current_stat_font_size}
				current_damage_label.color = pool.get_colorf(1, 1, 1, 1)
				current_damage_label.anchor = {Anchor.Anchor_T, Anchor.Anchor_B, pool.get_v3f(0, 0-1, 0)}
				stat_string:arg(1, math.abs(current_damage))
				current_damage_label.text = stat_string
				current_parent = current_damage_label
			end

			--Current melee weapon damage (Ash only)
			if is_ash and ash_melee_weapon then
				local current_damage = load.getStat("weapons", ash_melee_weapon, "damage", level)
				if current_damage then
					local weapon_icon_file = load.getStat("weapons", ash_melee_weapon, "info_icon", level)
					local damage_icon_asset = assets.findTextureByPathName(weapon_icon_file)
					local damage_icon = util.createWindowWithTexture(current_icon_parent, damage_icon_asset)
					damage_icon.anchor = current_icon_anchor
					current_icon_parent = damage_icon
					current_icon_anchor = {Anchor.Anchor_T, Anchor.Anchor_B, v3f(0, -8, 0)}

					local current_damage_label = current_parent:spawn(Text())
					current_damage_label.style = current_stat_font
					current_damage_label.options = {height = current_stat_font_size}
					current_damage_label.color = pool.get_colorf(1, 1, 1, 1)
					current_damage_label.anchor = {Anchor.Anchor_T, Anchor.Anchor_B, pool.get_v3f(0, 0-3, 0)}
					stat_string:arg(1, current_damage)
					current_damage_label.text = stat_string
					current_parent = current_damage_label
				end
			end

			--Current health
			local current_health = load.getStat(chunk_name, name, "health", level)
			if current_health then
				local health_icon = util.createWindowWithTexture(current_icon_parent, assets.textures.health_icon)
				health_icon.anchor = current_icon_anchor

				local current_health_label = current_parent:spawn(Text())
				current_health_label.style = current_stat_font
				current_health_label.options = {height = current_stat_font_size}
				current_health_label.color = pool.get_colorf(1, 1, 1, 1)
				current_health_label.anchor = {Anchor.Anchor_T, Anchor.Anchor_B, pool.get_v3f(0, 0-4, 0)}
				stat_string:arg(1, current_health)
				current_health_label.text = stat_string
			end
		end

		--Next level
		local next_lvl_label = self.m_upgrade_window:spawn(Text())
		next_lvl_label.style = current_lvl_font
		next_lvl_label.options = {height = current_lvl_font_size}
		next_lvl_label.color = pool.get_colorf(0, 1, 0, 1)
		next_lvl_label.anchor = {Anchor.Anchor_T, Anchor.Anchor_T, pool.get_v3f(lvl_offset, -(icon.size.y + border_height + padding_size)+6, 1)}
		level_string:arg(1, level + 1)
		next_lvl_label.text = level_string

		--Next info
		local next_parent = next_lvl_label
		local next_info = load.getStat(chunk_name, name, "info_text", level + 1)
		if next_info then
			local next_info_label = next_parent:spawn(Text())
			next_info_label.style = current_stat_font
			next_info_label.options = {height = current_stat_font_size}
			next_info_label.color = pool.get_colorf(0, 1, 0, 1)
			next_info_label.anchor = {Anchor.Anchor_T, Anchor.Anchor_B, pool.get_v3f(0, 0-1, 0)}
			stat_string:arg(1, next_info)
			next_info_label.text = stat_string

		else
			--Next damage
			local next_damage = getDamage(level + 1, is_ash and "secondary_weapon" or "weapon")
			if next_damage then
				local next_damage_label = next_parent:spawn(Text())
				next_damage_label.style = current_stat_font
				next_damage_label.options = {height = current_stat_font_size}
				next_damage_label.color = pool.get_colorf(0, 1, 0, 1)
				next_damage_label.anchor = {Anchor.Anchor_T, Anchor.Anchor_B, pool.get_v3f(0, 0-1, 0)}
				stat_string:arg(1, math.abs(next_damage))
				next_damage_label.text = stat_string
				next_parent = next_damage_label
			end

			--Next melee weapon damage (Ash only)
			if is_ash and ash_melee_weapon then
				local next_damage = load.getStat("weapons", ash_melee_weapon, "damage", level + 1)
				if next_damage then
					local next_damage_label = next_parent:spawn(Text())
					next_damage_label.style = current_stat_font
					next_damage_label.options = {height = current_stat_font_size}
					next_damage_label.color = pool.get_colorf(0, 1, 0, 1)
					next_damage_label.anchor = {Anchor.Anchor_T, Anchor.Anchor_B, pool.get_v3f(0, 0-3, 0)}
					stat_string:arg(1, next_damage)
					next_damage_label.text = stat_string
					next_parent = next_damage_label
				end
			end

			--Next health
			local next_health = load.getStat(chunk_name, name, "health", level + 1)
			if next_health then
				local next_health_label = next_parent:spawn(Text())
				next_health_label.style = current_stat_font
				next_health_label.options = {height = current_stat_font_size}
				next_health_label.color = pool.get_colorf(0, 1, 0, 1)
				next_health_label.anchor = {Anchor.Anchor_T, Anchor.Anchor_B, pool.get_v3f(0, 0-4, 0)}
				stat_string:arg(1, next_health)
				next_health_label.text = stat_string
				next_parent = next_health_label
			end
		end

		--Description
		local description_parent = self.m_upgrade_window:spawn(Window())
		description_parent.size = pool.get_v2f(self.m_upgrade_window.size.x - border_width * 2, 90)
		description_parent.anchor = {Anchor.Anchor_B, Anchor.Anchor_B, pool.get_v3f(0, border_width + padding_size, 1)}
		description_parent.hidden = true

		local description_text = name .. "_description"
		local description_label_format = {description_font, description_font_size, description_font_color}
		local description_label = util.createMultiLineLabel(description_parent, description_text, description_label_format, 275)

	--No stats: show description and flavor text
	else
		--Description
		local description_text = name .. "_description"
		local description_label_format = {description_font, description_font_size, description_font_color}
		local description_label = util.createMultiLineLabel(self.m_upgrade_window, description_text, description_label_format, 275)
		description_label.anchor = {Anchor.Anchor_T, Anchor.Anchor_T, pool.get_v3f(0, -(icon.size.y + padding_size * 2 + border_height), 1)}

		--Flavor text
		local flavor_text = load.getStat(chunk_name, name, "flavor_text")
		if flavor_text then
			local quotations = lstring("quotations")
			quotations:arg(1, lstring(name .. "_flavor"))
			local flavor_label = util.createMultiLineLabel(self.m_upgrade_window, quotations, {flavor_font, flavor_font_size, pool.get_colorf(1, 1, 1, 1)}, 275, 275)
			flavor_label.anchor = {Anchor.Anchor_B, Anchor.Anchor_B, pool.get_v3f(0, padding_size * 2 + border_width, 1)}
		end
	end

	--Upgrade button
	if (not locked) and ((has_next_level and chunk_name ~= "weapons") or unbought) then
		--Upgrade is affordable
		local upgrade_cost = load.getStat(chunk_name, name, "cost", unbought and level or (level + 1))
		local gold = ud:getGold()
		if upgrade_cost <= gold then
			--Upgrade button
			local upgrade_button = util.createButtonWithTexture(self.m_upgrade_window, assets.textures.upgrade_button_blank)
			upgrade_button.anchor = {Anchor.Anchor_T, Anchor.Anchor_B, pool.get_v3f(0, 20, 1)}
			upgrade_button.scale = pool.get_v3f(upgrade_scale, upgrade_scale, 1)
			upgrade_button.on_release = function(owner)
				self.m_upgrade_window:addAction(ScaleAction(0.25, InterpAction.Smooth, pool.get_v3f(0, 0, 1)))
				self.owner:addAction(AlphaAction(0.25, InterpAction.Smooth, 0))
				self.owner:addAction(DestroyAction(0))
				purchase_fn()
			end

			--Upgrade label
			local upgrade_label = upgrade_button:spawn(Text())
			upgrade_label.style = upgrade_font
			upgrade_label.options = {height = upgrade_font_size, max_width = 50, align = TextAlign.Center}
			upgrade_label.anchor = {Anchor.Anchor_C, Anchor.Anchor_C, pool.get_v3f(0, 7.5+4, 1)}
			upgrade_label.text = unbought and lstring("BUY") or lstring("UPGRADE")

			--Cost window
			local cost_window = upgrade_button:spawn(Window())
			cost_window.anchor = {Anchor.Anchor_B, Anchor.Anchor_B, pool.get_v3f(0, 10, 1)}
			cost_window.hidden = true

			--Gold Icon
			local gold_icon = util.createWindowWithTexture(cost_window, assets.textures.upgrade_gold_coin)
			gold_icon.scale = pool.get_v3f(0.5, 0.5, 1)
			gold_icon.anchor = {Anchor.Anchor_L, Anchor.Anchor_L, pool.get_v3f(0, 0, 1)}

			--Cost Label
			local cost_label = cost_window:spawn(Text())
			cost_label.style = upgrade_cost_font
			cost_label.options = {height = upgrade_cost_font_size, max_width = 34}
			cost_label.anchor = {Anchor.Anchor_R, Anchor.Anchor_R, pool.get_v3f(0, 0+3, 1)}
			stat_string:arg(1, tostring(upgrade_cost))
			cost_label.text = stat_string

			cost_window.size = pool.get_v2f(gold_icon.size.x * gold_icon.scale.x + cost_label.size.x * cost_label.scale.x, math.max(gold_icon.size.y * gold_icon.scale.y, cost_label.size.y * cost_label.scale.y))

		--Unaffordable: show buy coins button
		else
			local text = lstring("buy_coins")
			local format = {"Georgia Bold", 15.0, pool.get_colorf(0.93, 0.81, 0.27, 1.0), 45.0, TextAlign.Center, 0.0, 4.0, 0.0, false}
			local buy_gold =  util.createButtonWithDefaultOrLocale(self.m_upgrade_window, text, format, assets.textures.buy_coins_button_locale, assets.textures.buy_coins_button)
			buy_gold.anchor = {Anchor.Anchor_T, Anchor.Anchor_B, pool.get_v3f(0, 20, 1)}

			--IN APP PURCHASES
			buy_gold.on_release = function(owner)
				assets.playSound(assets.sounds.button_click_wav)
				self.m_upgrade_window:addAction(ScaleAction(0.25, InterpAction.Smooth, pool.get_v3f(0, 0, 1)))
				self.owner:addAction(AlphaAction(0.25, InterpAction.Smooth, 0))
				self.owner:addAction(DestroyAction(0))

                if chunks.iap.purchasingAvailable() then
        			local parent_window = self.owner:getOwner()
        			parent_window:addAction(LuaAction(0.25, function()
        				self.m_coin_store = parent_window:spawn(chunks.coinstore.spawn())
        			end))
                else
                    chunks.iap.displayPurchasingUnavailable()
                end
			end
		end
	end

	--[[Downgrade button
	if (level > 1) or ((not unbought) and must_be_bought) then
		-- display the option to downgrade
		local downgrade_cost = load.getStat(chunk_name, name, "cost", level) * globals.downgrade_gold_return * -1
		downgrade_cost = math.floor(downgrade_cost)
		local downgrade_button = util.createButtonWithTexture(self.m_upgrade_window, assets.textures.upgrade_button_blank)
		downgrade_button.anchor = {Anchor.Anchor_T, Anchor.Anchor_BL, pool.get_v3f(20, 20, 1)}
		downgrade_button.scale = pool.get_v3f(upgrade_scale, upgrade_scale, 1)
		downgrade_button.on_release = function(owner)
			self.m_upgrade_window:addAction(ScaleAction(0.25, InterpAction.Smooth, pool.get_v3f(0, 0, 1)))
			self.owner:addAction(AlphaAction(0.25, InterpAction.Smooth, 0))
			self.owner:addAction(DestroyAction(0))
			if unpurchase_fn ~= nil then
				unpurchase_fn()
			else
				print("unpurchase fnc doesn't exist. ):")
			end
		end
		
		local downgrade_label = downgrade_button:spawn(Label())
		downgrade_label.format = TextFormat(upgrade_font, upgrade_font_size, pool.get_colorf(1, 1, 1, 1))
		downgrade_label.anchor = {Anchor.Anchor_C, Anchor.Anchor_C, pool.get_v3f(0, 7.5, 1)}
		downgrade_label.text = "DOWNGRADE"
		constrainLabel(downgrade_label, 50)

		--Cost window
		local cost_window = downgrade_button:spawn(Window())
		cost_window.anchor = {Anchor.Anchor_B, Anchor.Anchor_B, pool.get_v3f(0, 10, 1)}
		cost_window.hidden = true

		--Gold Icon
		local gold_icon = util.createWindowWithTexture(cost_window, assets.textures.upgrade_gold_coin)
		gold_icon.scale = pool.get_v3f(0.5, 0.5, 1)
		gold_icon.anchor = {Anchor.Anchor_L, Anchor.Anchor_L, pool.get_v3f(0, 0, 1)}

		--Cost Label
		local cost_label = cost_window:spawn(Label())
		cost_label.format = TextFormat(upgrade_cost_font, upgrade_cost_font_size, pool.get_colorf(1.0, 1.0, 1.0, 1.0))
		cost_label.anchor = {Anchor.Anchor_R, Anchor.Anchor_R, pool.get_v3f(0, 0, 1)}
		cost_label.text = tostring(downgrade_cost)
		constrainLabel(cost_label, 34)

		cost_window.size = pool.get_v2f(gold_icon.size.x * gold_icon.scale.x + cost_label.size.x * cost_label.scale.x, math.max(gold_icon.size.y * gold_icon.scale.y, cost_label.size.y * cost_label.scale.y))
	end
	--]]
end

function LuaInfoWindow:dismiss()
    --For the Android Back Button
	self.owner:getGame().lua.m_dismissable = nil
    
    self.m_upgrade_window:addAction(ScaleAction(self.m_popup_time, InterpAction.Smooth, pool.get_v3f(0, 0, 1)))
	self.owner:addAction(AlphaAction(self.m_popup_time, InterpAction.Smooth, 0))
	self.owner:addAction(DestroyAction(0))
end

function LuaInfoWindow:destroy(window)
    if window:getGame().lua.m_dismissable == window then
        window:getGame().lua.m_dismissable = nil
    end
end