module(...,package.seeall)
-- includes --
local engine    = engine.instance
local chunks    = engine.chunks
local assets    = chunks.assets
local util      = chunks.util
local pool      = chunks.pool

local main_sound_set = nil
local main_sound_index = 0


-- Helper Functions
function hasNextLevel(data_file_name, type_name, force_level)
	local data_file = chunks[data_file_name]
	if data_file == nil then
		return false
	end

	local type_data = data_file[type_name]
	if type_data == nil then
		return false
	end

	local level = force_level or UserData.instance():getTypeLevel(type_name)
	level = level + 1
	local level_data = type_data.level_stat[level]

	return level_data ~= nil
end

function getLevelData(type_data, level)
	level = clampi(level, 1, #type_data.level_stat)
	return type_data.level_stat[level]
end

function getStat(data_file_name, type_name, stat_name, force_level)
	local data_file = chunks[data_file_name]
	if data_file == nil then
		return nil
	end

	local type_data = data_file[type_name]
	if type_data == nil then
		return nil
	end

	local level = force_level or UserData.instance():getTypeLevel(type_name)
	local level_data = getLevelData(type_data, level)

	if level_data == nil then
		return type_data[stat_name]
	end

	local data = level_data[stat_name] or type_data[stat_name]

	return data
end

function getHeight(data_file_name, type_name, force_level)
	return getStat(data_file_name, type_name, "height", force_level) or 65.0
end

function getRandomItemFromString(parse_string)
	if parse_string == nil then
		return nil
	end

	local string_list = util.splitString(parse_string, ";")
	local item_index = math.random(1, #string_list)

	return string_list[item_index]
end

local function findStringInSet(string, set)
	for i,v in ipairs(set) do
		if v == string then
			return true
		end
	end
	return false
end

function playNextSpawnSound(type_name, game)
	--local on_spawn_sound_list = level_data.on_spawn_sound or type_data.on_spawn_sound
	local on_spawn_sound_list = getStat("units", type_name, "on_spawn_sound")
	if on_spawn_sound_list == nil or #on_spawn_sound_list == 0 then
		return
	end

	local sound
	repeat
		main_sound_index = (main_sound_index % #main_sound_set) + 1
		sound = main_sound_set[main_sound_index]
		--print("trying " .. sound)
		--print("in list " .. on_spawn_sound_list)
	until findStringInSet(sound, util.splitString(on_spawn_sound_list, ";"))

	util.playSoundLeftOfCamera("?://" .. sound, game)

	--[[ TODO: FIXME 
	on_spawn_sound = getRandomItemFromString(on_spawn_sound_list)
	if on_spawn_sound then
		--Audio.get().sound:play("sound://" .. on_spawn_sound)
		util.playSoundLeftOfCamera("sound://" .. on_spawn_sound, {occurences=1})
	end
	]]
end

-- Local part loading functions
local function loadRenderPart(object, type_data, level_data)
	local texture_list = level_data.texture or type_data.texture
	local texture = getRandomItemFromString(texture_list)
	local model = level_data.model or type_data.model
	local scale = level_data.scale or type_data.scale

	if texture == nil then
		return
	end

	local render_part = object:get():addPart(PartManager.PART_RENDER)
	local scene = object:get():getGame():getScene()

	local default_render_actor = function()
		print("No Model! Creating A Blue Square")
		render_part:get():setRenderActor( util.spawnTestSpriteActor(object:get():getGame(), v2f(8, 8), colorf(0.0, 0.0, 1.0, 1.0)) )
	end

	-- print("\tLoading Part: render")
	if texture and model then
		if util.fileHasExtension(model, "gr2") then
			local granny_actor = scene:spawn(BasicSkeletonActor())
			granny_actor:enableExtraZOffset()
			granny_actor:setSkeleton("?://" .. model)
			granny_actor.texture = "?://" .. texture
			render_part:get():setRenderActor( granny_actor )
			if scale then
				granny_actor.scale = pool.get_v3f(scale, scale, 1.0)
			end
		elseif util.fileHasExtension(model, "obj") then
			local obj_actor = scene:spawn(ActorBase())
			obj_actor.mesh = "?://" .. string.gsub(model, "%.obj", "")  -- Strip the extension
			obj_actor.texture = "?://" .. texture
			render_part:get():setRenderActor( obj_actor )
			if scale then
				obj_actor.scale = pool.get_v3f(scale, scale, 1.0)
			end
		else
			default_render_actor()
		end

		return
	end

	if texture then
		local texture_asset = assets.findTextureByPathName(texture)
		if texture_asset then
			local sprite = scene:spawn(SpriteActor())
			assets.setTexture(sprite, texture_asset)
			render_part:get():setRenderActor( sprite )
			if scale then
				sprite.scale = pool.get_v3f(scale, scale, 1.0)
			end
		else
			default_render_actor()
		end

		return
	end

	default_render_actor()
end


local function loadMovementPart(object, type_data, level_data)
	local speed = level_data.speed or type_data.speed
	if speed == nil then
		return
	end
	-- print("\tLoading Part: movement")

	local rotation_speed = level_data.rotation_speed or type_data.rotation_speed or 0.0

	local movement_part = object:get():addPart(PartManager.PART_MOVEMENT)
	movement_part:get().speed = speed
	movement_part:get().rotation_speed = rotation_speed
end


local function loadHealtPart(object, type_data, level_data)
	local health = level_data.health or type_data.health
	if health == nil then
		return
	end
	-- print("\tLoading Part: health")

	local can_be_knocked_back = level_data.can_be_knocked_back or type_data.can_be_knocked_back
	if can_be_knocked_back == nil then
		can_be_knocked_back = true
	end

	local on_take_damage_sound = level_data.on_take_damage_sound or type_data.on_take_damage_sound

	local health_part = object:get():addPart(PartManager.PART_HEALTH)
	health_part:get().health = health
	health_part:get().can_be_knocked_back = can_be_knocked_back
	if on_take_damage_sound then
		local list = util.splitString(on_take_damage_sound, ";")
		for i,v in ipairs(list) do
			health_part:get().on_take_damage_sound = URI("?", v)
		end
	end
end


local function loadWeaponPart(object, type_data, level_data)
	local range = level_data.range or type_data.range
	if range == nil then
		return
	end

	local min_range = level_data.min_range or type_data.min_range or 0.0
	-- print("\tLoading Part: weapon")

	local ammo = level_data.ammo or type_data.ammo
	local damage = level_data.damage or type_data.damage
	local reload_delay = level_data.reload_delay or type_data.reload_delay or 0
	local min_knockback = level_data.min_knockback or type_data.min_knockback or 0
	local max_knockback = level_data.max_knockback or type_data.max_knockback or 0
	local splash_range = level_data.splash_range or type_data.splash_range or 0
	local flame_dps = level_data.flame_dps or type_data.flame_dps or 0
	local knockback_multiplier = level_data.knockback_multiplier or type_data.knockback_multiplier or 1
	local on_deal_damage_sound = level_data.on_deal_damage_sound or type_data.on_deal_damage_sound
	local on_animation_start_sound = level_data.on_animation_start_sound or type_data.on_animation_start_sound 
	local on_fire_effect = level_data.on_fire_effect or type_data.on_fire_effect or ""
	local on_fire_effect_scale = level_data.on_fire_effect_scale or type_data.on_fire_effect_scale or 1.0
	local on_fire_effect_info = EffectInfo(URI("particles", on_fire_effect), v3f(on_fire_effect_scale))

	local valid = false
	if ammo then
		local weapon_part = object:get():addPart(PartManager.PART_WEAPON)
		weapon_part:get().ammo = ammo
		weapon_part:get().range = range
		weapon_part:get().min_range = min_range
		weapon_part:get().reload_delay = reload_delay
		weapon_part:get().on_fire_effect = on_fire_effect_info
		valid = true
	elseif damage then
		local weapon_part = object:get():addPart(PartManager.PART_WEAPON)
		weapon_part:get().range = range
		weapon_part:get().min_range = min_range
		weapon_part:get().reload_delay = reload_delay
		weapon_part:get():setDamageInfo(DamageInfo(damage, DamageInfo.PHYSICAL_DAMAGE, splash_range, flame_dps, knockback_multiplier))
		weapon_part:get().min_knockback = min_knockback
		weapon_part:get().max_knockback = max_knockback
		weapon_part:get().on_fire_effect = on_fire_effect_info
		--weapon_part:get().on_fire_effect_scale = on_fire_effect_scale
		if on_deal_damage_sound then
			local list = util.splitString(on_deal_damage_sound, ";")
			for i,v in ipairs(list) do
				weapon_part:get().on_deal_damage_sound = URI("?", v)
			end
		end
		if on_animation_start_sound then
			local list = util.splitString(on_animation_start_sound, ";")
			for i,v in ipairs(list) do
				print("Setting on_animation_start_sound to", v)
				weapon_part:get().on_animation_start_sound = URI("?", v)
			end
		end

		valid = true
	end

	if valid then
		-- print("\tLoading Part: item")
		local item_part = object:get():addPart(PartManager.PART_ITEM)
		item_part:get().equip_slot = PartInventory.WEAPON_EQUIP_SLOT
	end
end


local function loadAmmoPart(object, type_data, level_data)
	local trajectory = level_data.trajectory or type_data.trajectory
	if trajectory == nil then
		return
	end

	local damage = level_data.damage or type_data.damage
	if damage == nil then
		return
	end
	-- print("\tLoading Part: ammo")

	local min_knockback = level_data.min_knockback or type_data.min_knockback or 0
	local max_knockback = level_data.max_knockback or type_data.max_knockback or 0
	local splash_range = level_data.splash_range or type_data.splash_range or 0
	local flame_dps = level_data.flame_dps or type_data.flame_dps or 0
	local knockback_multiplier = level_data.knockback_multiplier or type_data.knockback_multiplier or 1
	local on_fired_sound = level_data.on_fired_sound or type_data.on_fired_sound
	local on_deal_damage_sound = level_data.on_deal_damage_sound or type_data.on_deal_damage_sound
	local hit_ground = level_data.hit_ground or type_data.hit_ground or false


	local ammo = object:get():addPart(PartManager.PART_AMMO)
	ammo:get():setDamageInfo(DamageInfo(damage, DamageInfo.PHYSICAL_DAMAGE, splash_range, flame_dps, knockback_multiplier))
	ammo:get().min_knockback = min_knockback
	ammo:get().max_knockback = max_knockback
	ammo:get().hit_ground = hit_ground
	if trajectory == "arching" then
		ammo:get().trajectory = PartAmmo.ARCHING
	else
		ammo:get().trajectory = PartAmmo.STRAIGHT
	end
	if on_fired_sound then
		ammo:get().on_fired_sound = URI("?", getRandomItemFromString(on_fired_sound))
	end
	if on_deal_damage_sound then
		ammo:get().on_deal_damage_sound = URI("?", getRandomItemFromString(on_deal_damage_sound))
	end

end


-- Object loading
local function loadWeaponForObject(object, weapon_name, equip)
	local game = object:get():getMainGame()

	local inventory_part = object:get():addPart(PartManager.PART_INVENTORY)

	local level = object:get():getLevel()
	local weapon = game:createGameObject("weapon", weapon_name, level)
	if weapon:get() == nil then
		return
	end

	local weapon_item = PartItemIndex(weapon)
	if weapon_item:get() == nil then
		return
	end

	local inventory_object = inventory_part:get()
	if not equip and not inventory_object:getEquippedItemAtSlot(PartInventory.SECONDARY_WEAPON_EQUIP_SLOT):get() then
		--print("switching weapon to secondary slot")
		weapon_item:get().equip_slot = PartInventory.SECONDARY_WEAPON_EQUIP_SLOT
		equip = true
	end

	inventory_object:addItem(weapon_item)
	if equip then
		inventory_object:equipItem(weapon_item)
	end
end


local function loadHelmetForObject(object, helmet_name, equip)
	local game = object:get():getMainGame()

	local inventory_part = object:get():addPart(PartManager.PART_INVENTORY)

	local level = object:get():getLevel()
	local item = game:createGameObject("item", helmet_name, level)
	if item:get() == nil then
		return
	end

	local item_part = PartItemIndex(item)
	if item_part:get() == nil then
		return
	end

	item_part:get().equip_slot = PartInventory.HELMET_EQUIP_SLOT

	local inventory_object = inventory_part:get()
	inventory_object:addItem(item_part)
	if equip then
		inventory_object:equipItem(item_part)
	end
end


local function loadShieldForObject(object, shield_name, equip)
	local game = object:get():getMainGame()

	local inventory_part = object:get():addPart(PartManager.PART_INVENTORY)

	local level = object:get():getLevel()
	local item = game:createGameObject("item", shield_name, level)
	if item:get() == nil then
		return
	end

	local item_part = PartItemIndex(item)
	if item_part:get() == nil then
		return
	end

	item_part:get().equip_slot = PartInventory.SHIELD_EQUIP_SLOT

	local inventory_object = inventory_part:get()
	inventory_object:addItem(item_part)
	if equip then
		inventory_object:equipItem(item_part)
	end
end


local function loadItemsForObject(object, type_data, level_data)
	-- print("\tLoading Items: ")
	local primary_weapon = getRandomItemFromString(level_data.weapon) or getRandomItemFromString(type_data.weapon)
	if primary_weapon then
		loadWeaponForObject(object, primary_weapon, true)
	end

	local secondary_weapon = getRandomItemFromString(level_data.secondary_weapon) or getRandomItemFromString(type_data.secondary_weapon)
	if secondary_weapon then
		loadWeaponForObject(object, secondary_weapon, false)
	end

	local helmet = getRandomItemFromString(level_data.helmet) or getRandomItemFromString(type_data.helmet)
	if helmet then
		loadHelmetForObject(object, helmet, true)
	end

	local shield = getRandomItemFromString(level_data.shield) or getRandomItemFromString(type_data.shield)
	if shield then
		loadShieldForObject(object, shield, true)
	end
end


local function loadObject(object, type_data, type)
	local level = object:get():getLevel()
	local level_data = getLevelData(type_data, level) or {}

	object:get().label = type or type_data.title

	-- Load parts
	loadRenderPart(object, type_data, level_data)
	loadMovementPart(object, type_data, level_data)
	loadHealtPart(object, type_data, level_data)
	loadWeaponPart(object, type_data, level_data)
	loadAmmoPart(object, type_data, level_data)

	loadItemsForObject(object, type_data, level_data)

	-- Setup team
	if type_data.ally ~= nil then
		if type_data.ally then
			object:get().team = GameObject.TEAM_ALLY
		else
			object:get().team = GameObject.TEAM_ENEMY
		end
	end

	-- Set On Death Sound
	local on_death_sound = level_data.on_death_sound or type_data.on_death_sound
	if on_death_sound then
		object:get().on_death_sound = URI("?", getRandomItemFromString(on_death_sound))
	end

	-- Bind a custom script if we have one
	if type_data.script ~= nil then
		print("binding script", type_data.script)
		chunks[type_data.script].bindToObject(object)
	end

	--if math.random() <= 0.6 then
		--playNextSpawnSound(type_data, level_data)
	--end

	return true
end

-- Object loading functions called from c++
function unit(object, type, force_level)
	print("Loading unit:", type, force_level)

	local type_data = chunks.units[type]
	if type_data == nil then
		return false
	end

	local loaded = loadObject(object, type_data, type)
	if loaded then
		object:get():addPart(PartManager.PART_ORDERS)
	end

	return loaded
end

function weapon(object, type, force_level)
	print("Loading weapon:", type)

	local type_data = chunks.weapons[type]
	if type_data == nil then
		return false
	end

	local loaded = loadObject(object, type_data, type)
	if loaded then
		object:get():addPart(PartManager.PART_ITEM)
	end

	return loaded
end

function item(object, type, force_level)
	print("Loading item:", type)

	local type_data = chunks.items[type]
	if type_data == nil then
		return false
	end

	local loaded = loadObject(object, type_data, type)
	if loaded then
		object:get():addPart(PartManager.PART_ITEM)
	end

	return loaded
end

function ammo(object, type, force_level)
	print("Loading ammo:", type)

	local type_data = chunks.ammo[type]
	if type_data == nil then
		return false
	end

	local loaded = loadObject(object, type_data, type)
	return loaded
end

function shuffleMainSoundSet()

	math.randomseed( os.time() + os.clock() * 100000000 )

	local temp_sound_set = {}

	while #main_sound_set > 0 do
		local sound = table.remove(main_sound_set, math.random(1,#main_sound_set))
		table.insert(temp_sound_set, sound)
	end

	main_sound_set = temp_sound_set
end

function buildMainSoundSet()
	local temp_sound_set = {}

	for k,foo in pairs(chunks.units) do
		local sounds_string
		if foo.level_stat then
			sounds_string = getStat("units", k, "on_spawn_sound")
		else
			sounds_string = foo.on_spawn_sound
		end

		if sounds_string then
			local sounds_set = util.splitString(sounds_string, ";")
			--print(k)
			for bar,v in ipairs(sounds_set) do
				temp_sound_set[v] = true
				--print(v)
			end
		end
	end

	main_sound_set = {}
	main_sound_index = 0

	for k,v in pairs(temp_sound_set) do
		table.insert(main_sound_set, k)
		--print(k, #main_sound_set)
	end

end

buildMainSoundSet()


