-- module --
module(...,package.seeall)

-- includes --
local engine    = engine.instance
local chunks    = engine.chunks
local constants = engine.platform.constants
local assets    = chunks.assets
local pool      = chunks.pool
local util      = chunks.util
local music     = chunks.music
local globals   = chunks.globals
local constants = engine.platform.constants
local events    = chunks.events
local achievements = chunks.achievements
local analytics = chunks.analytics

--Difficulty settings
local enable_wave_difficulty_scaling = true --Disable this for wave balancing
local wave_difficulty_curve = 3 --The curve for applying wave difficulty scalar
local difficulty_dt_curve = 10 --The curve for applying dt scalar (steeper curve means it doesn't get harder 'till closer to the gate)
local difficulty_dt_factor_lo = 1 --Base scalar for delta-time difficulty modification
local difficulty_dt_factor_hi = 5 --Peak scalar for delta-time difficulty modification (at castle gate)

LuaMainGame = chunks.class.new
{
	m_last_spawn_time = 0.0,
	m_ash = nil,
	m_necro = nil,
	m_last_unit_spawned = false,
	m_game_over = false,
	m_gold_collected = 0,
	m_iron = 0,
	m_iron_rate = 1.15,
	m_game_window = nil,
    m_tutorial_window = nil,
	m_spell_target_area_actor = nil,
	m_buttons_unique = nil,
	m_units_this_wave = 0,
	m_spawned_so_far = 0,
	m_enemies_killed = 0,
	m_window_touch_id = -1,
	m_wave_action_actor = nil,
	m_wave_script_actor = nil,
	m_iron_level = 1,
	m_most_recent_enemy = nil,
    m_pit = nil,
    m_pit_open = nil,
    m_pit_closed = nil,
	m_asset_group = nil,
	m_front_line_tracker = nil,
	m_on_resign_active_event = nil,
	m_asset_group = nil,
	m_backup_group = nil,
	m_current_wave = nil,
}

-- The spawner.
spawn = chunks.class.spawner( MainGame, LuaMainGame )

local min_camera_x = 0
local max_camera_x = 0

function LuaMainGame:created(game)
	print("LuaMainGame:created")

	self.m_buttons_unique = {}
	self.m_current_wave = UserData.instance():getCurrentWave()

	local game_window = game:getWindow()

	-- Set clearing info.
	--game.clear_color    = constants.clear_color
	--game.clear_color_on = constants.clear_color_on
	--game.clear_depth_on = constants.clear_depth_on

	-- Scene camera.
	do
		local camera_actor = game:getScene():spawn(ActorBase())

		local camera = camera_actor:spawn( OrthographicCamera() )

		camera.viewport_offset = constants.game_camera.scene.viewport_offset
		camera.viewport_size   = constants.game_camera.scene.viewport_size
		camera.screen_size     = constants.game_camera.scene.screen_size
		camera.world_size      = constants.game_camera.scene.world_size
		camera.clipping_planes = constants.game_camera.scene.clipping_planes
		--camera.location = v3f(257, Map.get():getCameraYGivenX(257), 0);
		camera.clear_color     = constants.clear_color
		camera.clear_color_on  = constants.clear_color_on
		camera.clear_depth_on  = constants.clear_depth_on
		camera.scene           = game:getScene()

		game.scene_camera_actor = camera_actor
		self.m_scene_camera_actor = camera_actor
        self.m_scene_camera = camera_actor
		self.m_actual_camera = camera
	end

	-- Window camera.
	do
		local camera = game:getWindow():spawn( OrthographicCamera() )

		camera.viewport_offset = constants.game_camera.window.viewport_offset
		camera.viewport_size   = constants.game_camera.window.viewport_size
		camera.screen_size     = constants.game_camera.window.screen_size
		camera.world_size      = constants.game_camera.window.world_size
		camera.clipping_planes = constants.game_camera.window.clipping_planes
		camera.scene           = game_window

		game_window.camera = camera
	end

    self.m_backup_group = ResourceGroup("texture")
    Resources.get():addResourceGroup(self.m_backup_group)

    self.m_asset_group = ResourceGroup("texture")
    Resources.get():addResourceGroup(self.m_asset_group)

	--Resources.get():load("?://texture/game_atlas", TextureResource.class_info())

	--Set initial smithy level
	self.m_iron_level = math.max(tonumber(UserData.instance():getTypeLevel("smithy")), 1)
	for _ = 2, self.m_iron_level do
		self.m_iron_rate = self.m_iron_rate * globals.iron_rate_upgrade_factor
	end

	self.m_game_window = game_window:spawn(chunks.main_game_window.spawn())
    game.game_window = self.m_game_window

    local tutorial_window = game_window:spawn(Window())
    tutorial_window.size = constants.game_camera.window.world_size
    tutorial_window.receive_input = false
    tutorial_window.hidden = true
    self.m_tutorial_window = tutorial_window

    chunks.load.shuffleMainSoundSet()
    util.shuffleSpawnSlots()

	self:setupMapData()
	self:startDifficultyTracking()
	self:spawnInitialUnits()
	self:spawnWave()
	--self:spawnTestWave()
	self.m_game_window.lua:showCurrentWaveLabel()

	local current_wave = UserData.instance():getCurrentWave()
	local wave_name = string.format("wave_%02d", current_wave)
	local song = assets.findSongByPathName( chunks.waves[wave_name].song or "" ) or assets.music.in_game_1
	music.checkOrPlaySong(song)

	Resources.get():removeResourceGroup(self.m_asset_group)

	engine:setPauseCallback(
		function()
			if self.owner.paused == false and self.m_game_over == false then
				self.m_game_window.lua:showPauseWindow()
			end
		end
	)

	if engine.g_god_mode then
		self.m_iron_rate = 20

		if self.m_ash:get() then
			local ash_health = PartHealthIndex(self.m_ash):get()
			if ash_health then
				ash_health.health = 10000000
			end
		end

        --game.delta_time_factor = 2.0
	end

    --startFrameStats()
    --saveFrameStats(4*60.0)

	self:incrementWavePlayCount()
end

function LuaMainGame:incrementWavePlayCount()
	local play_count_id = globals.wave_play_count_id .. self.m_current_wave
	local new_count = util.getNumberValue(play_count_id) + 1

	util.setNumberValue(play_count_id, new_count)
end

function LuaMainGame:getWavePlayCount()
	local play_count_id = globals.wave_play_count_id .. self.m_current_wave

	return util.getNumberValue(play_count_id, 1)
end

function LuaMainGame:tick(game, dt)
	--don't advance stuff if the game is paused
	if game.paused then
		dt = 0
	end

	--don't do anything beyond this point if the game is paused
	if dt == 0 then
		return
	end

	--do normal game tick stuff here
	self:processIron(dt)

	local ash_object = self.m_ash:get()
	--local necro_object = self.m_necro:get()

	--[[
	if not ash_object or ash_object:isDead() or not necro_object or necro_object:isDead() then
		self:gameOver(GameObject.TEAM_ENEMY)
	end
	]]

	--Keep camera fixed on Ash
	if ash_object then
		local camera = self.m_scene_camera_actor
		local ash_location_x = ash_object.location.x
		local camera_location_x = camera.location.x
		if ash_location_x ~= camera_location_x then
			camera_location_x = clampf(ash_location_x, min_camera_x, max_camera_x)
			camera.location = pool.get_v3f(camera_location_x, Map.get():getCameraYGivenX(camera_location_x), camera.location.z)
		end
	end

	if self.m_current_wave <= globals.final_wave then
		if self.m_last_unit_spawned and game:getNumObjectsOnTeam(GameObject.TEAM_ENEMY)==0 then
			self.m_last_unit_spawned = false

			self:gameOver(GameObject.TEAM_ALLY)
		end
	end


	--[[
	if engine.g_god_mode then
		self.m_iron_rate = 20

		if ash_object then
			local ash_health = PartHealthIndex(ash_object:getObjectIndex()):get()
			if ash_health then
				ash_health.health = 1000000
			end
		end
	end
	]]
end

function LuaMainGame:onInputEvents(game, events)
	-- Input handling for the android back button.
	if AndroidEvent and self.m_game_over == false then
		for event in events:all() do
			local e = AndroidEvent.fast_cast( event )
			if e and e.type == AndroidEvent.AndroidEvent_BackButton then
				util.playSoundAtCamera("?://" .. assets.sounds.button_click_wav.path)
				self.m_game_window.lua:showHidePauseWindow()
				break
			end
		end
	end

	--print("some kind of touch happened")
	if self.m_game_window.lua:isSelectingSpellTarget() then
		self:doSelectSpellTargetInput(game, events)
	else
		self:doNormalInput(game, events)
	end
end

function LuaMainGame:destroy(game)
	util.tryToAutoSelectObjects()

	engine:clearGameSpawnStack()   -- Clear the game stack once we quit the main game
	Resources.get():unload(self.m_asset_group)

    Resources.get():removeResourceGroup(self.m_backup_group)
	Resources.get():unload(self.m_backup_group)

	-- submit updates to leaderboards
	local gc = GameCenterService.get()
	if gc then
		gc:postScore(UserData.instance():getTotalGold(), "ldr_most_gold_ever", function() end)
		
		gc:postScore(UserData.instance():getEndlessWaveScore(), "ldr_endless_mode_survivor", function() end)
	else
		print("LuaMainGame:destroy: gamecenter service unavailable")
	end
	achievements.goldChanged()

	engine:setPauseCallback(nil)
end

function LuaMainGame:doSelectSpellTargetInput(game, events)
	local camera = self.m_scene_camera_actor

	for event in events:all() do
		local touch = TouchEvent.fast_cast(event)
		if touch and (touch.type == TouchEvent.TouchStart and self.m_window_touch_id == -1) then
			self.m_window_touch_id = touch.id
			--print("make sure that removing touch.eaten is ok") --touch.eaten = true
			self:updateSpellTargetLocation(camera:screenToWorld(touch.location))

		elseif touch and touch.type == TouchEvent.Touching and touch.id == -1 then
			self.m_window_touch_id = touch.id
			--print("make sure that removing touch.eaten is ok") --touch.eaten = true
			self:updateSpellTargetLocation(camera:screenToWorld(touch.location))

		elseif touch and touch.type == TouchEvent.Touching and touch.id == self.m_window_touch_id then
			--print("make sure that removing touch.eaten is ok") --touch.eaten = true
			self:updateSpellTargetLocation(camera:screenToWorld(touch.location))

		elseif touch and touch.type == TouchEvent.TouchEnd and touch.id == self.m_window_touch_id then
			self.m_window_touch_id = -1
			if not self.m_ash:get() then
				return
			end
			--print("make sure that removing touch.eaten is ok") --touch.eaten = true

			self:spellTargetSelected(camera:screenToWorld(touch.location))
		end
	end
end

function LuaMainGame:doNormalInput(game, events)
	local camera = self.m_scene_camera_actor
	local map = Map.get()
	local orders = PartOrdersIndex(self.m_ash):get()

	if self:isAshBlocked() == true then
		self.m_window_touch_id = -1
		return
	end

	for event in events:all() do
		local touch = TouchEvent.fast_cast(event)
		if touch and touch.type == TouchEvent.TouchStart and self.m_window_touch_id == -1 then
			self.m_window_touch_id = touch.id
			if not self.m_ash:get() then
				return
			end

			--print("make sure that removing touch.eaten is ok") --touch.eaten = true

			local target_loc
			if touch.location.x < 0 then
				target_loc = map:getFirstPoint(GameObject.TEAM_ALLY)
				local render = PartRenderIndex(self.m_ash)
				render:get().render_actor.scale = pool.get_v3f(-1,1,1)
			else
				target_loc = map:getLastPoint(GameObject.TEAM_ALLY)
				local render = PartRenderIndex(self.m_ash)
				render:get().render_actor.scale = pool.get_v3f(1,1,1)
			end

			orders:clearOrders()
			orders:addOrder(MoveAlongPathOrder(target_loc), PartOrders.QUEUE_END)


		elseif touch and touch.type == TouchEvent.Touching and self.m_window_touch_id == -1 then
			self.m_window_touch_id = touch.id
			if not self.m_ash:get() then
				return
			end

			--print("make sure that removing touch.eaten is ok") --touch.eaten = true

			local target_loc
			if touch.location.x < 0 then
				target_loc = map:getFirstPoint(GameObject.TEAM_ALLY)
				local render = PartRenderIndex(self.m_ash)
				render:get().render_actor.scale = pool.get_v3f(-1,1,1)
			else
				target_loc = map:getLastPoint(GameObject.TEAM_ALLY)
				local render = PartRenderIndex(self.m_ash)
				render:get().render_actor.scale = pool.get_v3f(1,1,1)
			end

			orders:clearOrders()
			orders:addOrder(MoveAlongPathOrder(target_loc), PartOrders.QUEUE_END)

		elseif touch and (touch.type == TouchEvent.TouchEnd or touch.type == TouchEvent.TouchCancelled) and touch.id == self.m_window_touch_id then
			self.m_window_touch_id = -1
			if not self.m_ash:get() then
				return
			end
			--print("make sure that removing touch.eaten is ok") --touch.eaten = true
			local render = PartRenderIndex(self.m_ash)
			render:get().render_actor.scale = pool.get_v3f(1,1,1)
			orders:clearOrders()
			local logic = PartLogicIndex(self.m_ash)
			logic:get():requestProcess(1/30)
		end
	end
end

function LuaMainGame:isAshBlocked()
	local orders = PartOrdersIndex(self.m_ash):get()
	if orders == nil then
		return false
	end

	return orders:isBlocked()
end


function LuaMainGame:onCastSpell(caster, spell_type, location)
	--print("CASTING:", caster:get().label, spell_type, location)

	local on_cast_event_sounds = chunks.load.getStat("spells", spell_type, "on_cast_event_sound")
	if on_cast_event_sounds then
		local sounds = util.splitString(on_cast_event_sounds, ";")
		util.playRandomSoundFromTable(sounds, {priority=10.0})
	end

	local script_name = chunks.load.getStat("spells", spell_type, "script")
	local spell_cast = chunks[script_name].cast(caster, spell_type, location)

	if spell_cast == true then
		achievements.spellCast(spell_type)
		analytics.sendSpellCast(spell_type)
	end

	return spell_cast
end


function LuaMainGame:castCurrentSpell(spell_type, location)
	local ash_orders = PartOrdersIndex(self.m_ash):get()

	if ash_orders == nil then
		return
	end

	local render = PartRenderIndex(self.m_ash):get()
	if render == nil then
		return
	end
	render.render_actor.scale = pool.get_v3f(1,1,1)  -- Make ash face the enemy when casting

	local on_button_press_sounds = chunks.load.getStat("spells", spell_type, "on_button_press_sound")
	if on_button_press_sounds then
		local sounds = util.splitString(on_button_press_sounds, ";")
		util.playRandomSoundFromTable(sounds, {priority=10.0})
	end

	local auto_target_offset = chunks.load.getStat("spells", spell_type, "auto_target_offset") or 0.0
	local new_location = location or self.m_ash:get().location + globals.non_user_targeted_spell_offset + pool.get_v3f(auto_target_offset, 0.0, 0.0)
	local animation_name = chunks.load.getStat("spells", spell_type, "animation") or "cast"
	ash_orders:clearOrders()
	ash_orders:addOrder(CastOrder(spell_type, new_location, animation_name), PartOrders.QUEUE_END)
end


function LuaMainGame:updateSpellTargetLocation(location)
	if self.m_spell_target_area_actor == nil then
		self.m_spell_target_area_actor = self.owner:getScene():spawn(SpriteActor())
		self.m_spell_target_area_actor.color = pool.get_colorf(1.0, 0.0, 0.0, 0.25)
		self.m_spell_target_area_actor.texture = "builtin://white"
		self.m_spell_target_area_actor.size = pool.get_v2f(self.m_game_window.lua:getSpellRange() * 2, constants.game_camera.scene.world_size.y)
	end

	self.m_spell_target_area_actor.location = pool.get_v3f(location.x, self.m_scene_camera_actor.location.y, constants.game_camera.scene.clipping_planes.x)
end


function LuaMainGame:spellTargetSelected(location)
	self:destroySpellTargetDisplay()
	self.m_game_window.lua:spellTargetSelected(location)
end


function LuaMainGame:destroySpellTargetDisplay()
	self.m_spell_target_area_actor:destroy()
	self.m_spell_target_area_actor = nil
end


function LuaMainGame:spawnInitialUnits()
	local game = self.owner
	local map = Map.get()

	-- Spawn Ash
	do
		self.m_ash = util.createUnitOnPath(game, "ash")
		local ash = self.m_ash:get()
		local ash_health = PartHealthIndex(self.m_ash):get()
		local ash_inventory = PartInventoryIndex(self.m_ash):get()

		local ash_render = PartRenderIndex(self.m_ash):get()
		ash_render.render_actor.location = pool.get_v3f(0, 0, 0)

		--Regen Actions
		local damage_amount = -1.0 * chunks.load.getStat("units", "ash", "regen_rate")
		local damage_info = DamageInfo(damage_amount, DamageInfo.PHYSICAL_DAMAGE)
		local damage_packet = DamagePacket(self.m_ash, damage_info)

		local loop = LoopAction()
		loop:addAction(LuaAction(1, function()
			ash_health:takeDamage(damage_packet)
		end
		))

		local regen_actor = ash:spawn(ActorBase())
		regen_actor:addAction(loop)

		--Equip Weapons
		local destroyCurrentWeapon = function()
			local stump = ash_inventory:getEquippedItemAtSlot(PartInventory.WEAPON_EQUIP_SLOT)
			ash_inventory:removeItem(stump)
			stump:get():getObjectIndex():get():destroy()
		end

		if not UserData.instance():getLocked("chainsaw") then
			destroyCurrentWeapon()
			local level = ash:getLevel()
			local weapon = game:createGameObject("weapon", "chainsaw", level)
			local weapon_item = PartItemIndex(weapon)
			ash_inventory:addItem(weapon_item)
			ash_inventory:equipItem(weapon_item)

		elseif not UserData.instance():getLocked("iron_fist") then
			destroyCurrentWeapon()
			local level = ash:getLevel()
			local weapon = game:createGameObject("weapon", "iron_fist", level)
			local weapon_item = PartItemIndex(weapon)
			ash_inventory:addItem(weapon_item)
			ash_inventory:equipItem(weapon_item)
		end
	end
end

function LuaMainGame:startDifficultyTracking()
	--Difficulty tracker for this wave
	self.m_front_line_tracker = { count = 0, total = 0 }

	--Get average from previous waves
	local ud = UserData.instance()
	local current_wave = ud:getCurrentWave()
	local average_front_line = 0
	local total = 0
	for wave_i = 1, current_wave do
		local wave_total = tonumber(ud:getValue("wave_" .. wave_i .. "_front_line_total")) or 0
		local wave_count = tonumber(ud:getValue("wave_" .. wave_i .. "_front_line_count")) or 1
		total = total + (wave_total / wave_count)
--		print(string.format("wave %i: avg front line = %gx", wave_i, (wave_total / wave_count)))
	end
	average_front_line = total / current_wave

	--Determine how much to scale difficulty for the whole wave
	local scalar_in_far_half = math.max(0, average_front_line - 0.5) * 2
	scalar_in_far_half = math.pow(scalar_in_far_half, wave_difficulty_curve)

	--Comment this line out to disable wave difficulty scaling
	if enable_wave_difficulty_scaling then
		self.m_difficulty_dt_factor_lo = difficulty_dt_factor_lo + (difficulty_dt_factor_hi - difficulty_dt_factor_lo) * scalar_in_far_half
	else
		self.m_difficulty_dt_factor_lo = difficulty_dt_factor_lo
	end

--	print(string.rep("#", 80))
--	print(string.format("total avg front line = %gx", average_front_line))
--	print(string.format("Setting base difficulty for wave to %gx", self.m_difficulty_dt_factor_lo))
--	print(string.rep("#", 80))
end

function LuaMainGame:finishDifficultyTracking()
	local ud = UserData.instance()
	local current_wave = ud:getCurrentWave()
	local wave_total = tonumber(ud:getValue("wave_" .. current_wave .. "_front_line_total")) or 0
	local wave_count = tonumber(ud:getValue("wave_" .. current_wave .. "_front_line_count")) or 0
	local avg_for_this_play = self.m_front_line_tracker.total / self.m_front_line_tracker.count
	ud:setValue("wave_" .. current_wave .. "_front_line_total", tostring(wave_total + avg_for_this_play))
	ud:setValue("wave_" .. current_wave .. "_front_line_count", tostring(wave_count + 1))

--	print(string.rep("*", 80))
--	print(string.format("Avg front line for this wave: %gx", avg_for_this_play))
--	print(string.rep("*", 80))
end

function LuaMainGame:setupMapData()
	local game = self.owner
	local map = Map.get()
	local current_wave = UserData.instance():getCurrentWave()

	-- Camera Bounds
	do
		local path = map:getCameraPath().values
		local default = 480
		local size = constants.game_camera.scene.world_size.x
		local diff = size - default
		min_camera_x = path[1].x + (diff / 2)
		max_camera_x = path[#path].x
	end


	-- Spawn objects
	local initial_objects = map:getInitialObjects().values

	local spawn_objects = function(object_list, type, spawn_function)
		if object_list == nil then
			return {}
		end

		local objects = {}
		for _,spawn_data in pairs(object_list) do

			spawn_data.info.texture = string.gsub( spawn_data.info.texture, "texture://", "?://texture/" )

			if type == spawn_data.attrs.type then
				local spawned_object = spawn_function(spawn_data)
				if spawned_object ~= nil then
					table.insert(objects, spawned_object)
				end
			end
		end

		return objects
	end

	local spawn_units = function (type)
		return spawn_objects(initial_objects, type, function (spawn_data)
			--[[
			local unlock_wave = chunks.load.getStat("units", spawn_data.attrs.type, "unlock_wave") or 1
			if current_wave < unlock_wave then
				return nil
			end
			]]

			local spawn_count = chunks.load.getStat("units", spawn_data.attrs.type, "count") or 1
			local spawn_index = spawn_data.attrs.spawn_index or 1
			if spawn_index > spawn_count then
				return nil
			end

			local unit_index = game:createGameObject("unit", spawn_data.attrs.type)
			if unit_index:get() then
				unit_index:get().location = spawn_data.info.location
				unit_index:get().scale = spawn_data.info.scale

				local render_part = PartRenderIndex(unit_index):get()
				if render_part and render_part:getGrannyActor() then
					unit_index:get().location = unit_index:get().location - pool.get_v3f(0.0, spawn_data.size.y * 0.5, 0.0)
				end

				return unit_index
			end

			return nil
		end)
	end

	local sprite_spawner = function(spawn_data)
			local background = nil

			if spawn_data.attrs.mesh then
				background = game:getScene():spawn(ActorBase(), spawn_data.info)
				background.scale = v3f(spawn_data.size.x, spawn_data.size.y, 1.0)
				background.mesh = "?://" .. spawn_data.attrs.mesh
			else
				background = game:getScene():spawn(SpriteActor(), spawn_data.info)
				background.size = spawn_data.size

			end

			background.texture.filter = TextureResource.Filter_Nearest

			return background
	end

	local spawn_sprites = function (type)
		return spawn_objects(initial_objects, type, sprite_spawner)
	end

	-- Spawn Background pieces
	do
		if engine.platform.constants.background_loader then
			spawn_objects(initial_objects, "background", function(spawn_data)
				return engine.platform.constants.background_loader.spawn(engine, chunks, game, spawn_data, sprite_spawner)
			end)
		else
			spawn_sprites("background")
		end
	end

	-- Spawn Foreground pieces
	do
		spawn_sprites("foreground")
	end


	-- Spawn blacksmith
	spawn_units("smithy")

	-- Spawn Wall Archers
	if not UserData.instance():getLocked("inner_wall_archer") then
		spawn_units("inner_wall_archer")
	end

	-- Spawn catapults
	if not UserData.instance():getLocked("catapults") then
		spawn_units("catapults")
	end


	-- Spawn the Pit
	if not UserData.instance():getLocked("the_pit") then
		self.m_pit = spawn_units("the_pit")[1]
        if self.m_pit then
            self.m_pit_closed = spawn_sprites("the_pit_closed")[1]
            self.m_pit_open = spawn_sprites("the_pit_open")[1]
            self.m_pit_open.hidden = true
            self.m_pit_open.location = self.m_pit_closed.location
        end
	end

	-- Spawn Necronomicon
	do
		self.m_necro = spawn_units("necronomicon")[1]
	end

	--[[ Spawn BG Fill
	do
		spawn_objects(initial_objects, "bg_fill", function (spawn_data)
			local background = game:getScene():spawn(SpriteActor(), spawn_data.info)
			background.size = spawn_data.size
			background.material = Material.default_opaque
			background.texture = "builtin://white"
			background.color = colorf(0,0,0,1)
			background.texture.filter = TextureResource.Filter_Nearest
			return background
		end)
	end
	]]

	-- Spawn Moon
	do
		spawn_objects(initial_objects, "moon", function (spawn_data)
			local background = self.m_scene_camera_actor:spawn(SpriteActor(), spawn_data.info)
			background.size = spawn_data.size
			background.material = Material.default_opaque
			background.texture.filter = TextureResource.Filter_Nearest
			return background
		end)
	end
end

function LuaMainGame:spawnWave()
	local game = self.owner
	local far_left = Map.get():getFirstPoint(GameObject.TEAM_ALLY).x
	local far_right = Map.get():getLastPoint(GameObject.TEAM_ALLY).x
	local center = (far_left + far_right) / 2

	local wave_action_actor = game:getScene():spawn(ActorBase())
	wave_action_actor.hidden = true
	self.m_wave_action_actor = wave_action_actor

	local spawn_enemy = function(enemy, level)
		local object = util.createUnitOnPath(game, enemy, level)
		self.m_spawned_so_far = self.m_spawned_so_far + 1
		return object
	end

	--Difficulty management (this is called every time an enemy is spawned)
	local adjust_difficulty = function()
		--Tweak these as needed
		local ally_indices = game:getObjectsInRange(GameObject.TEAM_ALLY, far_left, far_right).values
		local enemy_indices = game:getObjectsInRange(GameObject.TEAM_ENEMY, far_left, far_right).values

		local skip_allies = {"smithy", "inner_wall_archer", "catapults", "necronomicon"} --Ally units to ignore in difficulty calculations

		--Determine how far our allies are out in the level
		local farthest_ally = far_left
		for _, i in ipairs(ally_indices) do
			local ally = i:get()
			local skip = false
			for _, j in ipairs(skip_allies) do
				if j == ally.label then
					skip = true
					break
				end
			end
			if not skip then
				farthest_ally = math.max(farthest_ally, ally.location.x)
			end
		end

		--Determine how close the enemies have come in the level
		local closest_enemy = far_right
		for _, i in ipairs(enemy_indices) do
			closest_enemy = math.min(closest_enemy, i:get().location.x)
		end

		--Determine the front line by averaging positions
		local total_length = (far_right - far_left)
		local front_line = math.max((farthest_ally + closest_enemy) / 2, farthest_ally) --Cap the line at the farthest ally, in case it is Ash and he is past the enemies
		local front_line_ratio = (front_line - far_left) / total_length
		local difficulty = math.pow(front_line_ratio, difficulty_dt_curve)
		local dt_factor = self.m_difficulty_dt_factor_lo + (difficulty_dt_factor_hi - self.m_difficulty_dt_factor_lo) * difficulty
		wave_action_actor.delta_time_factor = dt_factor

		--Tracking for storage
		self.m_front_line_tracker.total = self.m_front_line_tracker.total + front_line_ratio
		self.m_front_line_tracker.count = self.m_front_line_tracker.count + 1
		print(string.format("difficulty = %gx, line = %g%%", difficulty, front_line_ratio * 100))
		--print(string.format("Delta time factor is now %f (farthest_ally=%f, closest_enemy=%f, front_line=%f, difficulty=%f)", dt_factor, farthest_ally-far_left, closest_enemy-far_left, front_line_ratio, difficulty))
	end

	local wave_data = util.getCurrentWaveData()

	if wave_data == nil then
		return
	end

	local wave_script_actor = {}
	local wave_script = wave_data.script or "default"
	if chunks[wave_script] then
		wave_script_actor = ActorBase()
		chunks[wave_script].bindToObject(wave_script_actor)
		--wave_action_actor:spawn(wave_script_actor)
		game:getScene():spawn(wave_script_actor)
		wave_script_actor.hidden = true
	end

	self.m_units_this_wave = util.getNumUnitsInCurrentWave()

	local push_id = 0
	for i,v in ipairs(wave_data.unit_stat) do
		local delay = v.delay
		local unit = v.unit
		local level = v.level
		local tag = v.tag
		wave_action_actor:addAction(LuaAction(delay, function ()
			if tag == "push" then
				push_id = push_id + 1
				self.m_game_window.lua:showWavePushEffect(push_id)
			end

			self.m_most_recent_enemy = spawn_enemy(unit, level)
			adjust_difficulty()
		end))

		if wave_script_actor.lua and wave_script_actor.lua.unitSpawned then
			wave_action_actor:addAction(LuaAction(0, function ()
				wave_script_actor.lua:unitSpawned(i, self.m_most_recent_enemy, tag)
			end))
		end
	end

	wave_action_actor:addAction(LuaAction(0, function ()
		self.m_last_unit_spawned = true;
		if wave_script_actor.lua and wave_script_actor.lua.lastUnitSpawned then
			wave_script_actor.lua:lastUnitSpawned()
		end ;
		print("LAST UNIT SPAWNED");
	end))

	--wave_action_actor:addAction(DestroyAction(0))

	self.m_wave_script_actor = wave_script_actor

	if wave_script_actor.lua and wave_script_actor.lua.spawnWave then
		wave_script_actor.lua:spawnWave(game, self.m_wave_action_actor)
	end

	achievements.waveStarted(game, UserData.instance():getCurrentWave())
end

function LuaMainGame:unitDied(gameObjectIndex, weapon_index)

	local ash_object = self.m_ash:get()
	local necro_object = self.m_necro:get()

	if not ash_object or ash_object:isDead() or not necro_object or necro_object:isDead() then
		self:gameOver(GameObject.TEAM_ENEMY)
	end

	local unit = gameObjectIndex:get()
	if unit ~= nil then
		if unit.team == GameObject.TEAM_ENEMY then
			self.m_enemies_killed = self.m_enemies_killed  + 1
			self.m_game_window.lua:updateWaveProgressBar()
		end

		local weapon_name = ""
		local weapon = weapon_index:get()
		if weapon ~= nil then
			weapon_name = weapon.label
		end
		achievements.unitDied(unit.label, unit.team, unit.location.x, weapon_name)
	end

	local wave_script_actor = self.m_wave_script_actor
	if wave_script_actor and wave_script_actor.lua and wave_script_actor.lua.unitDied then
		wave_script_actor.lua:unitDied(gameObjectIndex)
	end
end

function LuaMainGame:objectSpawned(object_index, category, type_name)
	local wave_script_actor = self.m_wave_script_actor
	if wave_script_actor and wave_script_actor.lua and wave_script_actor.lua.objectSpawned then
		wave_script_actor.lua:objectSpawned(object_index, category, type_name)
	end
end

function LuaMainGame:spawnTestWave()
	local far_left = v3f(Map.get():getFirstPoint(GameObject.TEAM_ALLY))
	local far_right = Map.get():getLastPoint(GameObject.TEAM_ALLY)
	local center = (far_left + far_right) / 2
	local scene = self.owner:getScene()

	-- create good guys
	local good_models = {
		chunks.units.spearman,
		chunks.units.peasant,
		chunks.units.archer,
		chunks.units.armored_guard,
		chunks.units.spearman,
		chunks.units.peasant,
		chunks.units.archer,
		chunks.units.armored_guard,
		chunks.units.spearman,
		chunks.units.peasant,
		chunks.units.archer,
		chunks.units.peasant,
		chunks.units.archer,
		chunks.units.armored_guard,
		chunks.units.spearman,
		chunks.units.peasant,
		chunks.units.archer,
		chunks.units.armored_guard,
		chunks.units.spearman,
		chunks.units.peasant,
		chunks.units.archer
	}
	local i = 0
	for k,v in ipairs(good_models) do

		local granny_actor = scene:spawn(GrannyActor())
		granny_actor:setSkeleton("?://" .. v.model)
		granny_actor.texture = "?://" .. v.texture
		if i % 2 == 0 then
			granny_actor.location = v3f(far_left) + v3f(0,75,400)
		else
			granny_actor.location = v3f(far_left) + v3f(0,-75,400)
		end
		granny_actor:playAnimation("attack_a", { loop_count=0, start_offset=math.random() })

		local weapon = scene:spawn(ActorBase())
		weapon.mesh = "?://objects/weapons/peasant_weapon.obj"
		weapon.texture = "?://objects/weapons/broom"

		granny_actor:attach(weapon, "bone_weapon")

		if weapon:getOwner().id == granny_actor.id then
			weapon:destroy()
		end

		far_left.x = far_left.x + 20
		i = i+1
	end

	-- create bad guys
	do

	end

	startFrameStats()
	saveFrameStats(60.0)
end

function LuaMainGame:canPayIronCost(query)
	if query > self.m_iron then
		return false
	end
	return true
end

function LuaMainGame:spendIron(query)
	if self:canPayIronCost(query) == true then
		local previous_iron = math.floor(self.m_iron)
		self.m_iron = self.m_iron - query
		local current_iron = math.floor(self.m_iron)
		self.m_game_window.lua:ironTotalChanged(previous_iron, current_iron)
		return true
	else
		return false
	end
end

function LuaMainGame:incrementIron(amount)
	if self.m_game_over == true then
		return
	end

	if amount <= 0 then
		return
	end

	local cap = self:getIronCap()

	local previous_iron = math.floor(self.m_iron)
	self.m_iron = math.min(cap, self.m_iron + amount)
	local current_iron = math.floor(self.m_iron)

	if previous_iron ~= current_iron then
		self.m_game_window.lua:ironTotalChanged(previous_iron, current_iron)
	end
end

function LuaMainGame:processIron(dt)
	if self.m_game_over == true then
		return
	end

	local amount = dt * self.m_iron_rate
	self:incrementIron(amount)
end

function LuaMainGame:getIronLevel()
	return clampi(self.m_iron_level, 1, #globals.iron_level_cap)
end

function LuaMainGame:getMaxIronLevel()
	return #globals.iron_level_cap
end

function LuaMainGame:getIronCap()
	local level = self:getIronLevel()
	local cap = globals.iron_level_cap[level]

	return math.floor(cap)
end

function LuaMainGame:getIronCapUpgradePrice()
	local level = self:getIronLevel()
	local price = globals.iron_level_cap_upgrade_price[level]

	return math.floor(price)
end

function LuaMainGame:canUpgradeIronCap()
	if self:isAtMaxIronCap() == true then
		return false
	end

	local upgrade_price = self:getIronCapUpgradePrice()
	if self:canPayIronCost(upgrade_price) == false then
		return false
	end

	return true
end

function LuaMainGame:isAtMaxIronCap()
	local cap = self:getIronCap()
	local max_level = self:getMaxIronLevel()
	local max_cap = globals.iron_level_cap[max_level]

	if cap < max_cap then
		return false
	end

	return true
end

function LuaMainGame:upgradeIronCap()
	if self:canUpgradeIronCap() == false then
		return false
	end

	local upgrade_price = self:getIronCapUpgradePrice()
	if self:spendIron(upgrade_price) == false then
		return false
	end

	self.m_iron_level = self.m_iron_level + 1
	self.m_iron_rate = self.m_iron_rate * globals.iron_rate_upgrade_factor

	achievements.ironCapUpgraded(self.m_iron_level, self:getMaxIronLevel())

	return true
end

function LuaMainGame:gameOver(team)
	--if this has already been called then just return
	if self.m_game_over then
		return
	end

	if self.m_wave_script_actor and self.m_wave_script_actor.lua and self.m_wave_script_actor.lua.gameOver then
		local result = self.m_wave_script_actor.lua:gameOver(team)
		if result == nil then
			result = true
		end

		if result == false then
			print("DON'T END THE WAVE!!!")
			return
		end
	end

	self.m_game_over = true

	self:finishDifficultyTracking()

	--self:killAll(GameObject.TEAM_COIN)
	local won = (team == GameObject.TEAM_ALLY)
	local wave_ended_state = globals.wave_ended_state.WON

    --One of these may have already played it death sound
    --let's make sure the other doesn't
    local necro = self.m_necro:get()
    local ash = self.m_ash:get()
    if necro == nil or necro:isDead() then
        if ash then
        	ash:shouldPlayOnDeathSound(false)
        end

		if won == false then
			wave_ended_state = globals.wave_ended_state.BOOK_DIED
		end
    elseif ash == nil or ash:isDead() then
        if necro then
        	necro:shouldPlayOnDeathSound(false)
        end

		if won == false then
			wave_ended_state = globals.wave_ended_state.ASH_DIED
		end
    end

	local award = nil

	local current_wave = UserData.instance():getCurrentWave()
	if team == GameObject.TEAM_ALLY then
		--self:killAll(GameObject.TEAM_ENEMY)
		self:setWinAnimations(team)
		util.playSoundAtCamera("?://" .. assets.sounds.cheer.path)
		self:playRandomWinSound()
		award = self:getAwardForWave(current_wave)
		self:awardGoldForWave(award)
		UserData.instance():setCurrentWave(current_wave+1)
	elseif team == GameObject.TEAM_ENEMY then
		--self:killAll(GameObject.TEAM_ALLY)
		local ash = self.m_ash:get()
		if ash then
			ash:die()
			self.onInputEvents = function() end
		end

		music.checkOrPlaySong(assets.music.prologue)
	end

	self.m_game_window.lua:gameOver(team, award)


	achievements.waveEnded(current_wave, won)
	self:sendWaveEndedAnalytics(wave_ended_state)
end

function LuaMainGame:killAll(team)
	local team_members = self.owner:getObjectsInRange(team, -1000000, 1000000).values
	for i,object_index in ipairs(team_members) do
		local object = object_index:get()
		local item = PartItemIndex(object_index):get()
		if item and object then
			object:clearActions()
			object:addAction(MoveOffsetAction(0.5, InterpAction.Ramp, pool.get_v3f(0, -250, 0)))
			object:addAction(DestroyAction(0))
		elseif object then
			object:die()
		end
	end
end

function LuaMainGame:setWinAnimations(team)
	local options = {
						loop_count = 0,
						ease_out_current_animation_time = 0.3,
						ease_in_time = 0.1,
						start_new_animation_if_the_same = false,
					}
	local team_members = self.owner:getObjectsInRange(team, -1000000, 1000000).values
		for i,object_index in ipairs(team_members) do

			local render = PartRenderIndex(object_index):get()
			if render and object_index ~= self.m_ash then
				local part_orders = PartOrdersIndex(object_index):get()
				if part_orders then
					part_orders:clearOrders()
					part_orders:requestProcess(0)
				end
				local part_logic = PartLogicIndex(object_index):get()
				if part_logic then
					part_logic:requestProcess(0)
				end

				render.render_actor:addAction(LuaAction(math.random(), function (inner)
					render:playAnimation("win", options)
				end))
			end
		end
end


function LuaMainGame:addGold(amount)
	local new_gold = UserData.instance():getGold() + amount
	UserData.instance():setGold(new_gold)

	self.m_gold_collected = self.m_gold_collected + amount
end

function LuaMainGame:awardGoldForWave(award)
	local instance = UserData.instance()
	instance:setGold( instance:getGold() + award )
end

function LuaMainGame:getAwardForWave(force_wave)
	local current_wave = force_wave or UserData.instance():getCurrentWave()
	local wave_name = string.format("wave_%02d", current_wave)
	local award = chunks.waves[wave_name].completion_bonus
	return award
end

function LuaMainGame:playRandomWinSound()
	local sounds = {
						assets.sounds.boomstick,
						assets.sounds.buckle_up_bonehead,
						assets.sounds.cry_to_mama,
						assets.sounds.give_me_some_sugar_baby,
						assets.sounds.guy_with_the_gun,
						assets.sounds.hail_to_the_king_baby,
						assets.sounds.names_ash_housewares,
						assets.sounds.next_one_of_you_primates,
						assets.sounds.pillow_talk,
						assets.sounds.shop_s_mart,
						assets.sounds.hello_mr_fancy_pants,
						assets.sounds.london_bridge,
				   }

	local win_sound = sounds[math.random(1, #sounds)]
	if win_sound ~= nil then
		--assets.playSound(win_sound)
		self.m_actual_camera:addAction(LuaAction(1.5, function() util.playSoundAtCamera("?://" .. win_sound.path) end))
	end
end

function LuaMainGame:doCameraShake(x_offset, y_offset)
	if self.m_game_over == true then
		return
	end

	--print("do camera shake", x_offset, y_offset)
	self.m_actual_camera:addAction(MoveAction(1/30, InterpAction.Lerp, pool.get_v3f(x_offset, y_offset, 0)))
	self.m_actual_camera:addAction(MoveAction(1/30, InterpAction.Lerp, pool.get_v3f(0, 0, 0)))
end

function LuaMainGame:isObjectOnScreen(object_index)
	local object = object_index:get()

	if object == nil then
		return false
	end

	if object:isDead() == true then
		return false
	end

	local x_location = object:getWorldLocation().x
	local camera_x = self.m_scene_camera_actor.location.x
	local camera_half_size_x = self.m_actual_camera.world_size.x * 0.5

	if x_location > camera_x + camera_half_size_x then
		return false
	end

	if x_location < camera_x - camera_half_size_x then
		return false
	end

	return true
end

function LuaMainGame:sendWaveEndedAnalytics(wave_ended_state)
	-- do flurry stuff here
	local user_data = UserData.instance()
	local selected_unit_list = util.loadStringList(globals.selected_units_id)
	local selected_spell_list = util.loadStringList(globals.selected_spells_id)

	local get_type_data = function(type_name)
		if type_name == nil or #type_name == 0 then
			return ""
		end
		local level = user_data:getTypeLevel(type_name)

		return type_name .. "," .. level
	end

	local wave_analytics_data = {
		PLAY_COUNT = self:getWavePlayCount(),
		EXIT = wave_ended_state.flurry_info,
		UNIT1 = get_type_data(selected_unit_list[1]),
		UNIT2 = get_type_data(selected_unit_list[2]),
		UNIT3 = get_type_data(selected_unit_list[3]),
		UNIT4 = get_type_data(selected_unit_list[4]),
		UNIT5 = get_type_data(selected_unit_list[5]),
		SPELL1 = get_type_data(selected_spell_list[1]),
		SPELL2 = get_type_data(selected_spell_list[2]),
		SMITHY = user_data:getTypeLevel("smithy") .. "," .. self.m_iron_level
	}

	analytics.sendWaveEnd(self.m_current_wave, wave_analytics_data)
end
