-- module --
module(...,package.seeall)

-- includes --
local engine    = engine.instance
local chunks    = engine.chunks
local constants = engine.platform.constants
local assets    = chunks.assets
local pool      = chunks.pool
local util      = chunks.util
local load      = chunks.load
local globals   = chunks.globals

local stat_string = lstring("stat")

LuaMainGameWindow = chunks.class.new
{
	m_main_game = nil,
	m_ash_health_bar = nil,
	m_necro_warning_icon = nil,
	m_iron_label = nil,
	m_gold_label = nil,
	m_selecting_spell_target = false,
	m_select_spell_target_info_window = nil,
	m_spawn_buttons = nil,
	m_blood_overlay_base = nil,
	m_smithy_button = nil,
	m_spell_buttons = nil,
	m_current_spell = nil,
	m_asset_group = nil,
	m_danger_icons = nil,
	m_enemy_progress_bar = nil,
	m_enemy_progress_bar_container = nil,
	m_danger_popup = nil,
	m_pause_button = nil,
	m_pause_window = nil,
	m_platform_ui_handle = nil
}

-- The spawner.
spawn = chunks.class.spawner( Window, LuaMainGameWindow )

function LuaMainGameWindow:created(window)
	window.hidden = true
	window.receive_input = false
	window.size = constants.game_camera.window.world_size

	local game = window:getGame()
	self.m_main_game = game
	local current_wave = UserData.instance():getCurrentWave()

	self.m_asset_group = ResourceGroup("texture")
	Resources.get():addResourceGroup(self.m_asset_group)


	-- create HUD elements
	do
		-- create Pause button
		do
			local button = util.createButtonWithTexture(window, assets.textures.pause_button)

			button.anchor = { Anchor.Anchor_TL, Anchor.Anchor_TL, pool.get_v3f(1.0, -45.0, 1.0) }

			-- Spawn the pause window the game
			button.on_release = function()
				--assets.playSound(assets.sounds.button_click_wav)
				util.playSoundAtCamera("?://" .. assets.sounds.button_click_wav.path)
				self:showPauseWindow()
			end

			self.m_pause_button = button
		end
	end

	-- create Spawn Buttons
	do
		local buttons = {}
		local possible_unit_list = util.loadStringList(globals.selected_units_id)
		for i,unit_type in ipairs(possible_unit_list) do
			local unlock_wave = load.getStat("units", unit_type, "unlock_wave")
			if current_wave >= unlock_wave then
				local button = window:spawn(chunks.cooldown_button.spawn())
				--button.lua.m_unit_type = unit_type
				--assets.setTexture(button, assets.textures.temp_spawn_button)
				button.lua:setUnit(unit_type)
				button.anchor = {Anchor.Anchor_B, Anchor.Anchor_BL, pool.get_v3f(button.size.x*(i-0.5), 0.0, 1.0)}

				table.insert(buttons, button)
			end
		end

		self.m_spawn_buttons = buttons
	end

	-- create Spell Buttons
	do
		local buttons = {}
		local possible_spell_list = util.loadStringList(globals.selected_spells_id)
		
		-- if for some reason there is no saved value for selected spells, default to the first two: super_boomstick_blast and vortex
		-- this can occur the first time the user starts a new game after installing or due to a currently unknown bug that appears to sometimes cause selected spells not to be saved
		if #possible_spell_list == 0 then
			possible_spell_list = {
				"super_boomstick_blast",
				"vortex"
			}
		end
		
		local num_spell_slots_unlocked = util.getNumSpellSlotsUnlocked()
		for i,spell_type in ipairs(possible_spell_list) do
			if i > num_spell_slots_unlocked then
				break
			end
			
			local unlock_wave = load.getStat("spells", spell_type, "unlock_wave")
			
			-- If the player has reached the wave at which spells are unlocked
			if current_wave >= unlock_wave then
				local button = window:spawn(chunks.cooldown_button.spawn())
				local callback = function()
					if self:isSelectingSpellTarget() then
						return false
					end

					if game.lua:isAshBlocked() then
						return false
					end

					local user_targets = load.getStat("spells", spell_type, "user_targets")
					self.m_current_spell = spell_type
					if user_targets then
						self:setSelectingSpellTarget(true)
						return true
					end

					return self:castCurrentSpell()
				end

				button.lua:setSpell(spell_type, callback)

				table.insert(buttons, button)
			end
		end
		
		local num_buttons_in_row = math.min(2, #buttons)
		for i,button in ipairs(buttons) do
			if engine.platform.device == "ipad" then
				if i <= num_buttons_in_row then
					-- Put buttons along the bottom row first from the right...
					local x = -button.size.x * (num_buttons_in_row - i)
					button.anchor = {Anchor.Anchor_BR, Anchor.Anchor_BR, pool.get_v3f(x, 0.0, 1.0)}
				else
					-- ...then put them up the right side
					local y = button.size.y * (i - num_buttons_in_row)
					button.anchor = {Anchor.Anchor_BR, Anchor.Anchor_BR, pool.get_v3f(0.0, y, 1.0)}
				end
			else
				 if i <= num_buttons_in_row then
					 -- Put buttons along the top row first from the right...
					 local x = -button.size.x * (num_buttons_in_row - i)
					 button.anchor = {Anchor.Anchor_TR, Anchor.Anchor_TR, pool.get_v3f(x, 0.0, 1.0)}
				 else
					 -- ...then put them down the right side
					 local y = -button.size.y * (i - num_buttons_in_row)
					 button.anchor = {Anchor.Anchor_TR, Anchor.Anchor_TR, pool.get_v3f(0.0, y, 1.0)}
				 end
			end
		end

		self.m_spell_buttons = buttons
	end

	local create_progress_bar = function(parent, background_texture, foreground_texture)
		local background = parent:spawn(Window())
		background.receive_input = false
		assets.setTexture(background, background_texture)

		progress_bar = background:spawn(Window())
		progress_bar.receive_input = false
		progress_bar.color = pool.get_colorf(1.0, 1.0, 1.0, 1.0)
		progress_bar.size = background.size
		progress_bar.anchor = {Anchor.Anchor_L, Anchor.Anchor_L, pool.get_v3f(0.0, 0.0, 1.0)}

		local foreground = background:spawn(Window())
		foreground.receive_input = false
		assets.setTexture(foreground, foreground_texture)
		foreground.anchor = {Anchor.Anchor_TL, Anchor.Anchor_TL, pool.get_v3f(0.0, 0.0, 2.0)}

		return background, progress_bar, foreground
	end

	-- create Smithy button
	do
		local num_spawn_buttons = #self.m_spawn_buttons
		if num_spawn_buttons > 0 then
			local last_spawn_button = self.m_spawn_buttons[num_spawn_buttons]
			local offset_y = 0.0
			local offset_x = 10.0
			if last_spawn_button then
				local localtion = last_spawn_button.anchor["offset"]
				offset_x = offset_x + localtion.x + (last_spawn_button.size.x * 0.5)
			end

			local level_label = nil

			-- the button
			do
				self.m_smithy_button = window:spawn(Button())

				-- Dim the button.
				self.m_smithy_button.on_enter = function(owner)
					owner.color = colorf( 0.7, 0.7, 0.7, 1.0 )
				end

				-- Undim the button.
				self.m_smithy_button.on_exit = function(owner)
					owner.color = colorf( 1.0, 1.0, 1.0, 1.0 )
				end

				-- Undim the button.
				self.m_smithy_button.on_enable = function(owner)
					owner.color = colorf( 1.0, 1.0, 1.0, 1.0 )
					self.m_smithy_button.lua.upgrade_bar.enabled = true
				end

				self.m_smithy_button.on_disable = function(owner)
					self.m_smithy_button.lua.upgrade_bar.enabled = false
				end

				self.m_smithy_button.on_release = function()
					game.lua:upgradeIronCap()
					self:updateSmithyButton()

					--assets.playSound(assets.sounds.smithy_upgrade)
					util.playSoundAtCamera("?://"..assets.sounds.smithy_upgrade.path, {priority=1.0})
				end

				assets.setTexture(self.m_smithy_button, assets.textures.smithy_button)

				if engine.platform.device == "ipad" then
					self.m_smithy_button.anchor = {Anchor.Anchor_BL, Anchor.Anchor_BL, pool.get_v3f(0.0, last_spawn_button.size.y + offset_y, 1.0)}
				else
					self.m_smithy_button.anchor = {Anchor.Anchor_BL, Anchor.Anchor_BL, pool.get_v3f(offset_x, offset_y, 1.0)}
				end
			end

			do  -- Level label
				level_label = self.m_smithy_button:spawn(Text())

				level_label.style = "Helvetica-Bold"
				level_label.options = {height = 14.0, max_width = 25}
				level_label.anchor = {Anchor.Anchor_L, Anchor.Anchor_C, pool.get_v3f(-2.0+1, -1.0+3, 1.0)}

				local iron_level_string = lstring("level_abbreviation")
				iron_level_string:arg(1, game.lua:getIronLevel())
				level_label.text = iron_level_string
			end

			do  -- Spawn iron Label
				self.m_iron_label = self.m_smithy_button:spawn(Text())
				self.m_iron_label.style = "Helvetica-Bold"
				self.m_iron_label.options = {height = 18.0}
				self.m_iron_label.anchor = {Anchor.Anchor_BR, Anchor.Anchor_BR, pool.get_v3f(-8.0-1, 4.0+4, 1.0)}
			end

			do -- Upgrade Pop Out
				local text = lstring("upgrade")
				local format = {"Helvetica-Bold", 18.0, pool.get_colorf(0.2, 0.2, 1.0, 1.0), 62.0, TextAlign.Right, -13.0, 2.0, 0.4, false, false}
				local upgrade_bar = util.createButtonWithDefaultOrLocale(self.m_smithy_button, text, format, assets.textures.smithy_upgrade_bar_locale, assets.textures.smithy_upgrade_bar)
				upgrade_bar.anchor = {Anchor.Anchor_L, Anchor.Anchor_L, pool.get_v3f(11.0, 10.0, -2.0)}

				-- forward input to the button
				upgrade_bar.on_enter = function(owner)
					self.m_smithy_button:on_enter(self.m_smithy_button)
				end

				upgrade_bar.on_exit = function(owner)
					self.m_smithy_button:on_exit(self.m_smithy_button)
				end

				upgrade_bar.on_release = function(owner)
					self.m_smithy_button:on_release(self.m_smithy_button)
				end

				upgrade_bar.hidden = true
				upgrade_bar.scale = pool.get_v3f(0.0, 1.0, 1.0)
				self.m_smithy_button.lua = {}
				self.m_smithy_button.lua.upgrade_bar = upgrade_bar
			end

			-- Iron progress bar
			do
				local smithy_progress_bar_bg, smithy_progress_bar, smithy_progress_bar_foreground = create_progress_bar(self.m_smithy_button, assets.textures.smithy_progress_bar_back, assets.textures.smithy_progress_bar)
				smithy_progress_bar_bg.anchor = {Anchor.Anchor_BL, Anchor.Anchor_BR, pool.get_v3f(-9.0, 0.0, -5.0)}
				smithy_progress_bar_bg.scale = pool.get_v3f(0.0, 1.0, 1.0)

				smithy_progress_bar.color = pool.get_colorf(0.0, 0.0, 1.0, 1.0)
				smithy_progress_bar.anchor = {Anchor.Anchor_L, Anchor.Anchor_L, pool.get_v3f(2.0, 0.5, 1.0)}
				smithy_progress_bar.scale = pool.get_v3f(0.0, 1.0, 1.0)
				self.m_smithy_button.lua.progress_bar = smithy_progress_bar
				self.m_smithy_button.lua.progress_bar_bg = smithy_progress_bar_bg
				self.m_smithy_button.lua.progress_bar_foreground = smithy_progress_bar_foreground
			end

			self.m_smithy_button.lua.level_label = level_label
			self.m_smithy_button.enabled = false
			self:updateSmithyButton(true)
		end
	end

	-- Ash Health Bar
	do
		local ash_life_icon = window:spawn(Window())
		ash_life_icon.receive_input = false
		assets.setTexture(ash_life_icon, assets.textures.ash_life_icon)
		ash_life_icon.anchor = {Anchor.Anchor_TL, Anchor.Anchor_TL, pool.get_v3f(2.0, 0.0, 1.0)}

		local background = nil
		background, self.m_ash_health_bar = create_progress_bar(ash_life_icon, assets.textures.ash_lifebar_back, assets.textures.ash_lifebar)

		background.anchor = {Anchor.Anchor_TL, Anchor.Anchor_TR, pool.get_v3f(-5.0, -2.0, -4.0)}

		self.m_ash_health_bar.color = pool.get_colorf(0.0, 1.0, 0.0, 1.0)
		self.m_ash_health_bar.size = self.m_ash_health_bar.size * pool.get_v2f(0.88, 0.48)
		self.m_ash_health_bar.anchor = {Anchor.Anchor_L, Anchor.Anchor_L, pool.get_v3f(2.0, 0.5, 1.0)}
		--self.m_ash_health_bar.texture = "?://texture/" .. assets.textures.health_bar.path
		--self.m_ash_health_bar.texture.filter = TextureResource.Filter_Nearest
	end

	--[[ Gold Label
	do
		local coin = window:spawn(Window())
		coin.receive_input = false
		assets.setTexture(coin, assets.textures.gold_icon)
		coin.anchor = {Anchor.Anchor_TL, Anchor.Anchor_TL, pool.get_v3f(50.0, -29.0, 2.0)}

		self.m_gold_label = create_label(coin, nil, TextFormat("font://Helvetica-Bold", 17.0, pool.get_colorf(1.0, 0.89, 0.28, 1.0)))
		self.m_gold_label.anchor = {Anchor.Anchor_L, Anchor.Anchor_R, pool.get_v3f(0.0, -1.0, 1.0)}
	end
	]]

	-- Enemy Progress Bar
	do
		local foreground
		self.m_enemy_progress_bar_container, self.m_enemy_progress_bar, foreground = create_progress_bar(window, assets.textures.enemy_progress_bar_back, assets.textures.enemy_progress_bar_front)

		self.m_enemy_progress_bar_container.anchor = {Anchor.Anchor_T, Anchor.Anchor_T, pool.get_v3f(0.0, 0.0, 1.0)}
		assets.setTexture(self.m_enemy_progress_bar, assets.textures.enemy_progress_bar_mid)
		self.m_enemy_progress_bar.anchor = {Anchor.Anchor_R, Anchor.Anchor_R, pool.get_v3f(-8.0, 0.0, 1.0)}
		self.m_enemy_progress_bar.scale = pool.get_v3f(0.0, 1.0, 1.0)

		-- Put danger icons on the progress bar
		do
			self.m_danger_icons = {}
			local wave_data = util.getCurrentWaveData()

			if wave_data then
				local num_dudes = util.getNumUnitsInCurrentWave()
				for i,spawn_data in pairs(wave_data.unit_stat) do
					if spawn_data.tag == "push" then
						local danger_icon = foreground:spawn(Window())
						danger_icon.receive_input = false
						assets.setTexture(danger_icon, assets.textures.enemy_progress_danger)

						local percent = i / num_dudes
						local location = v3f(foreground.size) * pool.get_v3f(1.0 - percent, 0.0, 0.0)
						location.z = percent * 10.0
						danger_icon.anchor = {Anchor.Anchor_C, Anchor.Anchor_L, location}
						table.insert(self.m_danger_icons, danger_icon)
					end
				end
			end
		end
	end

	-- Necro Warning Icon
	do
		self.m_necro_warning_icon = window:spawn(Window())
		self.m_necro_warning_icon.receive_input = false
		assets.setTexture(self.m_necro_warning_icon, assets.textures.protect_the_book_icon)
		self.m_necro_warning_icon.anchor = {Anchor.Anchor_L, Anchor.Anchor_L, pool.get_v3f(0.0, 0.0, 2.0)}
		self.m_necro_warning_icon.hidden = true

		local should_hide_necro_icon = function (inner)
			local game = MainGame.fast_cast(self.owner:getGame())
			if game.lua.m_necro == nil then
				inner.hidden = true
				return
			end

			local necro_object = game.lua.m_necro:get()
			if necro_object == nil then
				inner.hidden = true
				return
			end

			if necro_object:isDead() == true then
				inner.hidden = true
				return
			end

			local last_damage_time = necro_object.lua.m_last_damage_time
			if last_damage_time <= 0.0 then
				inner.hidden = true
				return
			end

			local last_damage_dt = game:getGameTime() - necro_object.lua.m_last_damage_time
			if last_damage_dt > 2.0 then
				inner.hidden = true
				return
			end

			local icon_world_location = inner:getWorldLocation() + game.lua.m_scene_camera:getWorldLocation()
			local necro_world_location = necro_object:getWorldLocation()
			local location_dt = icon_world_location.x - necro_world_location.x - 60.0
			if location_dt <= 0.0 then
				inner.hidden = true
				return
			end

			inner.hidden = false
		end

		local sequence = SequenceAction()
		sequence:addAction(AlphaAction(0.5, InterpAction.Lerp, 0.5))
		sequence:addAction(LuaAction(0.0, should_hide_necro_icon))
		sequence:addAction(AlphaAction(0.5, InterpAction.Lerp, 1.0))
		sequence:addAction(LuaAction(0.0, should_hide_necro_icon))

		local loop = LoopAction()
		loop:addAction(sequence)
		self.m_necro_warning_icon:addAction(loop)
	end

	self:createBloodOverlay()

	Resources.get():removeResourceGroup(self.m_asset_group)
end

function LuaMainGameWindow:updateUI(window)
	--[[
	self:updateHealthBars()
	]]
end

function LuaMainGameWindow:updateWaveProgressBar()
	local game = MainGame.fast_cast(self.owner:getGame())
	--self.m_gold_label.text = "" .. UserData.instance():getGold()

	do-- Update the wave progress bar
		--local wave_percent = game.lua.m_spawned_so_far / game.lua.m_units_this_wave
		local wave_percent = game.lua.m_enemies_killed / math.max(1, game.lua.m_units_this_wave)
		wave_percent = math.min(1.0, wave_percent)

		local split = SplitAction()
		local uv_action = UVAction(0.0, InterpAction.Lerp)
		uv_action:setScale(v2f(self.m_enemy_progress_bar.scale.x, 1.0), v2f(wave_percent, 1.0))
		uv_action:setLocation(v2f(1.0 - self.m_enemy_progress_bar.scale.x, 0.0), v2f(1.0 - wave_percent, 0.0))
		split:addAction(uv_action)
		split:addAction(ScaleAction(0.0, InterpAction.Lerp, pool.get_v3f(wave_percent, 1.0, 1.0)))
		self.m_enemy_progress_bar:addAction(split)
	end
end

function LuaMainGameWindow:updateSmithyButton(instant)
	local game = self.owner:getGame()
	local iron_level = game.lua:getIronLevel()
	local max_iron_level = game.lua:getMaxIronLevel()
	local level_label = self.m_smithy_button.lua.level_label

	do	-- Update the progress bar size
		local action_time = 0.5
		if instant ~= nil and instant == true then
			action_time = 0.0
		end

		local new_x_scale = iron_level / max_iron_level
		local previous_x_scale = self.m_smithy_button.lua.progress_bar_bg.scale.x

		local create_uv_action = function()
			local split = SplitAction()
			local uv_action = UVAction(action_time, InterpAction.Lerp)
			uv_action:setScale(v2f(previous_x_scale, 1.0), v2f(new_x_scale, 1.0))
			uv_action:setLocation(v2f(1.0 - previous_x_scale, 0.0), v2f(1.0 - new_x_scale, 0.0))
			split:addAction(uv_action)

			return split
		end

		do -- Foreground Actions
			local split_action = create_uv_action()
			self.m_smithy_button.lua.progress_bar_foreground:addAction(split_action)
		end

		do -- Background Actions
			local split_action = create_uv_action()
			split_action:addAction(ScaleAction(action_time, InterpAction.Lerp, pool.get_v3f(new_x_scale, 1.0, 1.0)))
			self.m_smithy_button.lua.progress_bar_bg:addAction(split_action)
		end

		self.m_smithy_button.lua.progress_bar.size = self.m_smithy_button.lua.progress_bar_bg.size * pool.get_v2f(0.79 + ((iron_level - 1) * 0.05), 0.48)
	end

	if game.lua:isAtMaxIronCap() == true then
		level_label.text = lstring("MAX")
		level_label.scale = pool.get_v3f(0.65, 0.65, 1.0)
	else
		local iron_level_string = lstring("level_abbreviation")
		iron_level_string:arg(1, game.lua:getIronLevel())
		level_label.text = iron_level_string
		level_label.scale = pool.get_v3f(1.0, 1.0, 1.0)
	end
end

--[[
function LuaMainGameWindow:tick(window)
	--self:updateUI(window)
end
]]

function LuaMainGameWindow:setSelectingSpellTarget(value)
	self.m_selecting_spell_target = value

	if self.m_select_spell_target_info_window then
		self.m_select_spell_target_info_window:destroy()
		self.m_select_spell_target_info_window = nil
	end

	if value then
		self.m_select_spell_target_info_window = self.owner:spawn(Text())
		local format =  {"Georgia Bold", 20.0, pool.get_colorf(0.0, 1.0, 0.0, 1.0)}
		self.m_select_spell_target_info_window.style = format[1]
		self.m_select_spell_target_info_window.options = {height = format[2]}
		self.m_select_spell_target_info_window.color = format[3]
		self.m_select_spell_target_info_window.anchor = {Anchor.Anchor_T, Anchor.Anchor_T, pool.get_v3f(0.0, -25.0+5, 1.0)}
		self.m_select_spell_target_info_window.text = lstring("select_target") --"Select A Target!"
	end
end

function LuaMainGameWindow:isSelectingSpellTarget()
	return self.m_selecting_spell_target
end

function LuaMainGameWindow:getSpellRange()
	return chunks.load.getStat("spells", self.m_current_spell, "range") or 0
end

function LuaMainGameWindow:spellTargetSelected(location)
	if self:isSelectingSpellTarget() == false then
		return
	end

	local game = self.owner:getGame()
	self:setSelectingSpellTarget(false)

	self:castCurrentSpell(location)
end


function LuaMainGameWindow:castCurrentSpell(location)
	local game = self.owner:getGame()
	game.lua:castCurrentSpell(self.m_current_spell, location)

	return true
end

function LuaMainGameWindow:updateHealthBars()
	local main_game = MainGame.fast_cast(self.owner:getGame())

	local update_health_bar = function(health_bar, health_object)
		if health_bar == nil then
			return
		end

		local health = 0.0
		local health_remaining = 0.0
		if health_object then
			local part_health = PartHealthIndex(health_object):get()
			if part_health then
				health = part_health.health
				health_remaining = part_health:getHealthRemaining()
			end
		end

		if health == 0.0 then
			health_bar.scale = pool.v3f(0.0, 1.0, 1.0)
		else
			-- Show the health bar like a fuel gauge, so when you are empty you still have a bit of health left
			local health_diff = (1.0 - globals.ash_health_bar_show_percentage) * health
			local percent = math.max(0.0, (health_remaining - health_diff)) / (health - health_diff)

			health_bar.scale = pool.v3f(percent, 1.0, 1.0)

			-- Health color goes from green to yellow to red
			local red = 1.0 - percent
			local green = percent
			local multiplier = 1 / math.max(red, green)
			health_bar.color = pool.get_colorf(red * multiplier, green * multiplier, 0.0, 1.0)
		end
	end

	update_health_bar(self.m_ash_health_bar, main_game.lua.m_ash)
end

function LuaMainGameWindow:showRatingDialog()
	local title = lstring("rating_title")
    local message = lstring("rating_body")
	local no_text = lstring("rating_no_text")
	local yes_text = lstring("rating_yes_text")
	local ui = PlatformUIService.get()
	
	local callback = function()
		self.m_platform_ui_handle = nil
	end
	
	self.m_platform_ui_handle = ui:showRatingsDialog(title, message, yes_text, no_text, callback)
end

function LuaMainGameWindow:gameOver(team, award)
	Resources.get():addResourceGroup(self.m_asset_group)

	self.m_pause_button.enabled = false 	-- Disable the pause button since the game over screen will be coming up soon
	self.owner:getGame():getScene():addAction(LuaAction(3.0, function() self:delayedGameOver(team, award) end ))

	if self.m_danger_popup then
		self.m_danger_popup:destroy()
		self.m_danger_popup = nil
	end

	-- Victory/Defeat Assets
	local window
	if team == GameObject.TEAM_ALLY then
		local format = {"Gothic_HD", 65.0, pool.get_colorf(0.8, 0.7, 0.0, 1.0), 150.0, TextAlign.Center, 0.0, 13.0, 0.0, false}
		window = util.createWindowWithDefaultOrLocale(self.owner, lstring("victory"), format, assets.textures.victory_label_locale, assets.textures.victory_label)
		
		-- check for rating after wave 14, 29 and 44 (wave number has already been incremented at this point)
		local current_wave = UserData.instance():getCurrentWave()
		if current_wave == 15 or current_wave == 30 or current_wave == 45 then
			self:showRatingDialog()
		end
		
	else
		local format = {"Gothic_HD", 53.0, pool.get_colorf(1, 0, 0, 1), 200.0, TextAlign.Center, 0.0, 13.0, 0.4, false}
		window = util.createWindowWithDefaultOrLocale(self.owner, lstring("defeat"), format, assets.textures.defeat_label_locale, assets.textures.defeat_label)
		self:showBloodOverlay(0.0, 0.7)
	end
	window.anchor = { local_anchor=Anchor.Anchor_B, owner_anchor=Anchor.Anchor_C, offset=pool.get_v3f(0.0, 45.0, 475.0) }
	window.color = pool.get_colorf(1, 1, 1, 0);
	window:addAction(AlphaAction(0.25, InterpAction.Smooth, 1))

	local game_lua = self.owner:getGame().lua
	if game_lua.m_necro:get() == nil or game_lua.m_necro:get():isDead() == true then
		window.anchor = { local_anchor=Anchor.Anchor_B, owner_anchor=Anchor.Anchor_C, offset=pool.get_v3f(0.0, 55.0, 475.0) }

		local protect_window = util.createWindowWithTexture(window, assets.textures.tutorial_window)
		protect_window.anchor = { Anchor.Anchor_T, Anchor.Anchor_B, pool.get_v3f(0.0, 10.0, 1.0) }
		protect_window.scale = pool.get_v3f(1.0, 0.9, 1.0)

		local protect_book = util.createWindowWithTexture(protect_window, assets.textures.protect_the_book_icon)
		protect_book.anchor = { Anchor.Anchor_L, Anchor.Anchor_L, pool.get_v3f(12.0, 0.0, 1.0) }
		protect_book.scale = pool.get_v3f(0.8, 0.8, 1.0)

		local label = protect_window:spawn(Text())
		label.anchor = { Anchor.Anchor_C, Anchor.Anchor_C, pool.get_v3f(23, 5, 1) }
		label.style = "Helvetica-Bold"
		label.options = {height = 20.2, max_width = 230, align = TextAlign.Center}
		label.text = lstring("deadites_got_book") --"Deadites got the book!"
		label.scale = pool.get_v3f(1.0, 1.111111, 1.0)
	end

	Resources.get():removeResourceGroup(self.m_asset_group)
end

function LuaMainGameWindow:delayedGameOver(team, award)
	local window = self.owner
	local game = window:getGame()

	local do_continue_action = function()
		 --TODO: FIXME Audio.get().sound:unloadSounds("sound", ".*")
		 if team == GameObject.TEAM_ENEMY then
			if UserData.instance():getCurrentWave() == 1 then
				engine:spawnMainGame()
			else
				engine:spawnUpgradeUnitMenu()
			end
		 elseif team == GameObject.TEAM_ALLY then
			if UserData.instance():getCurrentWave() == globals.endless_wave then
				engine:spawnCreditsMenu()
				util.setEndlessModeUnlocked()
			else
				engine:spawnUnlockedMenu()
			end
		 end
	end

	Resources.get():addResourceGroup(self.m_asset_group)

	local game_over_window = window:spawn(Window())
	game_over_window.hidden = true
	game_over_window.size = window.size
	game_over_window.anchor = { local_anchor=Anchor.Anchor_C, owner_anchor=Anchor.Anchor_C, offset=pool.get_v3f(0.0, 0.0, 450.0) }

	--Stop Ash
	do
		local ash_index = self.owner:getGame().lua.m_ash
		local ash = ash_index:get()
		if ash then
			local render = PartRenderIndex(ash_index)
			render:get().render_actor.scale = pool.get_v3f(1,1,1)
			local orders = PartOrdersIndex(ash_index)
			orders:get():clearOrders()
			local logic = PartLogicIndex(ash_index)
			logic:get():requestProcess(1/30)
		end

	end

	-- Background
	do
		local bg = game_over_window:spawn(Window())
		bg.color = colorf(0, 0, 0, 0)
		bg.size = game_over_window.size
		bg:addAction(AlphaAction(0.125, InterpAction.Lerp, 0.5))
	end

	-- Continue button
	do
		local text = lstring("continue")
		local format = {"Georgia Bold", 25.0, pool.get_colorf(1.0, 1.0, 1.0, 1.0), 95.0, TextAlign.Center, 0.0, 4.0, 0.0, false}
		local continue_button = util.createButtonWithDefaultOrLocale(game_over_window, text, format, assets.textures.continue_button_locale, assets.textures.continue_button)
		continue_button.anchor = { local_anchor=Anchor.Anchor_C, owner_anchor=Anchor.Anchor_C, offset=pool.get_v3f(0.0, -75.0-36.0, 2.0) }

		-- Resume the game
		continue_button.on_release = function(owner)
			 do_continue_action()
			 util.playSoundAtCamera(assets.sounds.button_click_wav)
		end
	end

	--[[ Restart Button
	if team == GameObject.TEAM_ENEMY then
		local button = util.createButtonWithTexture(game_over_window, assets.textures.restart_button)
		button.anchor = {Anchor.Anchor_C, Anchor.Anchor_C, pool.get_v3f(0, -60, 2)}

		button.on_release = function(owner)
			engine:spawnMainGame()
			end
	end
	]]


	-- Gold Boxes (Plural!)
	if (award and award~=0) or (game.lua.m_gold_collected and game.lua.m_gold_collected~=0) then
		local gold_window = game_over_window:spawn(Window())
        gold_window.hidden = true
		gold_window.anchor = {Anchor.Anchor_C, Anchor.Anchor_C, v3f(0, -20, 1)}

		local y_pad = 8
		local y_offset = 0
        local text_height = 26
		local game_lua = self.owner:getGame().lua
		if game_lua.m_necro:get() == nil or game_lua.m_necro:get():isDead() == true then
			y_offset = -50
		end

        controls = {}
        local gold_line
		
		for i=(award and 0 or 1),(award and 2 or 1) do
			do
				-- the gold coin
				do
					--the gold amount label
					
					do
						local total_label = gold_window:spawn(Text())
						local format = {"Georgia Bold", text_height, pool.get_colorf(1.0, 0.89, 0.28, 1.0), 140}
						total_label.style = format[1]
						total_label.options = {height = format[2], max_width = format[4]}
						total_label.color = format[3]
						total_label.anchor = {Anchor.Anchor_TL, Anchor.Anchor_TL, pool.get_v3f(0, y_offset, 1)}
						total_label.text =  (i==0 and lstring("awarded")) or (i==1 and lstring("collected")) or (i==2 and lstring("total"))
						total_label.color = pool.get_colorf(1, 1, 1, 0)
						total_label:addAction(AlphaAction(0.25, InterpAction.Smooth, 1))

						-- Specific issue with German display
		                if engine.m_language == "de" then
		                    total_label.match_localized_scale = 0.78
		                end

						local gold_label = gold_window:spawn(Text())
						local format = {"Helvetica-Bold", text_height, pool.get_colorf(1.0, 0.89, 0.28, 1.0)}
						gold_label.style = format[1]
						gold_label.options = {height = format[2]}
						gold_label.color = format[3]
						
						local gold_stat_string = lstring("stat")
						
						if(i==0) then
							gold_stat_string:arg(1, tostring(award))
							
						elseif(i==1) then
							gold_stat_string:arg(1, tostring(game.lua.m_gold_collected))
							
						elseif(i==2) then
							gold_stat_string:arg(1, tostring(award + game.lua.m_gold_collected))
							
						end
						gold_label.text = gold_stat_string
						
						gold_label.color = pool.get_colorf(1, 1, 1, 0)

						gold_label:addAction(AlphaAction(0.25, InterpAction.Smooth, 1))

						local gold_icon = util.createWindowWithTexture(gold_label, assets.textures.upgrade_gold_coin)
						gold_icon.anchor = {Anchor.Anchor_L, Anchor.Anchor_R, pool.get_v3f(0, 0-5, 0)}

						gold_label.anchor = {Anchor.Anchor_TR, Anchor.Anchor_TR, pool.get_v3f(-gold_icon.size.x, y_offset, 1)}

                        table.insert(controls, total_label)
                        table.insert(controls, gold_label)
						
						y_offset = y_offset - (format[2] + y_pad + 8)  --8 = the space between the lines
					end
				end
			end
		end

        util.unifyTextHeight(controls)

		gold_window.size = pool.get_v2f(250, math.abs(y_offset) - 8)

		if award then
			local coin_box = util.createWindowWithTexture(gold_window, assets.textures.coin_box)
			coin_box.anchor = {Anchor.Anchor_B, Anchor.Anchor_B, pool.get_v3f(0, -2, 0)}
			coin_box.color = pool.get_colorf(1, 1, 1, 0)
			coin_box:addAction(AlphaAction(0.25, InterpAction.Smooth, 1))
		end
	end

	Resources.get():removeResourceGroup(self.m_asset_group)

	do -- check ratings dialog data
		if team == GameObject.TEAM_ALLY then
			local current_wave = UserData.instance():getCurrentWave()
			if util.table.simpleFind(globals.ratings_dialog_levels, current_wave - 1) ~= nil then
				events.on_ratings_dialog()
			end
		end
	end
end

function LuaMainGameWindow:ironTotalChanged(previous_iron, current_iron)
	--print("Iron total changed to:", previous_iron, current_iron)
	local game = self.owner:getGame()

	Resources.get():addResourceGroup(self.m_asset_group)

	for i,button in ipairs(self.m_spawn_buttons) do
		local button_lua = button.lua
		local cost = button_lua.m_iron_cost
		if previous_iron < cost and current_iron >= cost then   --now we can afford it
			--button_lua:lightenBorder()
			button_lua:makeReady(button, true)
		elseif previous_iron >= cost and current_iron < cost then   --spent some iron, now we can't
			--button_lua:darkenBorder()
			button_lua:makeDisabled()
		end
	end

	local iron_level = game.lua:getIronLevel()
	local max_iron_level = game.lua:getMaxIronLevel()
	local iron_level_percent = iron_level / max_iron_level

	if self.m_smithy_button then
		local upgrade_bar = self.m_smithy_button.lua.upgrade_bar
		local current_percent = upgrade_bar.scale.x / 1.0
		local move_time = 0.25
		local percent_left = 1 - current_percent

		if game.lua:canUpgradeIronCap() == true then
			self.m_smithy_button.enabled = true
			assets.setTexture(self.m_smithy_button, assets.textures.smithy_button_glow)


			upgrade_bar:clearActions()

			local sequence = SequenceAction()
			sequence:addAction(LuaAction(0.0, function (inner)
				inner.hidden = false
			end))

			local split = SplitAction()
			sequence:addAction(ScaleAction(move_time * percent_left, InterpAction.Lerp, pool.get_v3f(1.0, 1.0, 1.0)))

			local uv = UVAction(move_time * percent_left, InterpAction.Lerp)
			uv:setLocation(v2f(percent_left, 0.0), v2f(0, 0.0))
			split:addAction(uv)

			sequence:addAction(split)

			upgrade_bar:addAction(sequence)
		else
			self.m_smithy_button.enabled = false
			assets.setTexture(self.m_smithy_button, assets.textures.smithy_button)

			upgrade_bar:clearActions()

			local sequence = SequenceAction()
			local split = SplitAction()

			split:addAction(ScaleAction(move_time * current_percent, InterpAction.Lerp, pool.get_v3f(0.0, 1.0, 1.0)))

			local uv = UVAction(move_time * current_percent, InterpAction.Lerp)
			uv:setLocation(v2f(percent_left, 0.0), v2f(1, 0.0))
			split:addAction(uv)

			sequence:addAction(split)
			sequence:addAction(LuaAction(0.0, function (inner)
				inner.hidden = true
			end))
			upgrade_bar:addAction(sequence)
		end

		local percent_complete = game.lua.m_iron / game.lua:getIronCap()
		self.m_smithy_button.lua.progress_bar.scale = pool.get_v3f(percent_complete, 1.0, 1.0)

		local stat_string = lstring("stat")
		stat_string:arg(1, math.floor(game.lua.m_iron))
		self.m_iron_label.text = stat_string
	end

	Resources.get():removeResourceGroup(self.m_asset_group)
end

function LuaMainGameWindow:createBloodOverlay()
	local window = self.owner

	--The Center of the Window, up 250 on Z Axis
	local base = window:spawn(ActorBase())
	base.location = pool.get_v3f(0, 0, 250)

	--Bottom
	local blood_width = constants.main_menu_camera.window.viewport_size.x
	local blood = base:spawn(SpriteActor())
	assets.setTexture(blood, assets.textures.blood_overlay)
	blood.size = pool.get_v2f(blood_width, blood.size.y)
	blood.location = pool.get_v3f(0, window.size.y*-0.5 + blood.size.y*0.5, 0)

	--Top
	blood = base:spawn(SpriteActor())
	assets.setTexture(blood, assets.textures.blood_overlay)
	blood.size = pool.get_v2f(blood_width, blood.size.y)
	blood.rotation = pool.get_v3f(0, 0, 180)
	blood.location = pool.get_v3f(0, window.size.y*0.5 + blood.size.y*-0.5, 0)

	--Right
	blood = base:spawn(SpriteActor())
	assets.setTexture(blood, assets.textures.blood_overlay)
	blood.size = pool.get_v2f(blood_width, blood.size.y)
	blood.rotation = pool.get_v3f(0, 0, 90)
	blood.location = pool.get_v3f(window.size.x*0.5 + blood.size.y*-0.5, 0, 0)
	blood.scale = pool.get_v3f(window.size.y / window.size.x, 1, 1)

	--Left
	blood = base:spawn(SpriteActor())
	assets.setTexture(blood, assets.textures.blood_overlay)
	blood.size = pool.get_v2f(blood_width, blood.size.y)
	blood.rotation = pool.get_v3f(0, 0, 270)
	blood.location = pool.get_v3f(window.size.x*-0.5 + blood.size.y*0.5, 0, 0)
	blood.scale = pool.get_v3f(window.size.y / window.size.x, 1, 1)

	base.hidden_recursive = true

	self.m_blood_overlay_base = base
end

function LuaMainGameWindow:showBloodOverlay(fade_time, alpha)
	local base = self.m_blood_overlay_base
	local new_fade_time = fade_time or 0.75
	alpha = alpha or 1.0


	base:clearActions()
	base.hidden_recursive = false
	base.color = pool.get_colorf(1, 1, 1, alpha)

	if new_fade_time > 0.0 then
		base:addAction(AlphaAction(new_fade_time, InterpAction.Lerp, 0.0))
		base:addAction(LuaAction(0, function() base.hidden_recursive = true end))
	end
end


function LuaMainGameWindow:showCurrentWaveLabel()
	Resources.get():addResourceGroup(self.m_asset_group)

	local wave_box = self.owner:spawn(Window())
	assets.setTexture(wave_box, assets.textures.upgrade_box)
	wave_box.size = wave_box.size * pool.get_v2f(0.5, 0.75)
	wave_box.anchor = {Anchor.Anchor_C, Anchor.Anchor_C, pool.get_v3f(0.0, 0.0, 125.0)}

	local format = {"Georgia Bold", 30.0+3, colorf(0.93, 0.72, 0.26, 1.0)}

	local label = wave_box:spawn(Text())
	label.style = format[1]
	label.options = {height = format[2]}
	label.color = format[3]

	local fx = {drop_shadow = true, shadow_color = colorf( 0, 0, 0, 1.0 ), shadow_offset = v3f( 2, -2, 0), shadow_blur_amount = 1}
	label.fx = fx
	label.anchor = {Anchor.Anchor_C, Anchor.Anchor_C, pool.get_v3f(0.0, -2.0+7, 1.0)}

	local wave = util.getCurrentWave()
	local wave_string = lstring("wave")
	wave_string:arg(1, wave)
	label.text = wave_string

	wave_box:addAction(SleepAction(1.0))
	wave_box:addAction(AlphaAction(0.25, InterpAction.Lerp, 0.0))
	wave_box:addAction(DestroyAction(0.0))

	Resources.get():removeResourceGroup(self.m_asset_group)
end

function LuaMainGameWindow:showWavePushEffect(push_id)
	if not self.m_main_game.lua.m_game_over then
		util.playRandomSoundFromTable(assets.sounds.push_sounds, {priority=5.0})

		Resources.get():addResourceGroup(self.m_asset_group)

		self.m_danger_popup = self.owner:spawn(SpriteActor())
		util.setLocaleTexture(self.m_danger_popup, assets.textures.danger_popup_locale, assets.textures.danger_popup)
		self.m_danger_popup.location = pool.get_v3f(0.0, 65.0, 10.0)

		if not engine.m_default_language_used then
			local text_options = 	{
										height = 30,
									  	max_width = 470,
									  	align = TextAlign.Center,
									  	outline_color = pool.get_colorf(0,0,0,0),
									  	font_color = colorf(1,0,0,1),
									  	char_spacing = 1.0,
									  	line_spacing = 0.75
								 	}
			local vertical_offset = 28.0

			util.createText(self.m_danger_popup, "incoming_horde", "Gothic_Large", text_options, {x = 0.0, y = vertical_offset}, true)
		end

		local pulse_actor = function (parent, pulse_time, num_pulses, scale_offset)
			local sequence = SequenceAction()
			local loop = FiniteLoopAction(num_pulses)
			local half_time = pulse_time / 2

			loop:addAction(ScaleOffsetAction(half_time, InterpAction.Smooth, pool.get_v3f(scale_offset, scale_offset, 1.0)))
			loop:addAction(ScaleOffsetAction(half_time, InterpAction.Smooth, pool.get_v3f(-scale_offset, -scale_offset, 1.0)))

			sequence:addAction(loop)
			parent:addAction(sequence)
		end

		pulse_actor(self.m_danger_popup, 1.0, 3, 0.4)
		self.m_danger_popup:addAction(LuaAction(0, function() self.m_danger_popup = nil end ))
		self.m_danger_popup:addAction(DestroyAction(0.0))

		local danger_icon = self.m_danger_icons[push_id]
		if danger_icon ~= nil then
			pulse_actor(danger_icon, 1.0, 6, 0.1)
		end

		Resources.get():removeResourceGroup(self.m_asset_group)
	end
end

function LuaMainGameWindow:showPauseWindow()
	if self.m_pause_window == nil then
	    self:showHidePauseWindow()
	end
end

function LuaMainGameWindow:showHidePauseWindow()
    if self.m_pause_window == nil then
        self.m_pause_window = self.owner:spawn(chunks.pause_window.spawn())
	    self.m_pause_window.anchor = { Anchor.Anchor_C, Anchor.Anchor_C, pool.get_v3f(0.0, 0.0, 475.0) }
    else
        self.owner:getGame().paused = false
		self.m_pause_window:destroy()
        self.m_pause_window = nil
    end
end

function LuaMainGameWindow:destroy()
	Resources.get():unload(self.m_asset_group)
	
	self.m_platform_ui_handle = nil
end

