-- module --
module(...,package.seeall)

-- includes --
local engine    = engine.instance
local chunks    = engine.chunks
local constants = engine.platform.constants
local assets    = chunks.assets
local pool      = chunks.pool
local util      = chunks.util
local analytics = chunks.analytics
local globals   = chunks.globals
require "common.BigNum"

local DEBUG_RELEASE_ENABLED = false

LuaOptionsMenuWindow = chunks.class.new
{
    m_current_wave_label = nil,
    m_god_mode_label = nil,
    m_castle_units_label = nil,
    m_video_handle = nil
}

-- The spawner.
spawn = chunks.class.spawner( Window, LuaOptionsMenuWindow )

function LuaOptionsMenuWindow:created(window)
    window.hidden = true
    window.size = constants.main_menu_camera.window.world_size

    -- Background
    local background = window:spawn(Window())
    util.setLocaleTexture(background, assets.textures.options_screen_background_locale, assets.textures.options_screen_background)
    background.anchor = { Anchor.Anchor_L, Anchor.Anchor_L, pool.get_v3f(0,0,0) }
    background.size = window.size

    -- Back Button
    do
        local text = lstring("options_back")
        local format =  {"Georgia Bold", 16.0, pool.get_colorf(1.0, 1.0, 1.0, 1.0), 40.0, TextAlign.Center, 0.0, 2.4, 0.4}
        local back_button = util.createButtonWithDefaultOrLocale(window, text, format, assets.textures.blank_button_locale, assets.textures.back_button)
        back_button.anchor = {Anchor.Anchor_TR, Anchor.Anchor_TR, pool.get_v3f(0.0, 0.0, 1.0)}

        -- Quit the game
        back_button.on_release = function(owner)
            assets.playSound(assets.sounds.button_click_wav)
            engine:goToPreviousGame()
        end
    end

    -- Debug Options Button
    do
        if BFS_DEBUG or DEBUG_RELEASE_ENABLED then
            local format =  {"Debug Font", 20.0, pool.get_colorf(0.0, 0.0, 1.0, 1.0)}
            local debug_options = util.createButtonWithText(window, "**Debug Options**", format)
            debug_options.anchor = {Anchor.Anchor_L, Anchor.Anchor_L, pool.get_v3f(0.0, -50.0, 1.0)}

            -- Quit the game
            debug_options.on_release = function(owner)
                engine:spawnDebugOptionsMenu()
            end
        end
    end

    -- Display "OPTIONS" text
    do
        if not engine.m_default_language_used then
            local text_options =    {
                                        height = 65,
                                        max_width = 250,
                                        align = TextAlign.Left,
                                        outline_color = pool.get_colorf(0,0,0,1),
                                        font_color = colorf(0.63, 0.27, 1.0, 1.0),
                                        char_spacing = 2.7,
                                    }
            local options_text = util.createText(window, "options_options", "Gothic_HD", text_options, nil, true)
            options_text.anchor = {Anchor.Anchor_TL, Anchor.Anchor_TL, v3f( 10, -2, 1)}
        end
    end

    local toggle_button_controls = {}
    do -- Music/Sound options
        local create_onoff_check_button = function(locale_texture, texture, text, format, get_onoff_function, set_onoff_function)

            local button, text = util.createButtonWithDefaultOrLocale(window, text, format, locale_texture, texture)
            
            local check_mark = util.createWindowWithTexture(button, assets.textures.check_mark)
            check_mark.anchor = { Anchor.Anchor_R, Anchor.Anchor_R, pool.get_v3f(0.0, 0.0, 1.0) }
            check_mark.hidden = not get_onoff_function()

            button.on_release = function(inner)
                local button_on = not get_onoff_function()
                set_onoff_function(button_on)

                check_mark.hidden = not button_on
                assets.playSound(assets.sounds.button_click_wav)
            end

            return button, text
        end

        local y_location = 58.0
        -- Music on/off button
        do
            local format = {"Gothic_Large", 25.0, pool.get_colorf(1.0, 1.0, 1.0, 1.0), 65.0, TextAlign.Center, -12.0, 3, 0.1 }

            local music_button, music_text = create_onoff_check_button(assets.textures.music_button_locale, assets.textures.music_button, lstring("options_music"), format,
                function ()
                    return GameOptions.instance().music_on
                end,
                function(onoff)
                    GameOptions.instance().music_on = onoff
                    Audio.get():getPlaylistGroup().volume = GameOptions.instance().music_on and 1.0 or 0.0
                end
            )
            music_button.anchor = { Anchor.Anchor_L, Anchor.Anchor_L, pool.get_v3f(0.0, y_location, 1.0 ) }

            if music_text then
                table.insert(toggle_button_controls, music_text)
            end
        end



        -- Sound on/off button
        do
            local format = {"Gothic_Large", 25.0, pool.get_colorf(1.0, 1.0, 1.0, 1.0), 65.0, TextAlign.Center, -12.0, 3, 0.1}
            local sound_button, sound_text = create_onoff_check_button(assets.textures.music_button_locale, assets.textures.sound_button, lstring("options_sound"), format,
                function ()
                    return GameOptions.instance().sound_on
                end,
                function(onoff)
                    GameOptions.instance().sound_on = onoff
                    Audio.get():getSoundPoolGroup().volume = GameOptions.instance().sound_on and 1.0 or 0.0
                end
            )
            sound_button.anchor = { Anchor.Anchor_L, Anchor.Anchor_L, pool.get_v3f(120.0, y_location, 1.0 ) }

            if sound_text then
                table.insert(toggle_button_controls, sound_text)
            end
        end
    end

    if #toggle_button_controls > 1 then
       util.unifyTextHeight(toggle_button_controls) 
    end

    -- Legal Button
    do
        local text = lstring("legal")
        
        local format = {"Gothic_Large", 18.0, pool.get_colorf(1.0, 1.0, 1.0, 1.0), 75.0, TextAlign.Center, 0.0, 5.0}
        legal_button, legal_label = util.createButtonWithTextAndTexture(window, text, format, assets.textures.legal_button)
        legal_button.anchor = { Anchor.Anchor_L, Anchor.Anchor_L, pool.get_v3f(0.0, 0.0, 1.0) }
        legal_button.on_release = function(owner)
            local ui = PlatformUIService.get()
            if ui then
                local callback = function()
                    print("web view disappeared")
                    self.m_platform_handle = nil
                end
                self.m_platform_handle = ui:showWebView("https://iris-prod.beast-dev.com/query?q=license&game=aodd", "", UIOrientation.Any, callback)
            end
        end
    end

    local text_button_controls = {}
    do -- Intro Sequence
        local format = {"Gothic_Large", 25.0, pool.get_colorf(1.0, 1.0, 1.0, 1.0), 250.0, TextAlign.Right, -5.0, 15.0, 0, true}
        local intro_sequence_button, intro_text  = util.createButtonWithTextureOrLocaleText(window, lstring("options_intro"), format, assets.textures.intro_button)
        intro_sequence_button.anchor = { Anchor.Anchor_BR, Anchor.Anchor_R, pool.get_v3f(-10.0, -18.0, 1.0 ) }
        intro_sequence_button.on_release = function(inner)
            util.playIntroVideo( self.introStopped );
        end

        if intro_text then
            table.insert(text_button_controls, intro_text)
        end
    end

    do -- Credits Button

        local format =  {"Gothic_Large", 25.0, pool.get_colorf(1.0, 1.0, 1.0, 1.0), 250.0, TextAlign.Right, -5.0, 15.0, 0, true}
        local credits_button, credits_text = util.createButtonWithTextureOrLocaleText(window, lstring("options_credits"), format, assets.textures.credits_button)

        credits_button.anchor = { Anchor.Anchor_BR, Anchor.Anchor_R, pool.get_v3f(-10.0, -64.0, 1.0 ) }
        credits_button.on_release = function(inner)
            assets.playSound(assets.sounds.button_click_wav)
            engine:spawnCreditsMenu()
        end

        if credits_text then
            table.insert(text_button_controls, credits_text)
        end

    end

    do -- Reset All Data Button
        local format =  {"Gothic_Large", 25.0, pool.get_colorf(1.0, 1.0, 1.0, 1.0), 250.0, TextAlign.Right, -5.0, 15.0, 0, true}
        local reset_data_button, reset_text = util.createButtonWithTextureOrLocaleText(window, lstring("options_reset_game"), format, assets.textures.reset_game_button)

        reset_data_button.anchor = { Anchor.Anchor_BR, Anchor.Anchor_R, pool.get_v3f(-10.0, -112.0, 1.0 ) }
        reset_data_button.on_release = function(inner)

            local popup = window:spawn( chunks.options_reset_popup.spawn() )
            popup.lua:show(function()
                analytics.sendGameDataReset()
                local gold_from_iap = UserData.instance():getGoldFromIAP()
                engine:resetData()
                UserData.instance():setGold(gold_from_iap)
                UserData.instance():setGoldFromIAP(gold_from_iap)

            end)  -- end callback
        end -- end on_release

        if reset_text then
            table.insert(text_button_controls, reset_text)
        end
    end

    do -- Restore Purchases Button
        local format =  {"Gothic_Large", 25.0, pool.get_colorf(1.0, 1.0, 1.0, 1.0), 275.0, TextAlign.Right, -5.0, 15.0, 0, true}
        local restore_button, restore_text = util.createButtonWithTextureOrLocaleText(window, lstring("options_restore_purchases"), format, assets.textures.restore_purchases_button)

        restore_button.anchor = { Anchor.Anchor_BR, Anchor.Anchor_R, pool.get_v3f(-10.0, -160.0, 1.0 ) }
        restore_button.on_release = function()
            GameService.get():restorePurchases()
        end

        if restore_text then
            table.insert(text_button_controls, restore_text)
        end        
    end

    if #text_button_controls > 1 then
        util.unifyTextHeight(text_button_controls)
    end

    --Endless Mode Score
    if util.isEndlessModeUnlocked() == true then

        local score_window = util.createWindowWithTexture(window, assets.textures.coin_box, false)
        score_window.anchor = {Anchor.Anchor_BL, Anchor.Anchor_L, pool.get_v3f(8, -168, 1)}
        score_window.size = score_window.size - pool.get_v2f(96, 6)

        local score_label = score_window:spawn(Text())
        local format = {"Helvetica-Bold", 18.0+2}

        score_label.style = format[1]
        --TODO shadow blur doesnt seem to be working
        local fx = {drop_shadow = true, shadow_color = colorf( 0, 0, 0, 1.0 ), shadow_offset = v3f( 2, -2, 0), shadow_blur_amount = 100000}
        score_label.fx = fx
        score_label.options = {height = format[2], max_width = 160}
        score_label.anchor = { Anchor.Anchor_C, Anchor.Anchor_C, v3f(0,3,1) }

--        score_label.location = score_label.location + pool.get_v3f(2, 3, 2)

        local score = BigNum.new(UserData.instance():getEndlessWaveScore())
        local score_string = string.format("%09s", tostring(score))
        score_string = string.gsub(score_string, "(%d%d%d)", function(s) return s.." " end)
        local stat_string = lstring("endless_score")
        stat_string:arg(1, score_string)
        score_label.text = stat_string

        local format =  {"Gothic_Large", 36.0, pool.get_colorf(1.0, 1.0, 1.0, 1.0), 300.0, TextAlign.Left, 9.0, 23.0}
        --local endless = util.createButtonWithTextureOrLocaleText(score_window, lstring("endless"), format, assets.textures.endless_options)
        --endless.anchor = {Anchor.Anchor_BL, Anchor.Anchor_TL, pool.get_v3f(-8, 0, 1)}

    end
end

function LuaOptionsMenuWindow:introStopped()
    GameService.get():restoreAudioSession()
end

function LuaOptionsMenuWindow:destroy(window)
    if self.m_video_handle then
        self.m_video_handle:cancel()
    end
end