-- module --
module(...,package.seeall)

-- includes --
local engine    = engine.instance
local chunks    = engine.chunks
local constants = engine.platform.constants
local assets    = chunks.assets
local pool      = chunks.pool
local util      = chunks.util
local load      = chunks.load
local globals   = chunks.globals

LuaSelectUnitMenuGame = chunks.class.new
{
	m_asset_group = nil,
	m_active_window = nil,
}

-- The spawner.
spawn = chunks.class.spawner( SingleTouchWindowGame, LuaSelectUnitMenuGame )

function LuaSelectUnitMenuGame:created(game)
	print("LuaSelectUnitMenuGame:created")

	local game_window = game:getWindow()

	-- Window camera.
	do
		local camera = game:getWindow():spawn( OrthographicCamera() )

		camera.viewport_offset = constants.main_menu_camera.window.viewport_offset
		camera.viewport_size   = constants.main_menu_camera.window.viewport_size
		camera.screen_size     = constants.main_menu_camera.window.screen_size
		camera.world_size      = constants.main_menu_camera.window.world_size
		camera.clipping_planes = constants.main_menu_camera.window.clipping_planes
		camera.clear_color     = constants.clear_color
		camera.clear_color_on  = constants.clear_color_on
		camera.clear_depth_on  = constants.clear_depth_on
		camera.scene           = game_window

		game_window.camera = camera
	end

	self.m_asset_group = ResourceGroup("texture")
	self:preloadAssets()

	self:setup()

	chunks.music.checkOrPlaySong( assets.music.loop )
end

-- Input handling for the android back button.
function LuaSelectUnitMenuGame:onInputEvents( game, events )
	if engine.platform.device == "android" then
		for event in events:all() do
			local e = AndroidEvent.fast_cast( event )
			if e and e.type == AndroidEvent.AndroidEvent_BackButton then
				--engine:goToPreviousGame()
				self.m_active_window.lua.m_back_button.on_release()
			end
		end
	end
end



function LuaSelectUnitMenuGame:setup()
	local game = self.owner
	local game_window = game:getWindow()
	game_window.size = constants.main_menu_camera.window.world_size

	Resources.get():addResourceGroup(self.m_asset_group)

	-- Background
	do
		local background = game_window:spawn(Window())
		assets.setTexture(background, assets.textures.castle_front_background)
		background.anchor = { Anchor.Anchor_L, Anchor.Anchor_L, pool.get_v3f(0.0, 0.0, 0.0) }
	end

	-- Current Level
	do
		local label = game_window:spawn(Text())
		label.style = "Georgia Bold"
		local fx = {drop_shadow = true, shadow_color = colorf( 0, 0, 0, 1.0 ), shadow_offset = v3f( 2, -2, 0), shadow_blur_amount = 100000}
		label.fx = fx
		label.options = {height = 24.0, max_width = 80, font_color = pool.get_colorf(0.93, 0.72, 0.26, 1.0), align = TextAlign.Center}
		label.anchor = {Anchor.Anchor_T, Anchor.Anchor_T, pool.get_v3f(130.0, 0.0, 200.0)}
		local wave = util.getCurrentWave()
		local wave_string = lstring("wave")
		wave_string:arg(1, wave)
		label.text = wave_string
	end

	self:spawnBackgroundWaveUnits()

	Resources.get():removeResourceGroup(self.m_asset_group)

	self:spawnUnitSelectWindow()
end

function LuaSelectUnitMenuGame:spawnBackgroundWaveUnits()
	local game = self.owner
	local game_window = game:getWindow()

	local wave_data = util.getCurrentWaveData()
	if wave_data == nil then
		return
	end

	local units_to_spawn = {}
	do -- Set up the units to spawn array
		local unique_units = {}
		local num_unique_units = 0
		for k,v in pairs(wave_data.unit_stat) do
			if v.level then
				unique_units[v.unit] = math.max((unique_units[v.unit] or 1), v.level)
				num_unique_units = num_unique_units + 1
			end
		end


		for name,level in pairs(unique_units) do
			local unique = load.getStat("units", name, "unique", level) or false
			local unlock_wave = chunks.load.getStat("units", name, "unlock_wave", level) or 1
			table.insert(units_to_spawn, { name = name, level = level, unique = unique, unlock_wave = unlock_wave})
		end

		-- Put unique units at the front
		table.sort(units_to_spawn, function(left, right)
			local unlock_wave_check = function()
				return left.unlock_wave < right.unlock_wave
			end

			if left.unique == true then
				if right.unique == false then
					return true
				end
			else
				if right.unique == true then
					return false
				end
			end

			return unlock_wave_check()
		end)

		local num_units_to_insert = globals.min_units_in_wave_preview - #units_to_spawn
		if num_units_to_insert > 0 then
			for i=1, num_units_to_insert do
				local unit = units_to_spawn[i]
				if unit ~= nil and unit.unique == false then
					table.insert(units_to_spawn, unit)
				end
			end
		end
	end

	local animation_options = {
						loop_count = 0,
						ease_out_current_animation_time = 0.3,
						ease_in_time = 0.1,
					}

	local spawn_count = 0
	local num_units = #units_to_spawn
	local x_locs = { 130, 180, 230 }
	for i,unit_data in pairs(units_to_spawn) do
		if i > globals.max_units_in_wave_preview then
			break
		end

		local index = i - 1
		local unit = util.loadObjectRenderDataFromStats(game_window, "units", unit_data.name, unit_data.level)
		if unit ~= nil then
			unit:playAnimation("win", animation_options)
			local x = x_locs[(index % #x_locs) + 1]
			local y = (index * 20) + -125
			local z = (num_units - index) * 6
			unit.location = pool.get_v3f(x, y, z)
			unit.scale = unit.scale * pool.get_v3f(1.0, 1.0, 0.001)
		end
	end
end


function LuaSelectUnitMenuGame:spawnUnitSelectWindow()
	local game = self.owner
	local game_window = game:getWindow()

	Resources.get():addResourceGroup(self.m_asset_group)

	local unit_window = game_window:spawn(chunks.select_unit_window.spawn())
	unit_window.anchor = {Anchor.Anchor_TL, Anchor.Anchor_TL, pool.get_v3f(0.0, 0.0, 400.0)}
	self.m_active_window = unit_window

	Resources.get():removeResourceGroup(self.m_asset_group)
end

function LuaSelectUnitMenuGame:spawnSpellSelectWindow()
	local game = self.owner
	local game_window = game:getWindow()

	Resources.get():addResourceGroup(self.m_asset_group)

	local spell_window = game_window:spawn(chunks.select_spell_window.spawn())
	spell_window.anchor = {Anchor.Anchor_TL, Anchor.Anchor_TL, pool.get_v3f(0.0, 0.0, 400.0)}
	self.m_active_window = spell_window

	Resources.get():removeResourceGroup(self.m_asset_group)
end

function LuaSelectUnitMenuGame:destroy()
	self:unloadAssets()
end


function LuaSelectUnitMenuGame:preloadAssets()
	--TODO: FIXME local sound_manager = Audio.get().sound

	-- sounds
	--sound_manager:precache(URI(assets.getSound(assets.sounds.button_click)))
end

function LuaSelectUnitMenuGame:unloadAssets()
	Resources.get():unload(self.m_asset_group)
end
