-- module --
module(...,package.seeall)

-- includes --
local engine    = engine.instance
local chunks    = engine.chunks
local constants = engine.platform.constants
local assets    = chunks.assets
local pool      = chunks.pool
local util      = chunks.util
local load      = chunks.load
local globals   = chunks.globals

local stat_string = lstring("stat")

LuaSelectUnitWindow = chunks.class.new
{
	m_selected_unit_buttons = nil,
	m_possible_unit_buttons = nil,
	m_continue_button = nil,
	m_container_window = nil,
	m_transition_time = globals.selection_screen_slide_time,
}

-- The spawner.
spawn = chunks.class.spawner( Window, LuaSelectUnitWindow )

function LuaSelectUnitWindow:created(window)
	window.hidden = true
	window.size = window:getGame():getWindow().size

	self.m_selected_unit_buttons = {}
	self.m_possible_unit_buttons = {}
	local possible_units = util.getValidUnits()

	do --  Container Window
		local container_window = window:spawn(Window())
		container_window.receive_input = false
		container_window.hidden = true
		container_window.size = window:getGame():getWindow().size
		container_window.anchor = { Anchor.Anchor_L, Anchor.Anchor_L, pool.get_v3f(0,0,0) }
		self.m_container_window = container_window
	end

	do -- Setup Selection Box
		local selection_box = self.m_container_window:spawn(Window())
		selection_box.receive_input = false
		assets.setTexture(selection_box, assets.textures.selection_box)
		selection_box.size = selection_box.size + globals.selection_menu_selection_box_size_offset
		selection_box.anchor = {Anchor.Anchor_TL, Anchor.Anchor_TL, pool.get_v3f(0.0, 0.0, 1.0)}

		do -- setup buttons
			local num_columns = 3
			local column = 1
			local row = 1
			local icon_spacing = v2f(100, 57.5)
			local offset = v2f(10, -23)
			for i,v in ipairs(possible_units) do
				local icon = self:createUnitIcon(selection_box, v)
				local x = offset.x + ((column - 1) * icon_spacing.x)
				local y = offset.y + ((row - 1) * -icon_spacing.y)
				icon.anchor = {Anchor.Anchor_TL, Anchor.Anchor_TL, pool.get_v3f(x, y, 1.0)}
				v.button = icon

				table.insert(self.m_possible_unit_buttons, icon)

				column = column + 1
				if column > num_columns then
					column = 1
					row = row + 1
				end
			end
		end

		do -- Select troops label
			local label = selection_box:spawn(Text())

			label.style = "Georgia Bold"
			local fx = {drop_shadow = true, shadow_color = colorf( 0, 0, 0, 1.0 ), shadow_offset = v3f( 1, -1, 0), shadow_blur_amount = 100000}
			label.fx = fx
			label.options = {height = 16.0}
			label.anchor = {Anchor.Anchor_TL, Anchor.Anchor_TL, pool.get_v3f(12.0, -4.0+2, 1.0)}
			label.text = lstring("select_troops") --"Select Troops"
		end
	end

	for i=1,globals.max_selected_units do
		local button = util.createButtonWithTexture(self.m_container_window, assets.textures.icon_blank)
		button.anchor = {Anchor.Anchor_B, Anchor.Anchor_BL, pool.get_v3f(button.size.x*(i-0.5), 0.0, 0.5)}
		button.on_release = function() assets.playSound(assets.sounds.error) end
	end

	-- the Back button
	do
		local text = lstring("options_back")
		local format =  {"Georgia Bold", 16.0, pool.get_colorf(1.0, 1.0, 1.0, 1.0), 48.0, TextAlign.Center, 0.0, 2.4, 0.4}
		local back_button = util.createButtonWithDefaultOrLocale(window, text, format, assets.textures.blank_button_locale, assets.textures.back_button)
		back_button.anchor = {Anchor.Anchor_TR, Anchor.Anchor_TR, pool.get_v3f(0.0, 0.0, 1.0)}

		-- Quit the game
		back_button.on_release = function(owner)
			assets.playSound(assets.sounds.button_click_wav)
			self:saveSelectedUnits()
			self:transitionOut(function()
				engine:goToPreviousGame()
			end)
		end
		self.m_back_button = back_button
	end

	-- the Continue button
	do
	    local format = {"Georgia Bold", 25.0, pool.get_colorf(1.0, 1.0, 1.0, 1.0), 95.0, TextAlign.Center, 0.0, 4.0, 0.0, false}
		self.m_continue_button = util.createButtonWithDefaultOrLocale(window, lstring("continue"), format, assets.textures.continue_button_locale, assets.textures.continue_button)

		self.m_continue_button.anchor = {Anchor.Anchor_BR, Anchor.Anchor_BR, chunks.globals.continue_button_offset}

		-- Start the game
		self.m_continue_button.on_release = function(owner)
			if #self.m_selected_unit_buttons > 0 then
				assets.playSound(assets.sounds.button_click_wav)
				self:saveSelectedUnits()

				self:transitionOut(function()
					local game = self.owner:getGame()
					self.owner:destroy()
					game.lua:spawnSpellSelectWindow()
				end)
			else
				assets.playSound(assets.sounds.error)
				local info = window:spawn(chunks.infopopup.spawn())
				if info then
					info.lua:setTitle("select_unit")
					info.lua:setMessage({ "select_one_unit" })
					info.lua:show()
				end
			end
		end
	end

	do -- Load selected units
		local unit_list = util.loadStringList(globals.selected_units_id)

		for i,v in ipairs(unit_list) do
			self:createSelectedUnitButton(v)
		end
	end

	self:updateButtons()

	self:transitionIn()
end

function LuaSelectUnitWindow:transitionIn(callback_function)
	local sequence = SequenceAction()

	sequence:addAction(MoveAction(self.m_transition_time, InterpAction.Smooth, pool.get_v3f(0.0, 0.0, 0.0)))
	if callback_function ~= nil then
		sequence:addAction(LuaAction(0.0, function(inner)
			callback_function()
		end))
	end

	self.m_container_window:addAction(sequence)
end

function LuaSelectUnitWindow:transitionOut(callback_function)
	local sequence = SequenceAction()

	util.setReceiveInputRecursive(self.owner, false)
	sequence:addAction(MoveAction(self.m_transition_time, InterpAction.Smooth, pool.get_v3f(-self.m_container_window.size.x, 0.0, 0.0)))
	if callback_function ~= nil then
		sequence:addAction(LuaAction(0.0, function(inner)
			callback_function()
		end))
	end

	self.m_container_window:addAction(sequence)
end

function LuaSelectUnitWindow:saveSelectedUnits()
	local unit_list = {}
	for i,v in ipairs(self.m_selected_unit_buttons) do
		table.insert(unit_list, v.label)
	end

	util.saveStringList(globals.selected_units_id, unit_list)
end

function LuaSelectUnitWindow:createUnitIcon(owner, unit_type)
	local name = unit_type.name
	local level = UserData.instance():getTypeLevel(name)
	local current_wave =UserData.instance():getCurrentWave()

	--create Icon
	local available = current_wave >= unit_type.unlock_wave

	local icon_asset = assets.findTextureByPathName( chunks.load.getStat("units", name, "icon_texture") ) or assets.textures.icon_generic
	if not available then
		icon_asset = assets.textures.icon_locked
	end

	local icon = util.createButtonWithTexture(owner, icon_asset)
	assets.setTexture(icon, icon_asset)
	icon.anchor = {Anchor.Anchor_L, Anchor.Anchor_L, pool.get_v3f(0.0, 0.0, 1.0)}
	icon.label = name

	if available then
		-- create Iron Cost Label
		do
			local iron_label = icon:spawn(Text())
			local format = {"Helvetica-Bold", 14.0}
			iron_label.style = format[1]
			iron_label.options = {height = format[2]}
			iron_label.anchor = {Anchor.Anchor_C, Anchor.Anchor_C, pool.get_v3f(13,-19+4,1)}
			stat_string:arg(1, tostring( load.getStat("units", name, "iron_cost") ))
			iron_label.text = stat_string
		end

		-- create Attack Label
		do
			local attack_label = icon:spawn(Text())
			local format = {"Helvetica-Bold", 12.0}
			attack_label.style = format[1]
			attack_label.options = {height = format[2]}
			attack_label.anchor = {Anchor.Anchor_TL, Anchor.Anchor_TR, pool.get_v3f(12+1, -7.5+3, 1.0)}

			local damage
			local weapon_name = load.getStat("units", name, "weapon", level)
			local ammo_name = load.getStat("weapons", weapon_name, "ammo", level)
			if ammo_name then
				damage = load.getStat("ammo", ammo_name, "damage", level)
			else
				damage = load.getStat("weapons", weapon_name, "damage", level)
			end
			stat_string:arg(1, tostring(math.abs(damage)))
			attack_label.text = stat_string

			--create Attack Icon
			do
				local attack = nil
				if damage < 0 then
					attack = util.createWindowWithTexture(icon, assets.textures.heal_icon)
				else
					attack = util.createWindowWithTexture(icon, assets.textures.attack_icon)
				end
				attack.anchor = {Anchor.Anchor_TL, Anchor.Anchor_TR, pool.get_v3f(-3.0, -7.5, 1.0)}
			end
		end

		--create Health Icon
		do
			local health = util.createWindowWithTexture(icon, assets.textures.health_icon)
			health.anchor = {Anchor.Anchor_BL, Anchor.Anchor_BR, pool.get_v3f(-3.0, 7.5, 1.0)}
		end

		-- create Health Label
		do
			local health_label = icon:spawn(Text())
			local format = {"Helvetica-Bold", 12.0}
			health_label.style = format[1]
			health_label.options = {height = format[2]}
			health_label.anchor = {Anchor.Anchor_BL, Anchor.Anchor_BR, pool.get_v3f(13, 10.5, 1.0)}
			stat_string:arg(1, tostring( load.getStat("units", name, "health", level) ))
			health_label.text = stat_string
		end

		-- Should we show the new button?
		if current_wave == unit_type.unlock_wave then
			local font_height = 7.0
			local offset = 7.0

			-- In Chinese we want larger text to make it legible
			if ( engine.m_language == "zh-Hant" or engine.m_language == "zh-Hans" or engine.m_language == "zh" ) then
				font_height = 12.0
				offset = 7.8
			end

			local format = {"Helvetica", font_height, pool.get_colorf(0.0, 0.0, 0.0, 1.0), 25.0, TextAlign.Center, -offset, offset, 1.0, false, false}
			local new_plaque, text_label = util.createWindowWithDefaultOrLocale(icon, lstring("new_plaque"), format, assets.textures.new_plaque_locale, assets.textures.new_plaque)

			new_plaque.anchor = {Anchor.Anchor_TL, Anchor.Anchor_TL, pool.get_v3f(2.0, -2.0, 1.0)}

			-- text_label is only created when non-English language is in use
			if ( text_label ) then
				text_label.rotation = pool.get_v3f(0, 0, 45)
			end
		end

		icon.on_release = function(inner)
			if #self.m_selected_unit_buttons < globals.max_selected_units then
				assets.playSound(assets.sounds.button_click_wav)
				inner.enabled = false

				self:createSelectedUnitButton(inner.label)
			else
				assets.playSound(assets.sounds.error)
			end
		end
	else
		icon.on_release = function() assets.playSound(assets.sounds.error) end
	end

	return icon
end

function LuaSelectUnitWindow:createSelectedUnitButton(name)
	local level = UserData.instance():getTypeLevel(name)
	local icon_asset = assets.findTextureByPathName( chunks.load.getStat("units", name, "icon_texture") ) or assets.textures.icon_generic

	local button = util.createButtonWithTexture(self.m_container_window, icon_asset)
	assets.setTexture(button, icon_asset)
	button.label = name

	-- create Iron Cost Label
	do
		local iron_label = button:spawn(Text())
		local format = {"Helvetica-Bold", 14.0}
		iron_label.style = format[1]
		iron_label.options = {height = format[2]}
		iron_label.anchor = {Anchor.Anchor_C, Anchor.Anchor_C, pool.get_v3f(13, -19+4, 1)}
		stat_string:arg(1, tostring( load.getStat("units", name, "iron_cost") ) )
		iron_label.text = stat_string
	end

	button.on_release = function(inner)
		local index_to_remove = nil
		for i,v in ipairs(self.m_selected_unit_buttons) do
			if inner.label == v.label then
				index_to_remove = i
				break
			end
		end

		if index_to_remove then
			for i,v in ipairs(self.m_possible_unit_buttons) do
				if inner.label == v.label then
					v.enabled = true
					break
				end
			end

			table.remove(self.m_selected_unit_buttons, index_to_remove)
			inner:destroy()
			self:updateButtons()
			assets.playSound(assets.sounds.button_click_wav)
		else
			assets.playSound(assets.sounds.error)
		end
	end

	do --Create the "x" icon
		local x_icon = util.createWindowWithTexture(button, assets.textures.x_button, true)
		x_icon.anchor = {Anchor.Anchor_TR, Anchor.Anchor_TR, pool.get_v3f(8.0, 8.0, 1.0)}

		-- Foward input to the parent
		x_icon.on_enter = function(inner)
			inner:getOwner():on_enter()
		end

		x_icon.on_exit = function(inner)
			inner:getOwner():on_exit()
		end

		x_icon.on_release = function(inner)
			inner:getOwner():on_release()
		end
	end

	table.insert(self.m_selected_unit_buttons, button)
	self:updateButtons()
end

function LuaSelectUnitWindow:updateButtons()
	for i,button in ipairs(self.m_selected_unit_buttons) do
		button.anchor = {Anchor.Anchor_B, Anchor.Anchor_BL, pool.get_v3f(button.size.x*(i-0.5), 0.0, 1.0)}

		for j,button2 in ipairs(self.m_possible_unit_buttons) do
			if button.label == button2.label then
				button2.enabled = false
			end
		end
	end
end
