module(...,package.seeall)
-- includes --
local engine	= engine.instance
local chunks	= engine.chunks
local util		= chunks.util
local assets	= chunks.assets
local pool		= chunks.pool
local globals   = chunks.globals

-- watch for either ONE or ALL of a set of supplied functions
-- to return 'true', then call the triggered_function
function watchForSet(owner,funcs, watch_type, triggered_function)
	-- print("watching for a set")
	if watch_type == "all" then
		local bool = true
		for i,v in ipairs(funcs) do
			if v() == false then
				-- shortcut try again later
				owner:addAction( LuaAction(0.2, function()
					watchForSet(owner, funcs, watch_type, triggered_function)
				end) )
				return
			end
		end
		triggered_function()
	elseif watch_type == "one" then
		for i,v in ipairs(funcs) do
			if v() == true then
				print("ended on function set "..i)
				triggered_function()
				return
			end
		end
		-- no true, try again
		owner:addAction( LuaAction(0.2, function()
			watchForSet(owner, funcs, watch_type, triggered_function)
		end) )
	end
end

function watchForIronCountOver(owner, game, trigger_count, triggered_function)
	local actor = owner:spawn(ActorBase())
	actor.label = "tutorial.watchForIronCountOver: "..trigger_count

	local loop = LoopAction()
	loop:addAction( LuaAction(globals.tutorial_interval, function()
		local iron_count = game.lua.m_iron
		if iron_count > trigger_count then
			if triggered_function then
				triggered_function()
			end
			actor:destroy()
		end
	end) )

	actor:addAction(loop)
end

function watchForAshGreaterThanLocation(owner, trigger_location, triggered_function)
	print("watching for Ash location greater than "..trigger_location)

	local actor = owner:spawn(ActorBase())
	actor.label = "watchForAshGreaterThanLocation: "..trigger_location

	local loop = LoopAction()
	loop:addAction( LuaAction(globals.tutorial_interval, function()
		print("ash location check loop")
		local ash_index = owner:getGame().lua.m_ash
		local ash = ash_index:get()

		if ash then
			if ash.location.x > trigger_location then
				if triggered_function then
					triggered_function()
				end
				actor:destroy()
			end
		end
	end) )

	actor:addAction(loop)
	print("loop action added successfully")
end

function watchForUnitDamage(owner, object_index, trigger_ratio, triggered_function)
	local actor = owner:spawn(ActorBase())
	actor.label = "watchForUnitDamage: "..trigger_ratio

	local loop = LoopAction()
	loop:addAction( LuaAction(globals.tutorial_interval, function()
		local object = object_index:get()
		if object == nil or object:isDead() then
			if triggered_function then
				triggered_function()
			end
			actor:destroy()
			return --shortcut return so you don't access health of a null actor
		end

		local health_part = PartHealthIndex(object_index):get()
		local test_ratio = health_part:getHealthRemaining() / health_part.health

		if test_ratio < trigger_ratio then
			if triggered_function then
				triggered_function()
			end
			actor:destroy()
		end
	end) )

	actor:addAction(loop)
end

function watchForAshDistanceToEnemies(owner, trigger_distance, triggered_function)
	local actor = owner:spawn(ActorBase())
	actor.label = "watchForAshDistanceToEnemies: "..trigger_distance

	local loop = LoopAction()

	loop:addAction( LuaAction(globals.tutorial_interval, function()
		local ash_index = owner:getGame().lua.m_ash
		local ash = ash_index:get()

		if ash == nil then
			print("Ash is dead, cannot watch for distance")
			return
		end

		local ash_location = ash.location.x
		local nearest_target = ash:getMainGame():getClosestObjectToAttack(GameObject.TEAM_ENEMY, ash_location)
		if nearest_target ~= GameObjectIndex() then
			local target_location = nearest_target:get().location.x
			if math.abs(target_location - ash_location) < trigger_distance then
				if triggered_function then
					triggered_function()
				end
				actor:destroy()
			end
		end
	end) )

	actor:addAction(loop)
end

function watchForAshDamage(owner, trigger_ratio, triggered_function)
	local actor = owner:spawn(ActorBase())
	actor.label = "watchForAshDamage: "..trigger_ratio

	local loop = LoopAction()
	loop:addAction( LuaAction(globals.tutorial_interval, function()
		local ash_index = owner:getGame().lua.m_ash
		local ash = ash_index:get()

		if ash then
			local health_part = PartHealthIndex(ash_index):get()
			local test_ratio = health_part:getHealthRemaining() / health_part.health

			if test_ratio < trigger_ratio then
				if triggered_function then
					triggered_function()
				end
				actor:destroy()
			end
		else
			print("NO ASH!! DX")
		end
	end) )

	actor:addAction(loop)
end

function createWindow(parent, location, offset)
	if offset == nil then
		if location == "bottom" then
			offset = v3f(0, 48, 1)
		elseif location == "bottom-right" then
			offset = v3f(-16, 20, 1)
		elseif location == "top" then
			offset = v3f(0, -24, 1)
		else
			offset = v3f(0, 0, 1)
		end
	end

	if is_bottom == nil then
		is_bottom = false
	end

	local tutorial_window = util.createWindowWithTexture(parent, assets.textures.tutorial_window, false)
	if location == "bottom" then
		tutorial_window.location = pool.get_v3f(0, parent.size.y*-0.5 + tutorial_window.size.y*0.5, 0) + offset
	elseif location == "bottom-right" then
		tutorial_window.location = pool.get_v3f(parent.size.x*0.5 - tutorial_window.size.x*0.5, parent.size.y*-0.5 + tutorial_window.size.y*0.5, 0) + offset
	elseif location == "top" then
		tutorial_window.location = pool.get_v3f(0, parent.size.y*0.5 + tutorial_window.size.y*-0.5, 0) + offset
	else
		tutorial_window.location = pool.get_v3f(0, 0, 0) + offset
	end

	return tutorial_window
end

function createBouncingArrow(parent, direction, magnitude, speed, pause, delay)
	delay = delay or 0
	pause = pause or 0
	rest = rest or 0
	magnitude = magnitude or 1
	speed = speed or 0.4
	local arrow = parent:spawn(SpriteActor())
	arrow.texture = "?://"..assets.textures.tutorial_pointer.path
	arrow.size = assets.textures.tutorial_pointer.size

	local loop = LoopAction()
	local sequence = SequenceAction()

	if direction == "left" then
		sequence:addAction(MoveOffsetAction(speed, InterpAction.Fall, pool.get_v3f(32*magnitude, 0, 0)))
		sequence:addAction(MoveOffsetAction(speed, InterpAction.Ramp, pool.get_v3f(-32*magnitude, 0, 0)))
	elseif direction == "down" then
		arrow.rotation = pool.get_v3f(0, 0, 90)
		sequence:addAction(MoveOffsetAction(speed, InterpAction.Fall, pool.get_v3f(0, 32*magnitude, 0)))
		sequence:addAction(MoveOffsetAction(speed, InterpAction.Ramp, pool.get_v3f(0, -32*magnitude, 0)))
	elseif direction == "up" then
		arrow.rotation = pool.get_v3f(0, 0, -90)
		sequence:addAction(MoveOffsetAction(speed, InterpAction.Fall, pool.get_v3f(0, -32*magnitude, 0)))
		sequence:addAction(MoveOffsetAction(speed, InterpAction.Ramp, pool.get_v3f(0, 32*magnitude, 0)))
	else -- direction = "right"
		arrow.scale = pool.get_v3f(-1, 1, 1)
		sequence:addAction(MoveOffsetAction(speed, InterpAction.Fall, pool.get_v3f(-32*magnitude, 0, 0)))
		sequence:addAction(MoveOffsetAction(speed, InterpAction.Ramp, pool.get_v3f(32*magnitude, 0, 0)))
	end

	if pause > 0 then
		sequence:addAction(SleepAction(pause))
	end

	loop:addAction(sequence)

	-- add a delay before the arrow's loop is added
	if delay > 0 then
		arrow:addAction( LuaAction(delay, function()
			arrow:addAction(loop)
		end) )
	else
		arrow:addAction(loop)
	end

	return arrow
end

function addAlphaPulse(object, duration, off_alpha)
	duration = duration or 1
	local half = duration / 2
	off_alpha = off_alpha or 0.3
	local loop = LoopAction()
	local sequence = SequenceAction()
	sequence:addAction(AlphaAction(half, InterpAction.Smooth, 1.0))
	sequence:addAction(AlphaAction(half, InterpAction.Smooth, off_alpha))
	loop:addAction(sequence)
	object:addAction(loop)
end

function getAnchorLocation(actor, location)
	local c = actor.location
	local w = actor.size.x/2
	local h = actor.size.y/2
	local s = actor.scale

	if location == "C" then
		return c
	elseif location == "L" then
		return v3f(c.x-w*math.abs(s.x), c.y, c.z)
	elseif location == "B" then
		return v3f(c.x, c.y-h*math.abs(s.y), c.z)
	elseif location == "R" then
		return v3f(c.x+w*math.abs(s.x), c.y, c.z)
	elseif location == "T" then
		return v3f(c.x, c.y+h*math.abs(s.y), c.z)
	elseif location == "TL" then
		return v3f(c.x-w*math.abs(s.x), c.y+h*math.abs(s.y), c.z)
	elseif location == "TR" then
		return v3f(c.x+w*math.abs(s.x), c.y+h*math.abs(s.y), c.z)
	elseif location == "BL" then
		return v3f(c.x-w*math.abs(s.x), c.y-h*math.abs(s.y), c.z)
	elseif location == "BR" then
		return v3f(c.x+w*math.abs(s.x), c.y-h*math.abs(s.y), c.z)
	end
end

function getAnchorLocationLocal(actor, location)
	local w = actor.size.x/2
	local h = actor.size.y/2
	local s = actor.scale

	if location == "C" then
		return v3f(0, 0, 0)
	elseif location == "L" then
		return v3f(-w*math.abs(s.x), 0, 0)
	elseif location == "B" then
		return v3f( 0, -h*math.abs(s.y), 0)
	elseif location == "R" then
		return v3f( w*math.abs(s.x), 0, 0)
	elseif location == "T" then
		return v3f (0,  h*math.abs(s.y), 0)
	elseif location == "TL" then
		return v3f(-w*math.abs(s.x),  h*math.abs(s.y), 0)
	elseif location == "TR" then
		return v3f( w*math.abs(s.x),  h*math.abs(s.y), 0)
	elseif location == "BL" then
		return v3f(-w*math.abs(s.x), -h*math.abs(s.y), 0)
	elseif location == "BR" then
		return v3f( w*math.abs(s.x), -h*math.abs(s.y), 0)
	end
end

-- Match anchor spots assuming that A and B are siblings (both have the same parent)
function matchAnchorAtoB(a, aloc, b, bloc, offset)
	-- a is source
	-- b is destination
	offset = offset or pool.get_v3f(0, 0, 0)

	local anchor_a = getAnchorLocation(a, aloc)
	local anchor_b = getAnchorLocation(b, bloc)

	a.location = pool.get_v3f(
		anchor_b.x - anchor_a.x + offset.x,
		anchor_b.y - anchor_a.y + offset.y,
		anchor_b.z + offset.z)
end

-- Match anchor spot from object A to an anchor spot on object B relative to where B's coordinates
-- This is best used is A is a child of B
function matchAnchorAtoLocalB(a, aloc, b, bloc, offset)
	-- a is source
	-- b is destination
	offset = offset or pool.get_v3f(0, 0, 0)

	local anchor_a = getAnchorLocationLocal(a, aloc)
	local anchor_b = getAnchorLocationLocal(b, bloc)

	a.location = pool.get_v3f(
		anchor_b.x - anchor_a.x + offset.x,
		anchor_b.y - anchor_a.y + offset.y,
		anchor_b.z + offset.z)
end

-- Match anchor spot from object A to an anchor spot on object B adjusting for B's location
-- This is best used is A is a nephew of B
function matchAnchorAtoParentB(a, aloc, b, bloc, offset)
	-- a is source
	-- b is destination
	offset = offset or pool.get_v3f(0, 0, 0)
	local parent = a:getOwner()
	local inverse_parent = v3f(-parent.location.x, -parent.location.y, 0)

	local anchor_a = getAnchorLocation(a, aloc)
	local anchor_b = getAnchorLocation(b, bloc)

	a.location = pool.get_v3f(
		anchor_b.x - anchor_a.x + offset.x,
		anchor_b.y - anchor_a.y + offset.y,
		anchor_b.z + offset.z) + inverse_parent
end
