-- module --
module(...,package.seeall)

local engine  = engine.instance
local chunks  = engine.chunks
local pool    = chunks.pool
local globals   = chunks.globals

local spawn_slots = {}
local unique_spawn_slots = {}
local center_spawn_slots = {}

for i=-18.00,18.75,0.02 do
	if (i<-1.0 or i>1.0) and (i<-9.1 or i>-8.9) then  --reserve space for ash and the deathcoaster
		table.insert(spawn_slots, i)
	end
end

for i=-18.75,-18.05,0.02 do
	table.insert(unique_spawn_slots, i)
end

for i=-0.9,-0.1,0.02 do
	table.insert(center_spawn_slots, i)
end


--See if a file has the specified extension
function fileHasExtension(filename, extension)

	if filename == nil or extension == nil then
		return false
	end

	local start_index, end_index  = string.find(filename, "%." .. extension)
	if start_index == nil then
		return false
	end

	if end_index ~= #filename then
		return false
	end

	return true
end

--Remove leading/trailing whitespace
function stripWhitespace(s)
	return s:gsub("^%s*(.-)%s*$", "%1")
end


function splitString(str, pat)
   local t = {}  -- NOTE: use {n = 0} in Lua-5.0
   local fpat = "(.-)" .. pat
   local last_end = 1
   local s, e, cap = str:find(fpat, 1)
   while s do
      if s ~= 1 or cap ~= "" then
	 table.insert(t,cap)
      end
      last_end = e+1
      s, e, cap = str:find(fpat, last_end)
   end
   if last_end <= #str then
      cap = str:sub(last_end)
      table.insert(t, cap)
   end
   return t
end

--Read an integer of the specified number of bytes and endianess
function readInt(data, bytes, little_endian)
	local val = 0
	for i = 1, bytes do
		local c = data:sub(i, i):byte()
		local m = little_endian and (i - 1) or (bytes - i)
		val = val + c * (0xFF * m + 1)
	end
	return val
end

--Parse json data into a lua object
function loadJson(json_str)
	--Temp info used in parsing
	local parse_info = { stack = {}, key = nil }

	--Push a parsed value
	local function pushValue(v)
		local current = (#parse_info.stack > 0) and parse_info.stack[#parse_info.stack] or parse_info.stack
		if (parse_info.key) then
			current[parse_info.key] = v
			parse_info.key = nil
		else
			table.insert(current, v)
		end
	end

	--Parsing callback
	local function onParse(v)
		if (v.type == json.ARRAY_BEGIN or v.type == json.OBJECT_BEGIN) then
			local t = {}
			pushValue(t)
			table.insert(parse_info.stack, t)
		elseif (v.type == json.ARRAY_END or v.type == json.OBJECT_END) then
			table.remove(parse_info.stack)
		elseif (v.type == json.INTEGER) then
			pushValue(v.int)
		elseif (v.type == json.FLOAT) then
			pushValue(v.flt)
		elseif (v.type == json.NULL) then
			pushValue(nil)
		elseif (v.type == json.TRUE) then
			pushValue(true)
		elseif (v.type == json.FALSE) then
			pushValue(false)
		elseif (v.type == json.STRING) then
			pushValue(v.str)
		elseif (v.type == json.KEY) then
			parse_info.key = v.str
		end
	end

	json.parse(json_str, onParse)
	return parse_info.stack[1]
end

function table.find(l, f) -- find element v of l satisfying f(k,v)
  for k,v in pairs(l) do
    if f(k,v) then
      return v
    end
  end

  return nil
end

function table.simpleFind(l, f) -- find element f == v
	return table.find(l, function(k, v)
		return v == f
	end)
end

function randomPoint(world_box)
	return v3f(math.random(world_box.left, world_box.right),
			   math.random(world_box.bottom, world_box.top),
			   1)
end

function spawnTestSpriteActor(game, size, color)
	local sprite = game:getScene():spawn(SpriteActor())
	sprite.texture = "builtin://white"
	sprite.size = size

	if color then
		sprite.color = color
	else
		sprite.color = colorf(math.random(0.0, 1.0), math.random(0.0, 1.0), math.random(0.0, 1.0), 1.0)
	end

	return sprite
end

function createButtonWithText(window, text, input_format)
	local button = window:spawn(Button())
	local label = button:spawn(Text())
	local format = input_format or {"Debug Font", 24.0, pool.get_colorf(0.0, 0.0, 1.0, 1.0)}
	label.style = format[1]
	label.options = {height = format[2]}
	label.color = format[3]
	local stat_string = lstring("stat")
	stat_string:arg(1, text)
	label.text = stat_string or ""
	label.location = pool.get_v3f(0.0, 4, 0.1)
	button.size = label.size + pool.get_v2f(6, 0)
	button.color = pool.get_colorf(1.0, 1.0, 1.0, 1.0)

	-- Dim the button.
	button.on_enter = function(owner)
		owner.color = colorf( 0.7, 0.7, 0.7, 1.0 )
	end

	-- Undim the button.
	button.on_exit = function(owner)
		owner.color = colorf( 1.0, 1.0, 1.0, 1.0 )
	end

	return button, label
end

function createButtonWithTexture(parent, texture)
	local button = parent:spawn(Button())
	--print("--- Texture name is: " .. texture)
	chunks.assets.setTexture(button, texture)

	-- Dim the button.
	button.on_enter = function(owner)
		owner.color = colorf( 0.7, 0.7, 0.7, 1.0 )
	end

	-- Undim the button.
	button.on_exit = function(owner)
		owner.color = colorf( 1.0, 1.0, 1.0, 1.0 )
	end

	button.on_enable = function(owner)
		button.color = pool.get_colorf(1.0, 1.0, 1.0, 1.0)
	end

	button.on_disable = function(owner)
		button.color = colorf( 0.7, 0.7, 0.7, 1.0 )
	end

	return button
end

function createWindowWithTexture(parent, texture, receives_input)
	local window = parent:spawn(Window())
	chunks.assets.setTexture(window, texture)
	window.receive_input = receives_input or false
	return window
end

function createButtonWithTextAndTexture(parent, text, input_format, texture)
	local button = parent:spawn(Button())

	local label = button:spawn(Text())
	local format = input_format or {"Debug Font", 24.0, colorf(0.0, 0.0, 1.0, 1.0), 0.0, TextAlign.Center, 0.0, 0.0}
	local alignment = format[5] or TextAlign.Center

	label.style = format[1]

    local options = {height = format[2], max_width = format[4] or 0.0, align = alignment, char_spacing = format[8] or 0}
    if format.wrapping then
        options.wrapping = format.wrapping
    end

    if format.max_lines then
        options.max_lines = format.max_lines
    end

    if format.localization_scaling_limit then
        options.localization_scaling_limit = format.localization_scaling_limit
    end

	label.options = options
 	label.color = format[3]
 	local original_color = format[3]
	scale_button = format[9] or false

	local text_offset_x = format[6] or 0.0
	local text_offset_y = format[7] or 0.0

 	-- The alignment of the text in 'format' determines the anchoring in this case as well
 	if ( alignment == TextAlign.Left ) then
 		label.anchor = { Anchor.Anchor_L, Anchor.Anchor_L, pool.get_v3f(text_offset_x, text_offset_y, 0.1) }
 	elseif ( alignment == TextAlign.Right ) then
 		label.anchor = { Anchor.Anchor_R, Anchor.Anchor_R, pool.get_v3f(text_offset_x, text_offset_y, 0.1) }
 	else
 		label.anchor = { Anchor.Anchor_C, Anchor.Anchor_C, pool.get_v3f(text_offset_x, text_offset_y, 0.1) }
 	end

	label.text = text or ""

	if ( nil == format[10] ) or ( true == format[10] ) then
		local fx = {drop_shadow = true, shadow_color = colorf( 0, 0, 0, 1.0 ), shadow_offset = v3f( 2, -2, 0), shadow_blur_amount = 100000}
		label.fx = fx;
	end

	if ( texture ) then
		chunks.assets.setTexture(button, texture)
		button.color = colorf(1.0, 1.0, 1.0, 1.0)
	else
		button.color = colorf(1.0, 1.0, 1.0, 0.0)
		label.force_color = true
	end


	if scale_button then
		button.size = label.size + pool.get_v2f(6, 0)
	end

	-- Dim the button.
	button.on_enter = function(owner)
		if ( texture ) then
			owner.color = colorf( 0.7, 0.7, 0.7, 1.0 )
		else
			local current_color = label.color
			label.color = colorf( current_color.r * 0.7, current_color.g * 0.7, current_color.b * 0.7, current_color.a)
			label.force_color = true
		end
	end

	-- Undim the button.
	button.on_exit = function(owner)
		if ( texture ) then
			owner.color = colorf( 1.0, 1.0, 1.0, 1.0 )
		else
			label.color = original_color
			label.force_color = true
		end
	end

	return button, label
end

function createButtonWithDefaultOrLocale(parent, locale_text, locale_input_format, locale_texture, default_texture)

	if engine.m_default_language_used then
		return createButtonWithTexture(parent, default_texture)
	else
		return createButtonWithTextAndTexture(parent, locale_text, locale_input_format, locale_texture)
	end
end

function createWindowWithDefaultOrLocale(parent, locale_text, locale_input_format, locale_texture, default_texture)
	if engine.m_default_language_used then
		return createWindowWithTexture(parent, default_texture)
	else
		return createWindowWithTextAndTexture(parent, locale_text, locale_input_format, locale_texture)
	end
end

function createButtonWithTextureOrLocaleText(parent, locale_text, locale_input_format, default_texture)
	if engine.m_default_language_used then
		return createButtonWithTexture(parent, default_texture)
	else
		return createButtonWithTextAndTexture(parent, locale_text, locale_input_format, nil)
	end
end

function createWindowWithTextAndTexture(parent, text, input_format, texture, window_anchor)
	local window = parent:spawn(Window())
	chunks.assets.setTexture(window, texture)
	window.receive_input = false

	if ( window_anchor ) then
		window.anchor = window_anchor
	end

	local label = window:spawn(Text())
	local format = input_format or {"Debug Font", 24.0, pool.get_colorf(0, 0, 1, 1), 0.0, TextAlign.Center, 0.0, 0.0}
    local alignment = format[5] or TextAlign.Center

	label.style = format[1]
	label.options = {height = format[2], max_width = format[4] or 0.0, align = format[5] or TextAlign.Center, char_spacing = format[8] or 0, localization_scaling_limit = 0.1}
	label.color = format[3]
	label.text = text or lstring("stat")


	local text_offset_x = format[6] or 0.0
	local text_offset_y = format[7] or 0.0

	-- The alignment of the text in 'format' determines the anchoring in this case as well
 	if ( alignment == TextAlign.Left ) then
 		label.anchor = { Anchor.Anchor_L, Anchor.Anchor_L, pool.get_v3f(text_offset_x, text_offset_y, 0.1) }
 	elseif ( alignment == TextAlign.Right ) then
 		label.anchor = { Anchor.Anchor_R, Anchor.Anchor_R, pool.get_v3f(text_offset_x, text_offset_y, 0.1) }
 	else
 		label.anchor = { Anchor.Anchor_C, Anchor.Anchor_C, pool.get_v3f(text_offset_x, text_offset_y, 0.1) }
 	end

    if ( nil == format[10] ) or ( true == format[10] ) then
		local fx = {drop_shadow = true, shadow_color = colorf(0, 0, 0, 1), shadow_offset = v3f( 2, -2, 0), shadow_blur_amount = 100000}
		label.fx = fx;
	end

	local scale_back = format[9] or false

	if scale_back then
		window.size = label.size + pool.get_v2f(6, 0)
	end

    return window, label
end


function createLabelWithTextureIfNotDefaultLocale(parent, text, input_format, texture, window_anchor)

	if engine.m_default_language_used then
		return nil
	else
		return createWindowWithTextAndTexture(parent, text, input_format, texture, window_anchor)
	end
end


function createText(parent, text, style, text_options, text_offset, use_drop_shadow)

	local label = parent:spawn(Text())

	label.style = style or "Debug Font" -- "Debug_Font" must be defined in your styles sheet this to work.

	if not text_options then
		print("Warning: No text_options were passed into util.lua createText method.")
		text_options = {height = 24.0, align = TextAlign.Center, font_color = colorf(1,0,1,1)}
	end

	-- If text_options is nil then the text will be green (uncommon enough color to be noticably broken).
	label.options = text_options

	if not text_options["font_color"] then
		print("Warning: No font color was passed into util.lua createText method.")
		text_options["font_color"] = colorf(0,1,0,1)
	end

    if ( text ) then
   		label.text = lstring(text)
   	else
   		print("Warning: No text was passed into util.lua createText method.")
   		label.text = lstring("Invalid_Text")
   	end

   	if ( use_drop_shadow ) then
		local fx = {drop_shadow = true, shadow_color = colorf(0, 0, 0, 1), shadow_offset = v3f( 2, -2, 0), shadow_blur_amount = 100000}
		label.fx = fx;
	end

	if not text_offset then
		text_offset = {x = 0.0, y = 0.0}
	end

	label.location = v3f(text_offset.x, text_offset.y, 0.1)

	-- The alignment of the text determines the anchoring in this case as well
 	if ( alignment == TextAlign.Left ) then
 		label.anchor = { Anchor.Anchor_L, Anchor.Anchor_L, v3f(text_offset.x, text_offset.y, 0.1) }
 	elseif ( alignment == TextAlign.Right ) then
 		label.anchor = { Anchor.Anchor_R, Anchor.Anchor_R, v3f(text_offset.x, text_offset.y, 0.1) }
 	else
 		label.anchor = { Anchor.Anchor_C, Anchor.Anchor_C, v3f(text_offset.x, text_offset.y, 0.1) }
 	end

	return label
end

function setLocaleTexture(actor, locale_texture, default_texture)
	-- If English is being used we try to use the default_texture, if that's nil we try locale_texture next.
	-- If English isn't being used we do the opposite.
	if ( engine.m_default_language_used ) then
		assets.setTexture(actor, default_texture or locale_texture)
	else
		assets.setTexture(actor, locale_texture or default_texture)
	end
end

function showBubbleSprite(game_object, asset, height, scale)
	local renderer = PartRenderIndex(game_object:getObjectIndex()):get()
	if not renderer then
		return
	end
	scale = scale or 1.0

	local render = renderer.render_actor
	local sprite = render:spawn(SpriteActor())
	assets.setTexture(sprite, asset)
	sprite.location = pool.get_v3f(0, scale*asset.size.y/2, -10)

	local length = 1.3
	local height = height or 100.0

	sprite.scale = pool.get_v3f(scale, scale, scale)

	sprite:addAction(MoveOffsetAction(length*0.5, InterpAction.Fall, pool.get_v3f(0, height, 0)))

	local split = SplitAction()
	split:addAction(AlphaAction(length*0.5, InterpAction.Lerp, 0))

	local sequence = SequenceAction()
	sequence:addAction(ScaleAction(length*0.25, InterpAction.Smooth, pool.get_v3f(scale, scale, scale)*2.0 ))
	sequence:addAction(ScaleAction(length*0.25, InterpAction.Smooth, pool.get_v3f(scale, scale, scale)))
	split:addAction(sequence)

	sprite:addAction(split)

	sprite:addAction(DestroyAction(0))

end

function showBubbleText(game_object, text, height)
	--local game = game_object:getGame()
	--local window = game:getWindow()
	local bubble_text_scale = engine.platform.constants.bubble_text_scale
	local renderer = PartRenderIndex(game_object:getObjectIndex()):get()
	if not renderer then
		return
	end

	local actor = renderer.render_actor
	local label = actor:spawn(Label())
	label.location = pool.get_v3f(0, 0, -10)
	label.format = format or TextFormat("font://Helvetica-Bold", 20.0/bubble_text_scale, pool.get_colorf(1.0, 0.0, 0.0, 1.0))
	label.text = text or ""
	label.scale = pool.get_v3f(bubble_text_scale, bubble_text_scale, 1.0)

	local length = 1.3
	local height = height or 100.0

	label:addAction(MoveOffsetAction(length*0.5, InterpAction.Fall, pool.get_v3f(0, height, 0)))

	local split = SplitAction()
	split:addAction(AlphaAction(length*0.5, InterpAction.Lerp, 0))

	local sequence = SequenceAction()
	sequence:addAction(ScaleAction(length*0.25, InterpAction.Smooth, pool.get_v3f(bubble_text_scale*2, bubble_text_scale*2, 1)))
	sequence:addAction(ScaleAction(length*0.25, InterpAction.Smooth, pool.get_v3f(bubble_text_scale, bubble_text_scale, 1)))
	split:addAction(sequence)

	label:addAction(split)

	label:addAction(DestroyAction(0))

end

--[[ DEPRECATED
function showDamagehBubbleText(game_object, damage)
	local format = nil
	local text = nil

	if damage > 0.0 then        -- Damaged
		format = TextFormat("font://Helvetica-Bold", 12.0, pool.get_colorf(1.0, 0.0, 0.0, 1.0))
		text = "-" .. damage
	elseif damage < 0.0 then    -- Healing
		format = TextFormat("font://Helvetica-Bold", 12.0, pool.get_colorf(0.0, 1.0, 0.0, 1.0))
		text = "+" .. math.abs(damage)
	else						-- No Damage or Heal
		format = TextFormat("font://Helvetica-Bold", 12.0, pool.get_colorf(1.0, 1.0, 1.0, 1.0))
		text = "" .. damage
	end

	showBubbleText(game_object, text, format)
end
]]

function showHealSign(game_object, asset, length)
	local game = game_object:getGame()

	local renderer = PartRenderIndex(game_object:getObjectIndex()):get()
	if not renderer then
		return
	end

	local length = length or 1.1
	local height = {100.0, 90, 80}
	local width = {0, 10, -10}

	local actor = renderer.render_actor
	for i=1,3 do
		local heal_sign = actor:spawn(SpriteActor())
		heal_sign.location = pool.get_v3f(0, 0, 1)
		asset = asset or chunks.assets.textures.heal_effect
		chunks.assets.setTexture(heal_sign, asset)
		heal_sign.color = pool.get_colorf(1, 1, 1, 0)



		heal_sign:addAction(SleepAction(i*length*0.1))

		heal_sign:addAction(AlphaAction(0, InterpAction.Lerp, 1.0))

		heal_sign:addAction(MoveOffsetAction(length*0.5, InterpAction.Fall, pool.get_v3f(width[i], height[i], 0)))

		local split = SplitAction()
		split:addAction(AlphaAction(length*0.5, InterpAction.Lerp, 0))

		local sequence = SequenceAction()
		sequence:addAction(ScaleAction(length*0.25, InterpAction.Smooth, pool.get_v3f(2.0, 2.0, 1)))
		sequence:addAction(ScaleAction(length*0.25, InterpAction.Smooth, pool.get_v3f(1.0, 1.0, 1)))
		split:addAction(sequence)

		heal_sign:addAction(split)

		heal_sign:addAction(DestroyAction(0))
	end

end

function createHealthBarSprite(parent, size, color)
	local container = parent:spawn(SpriteActor())
	container.hidden = true
	container.size = size

	local parent_scale = parent.scale
	container.scale = pool.get_v3f(1/parent_scale.x, 1/parent_scale.y, 1.0)

	local health_bar_background = container:spawn(SpriteActor())
	health_bar_background.color = pool.get_colorf(0.2, 0.2, 0.2, 0.75)
	health_bar_background.size = container.size
	health_bar_background.location = pool.get_v3f(0.0, 0.0, 0.0)
	health_bar_background.scale = pool.get_v3f(1, -1, 1)
	health_bar_background.texture = "?://" .. assets.textures.health_bar.path
	health_bar_background.texture.filter = TextureResource.Filter_Nearest

	local health_bar = container:spawn(SpriteActor())
	health_bar.color = color
	health_bar.size = container.size
	health_bar.location = pool.get_v3f(0.0, 0.0, 1.0)
	health_bar.texture = "?://" .. assets.textures.health_bar.path
	health_bar.texture.filter = TextureResource.Filter_Nearest

	return health_bar, container
end

function createEnemyHealthBar(parent)
	return createHealthBarSprite(parent, v2f(25, 4), colorf(1.0, 0.0, 0.0, 1.0))
end

function createAllyHealthBar(parent)
	return createHealthBarSprite(parent, v2f(25, 4), colorf(0.0, 0.7, 0.0, 1.0))
end

function scaleHealthBarForObject(game_object)
	local health_bar = game_object.lua.m_health_bar

	if health_bar == nil then
		return
	end

	local health_part = PartHealthIndex(game_object:getObjectIndex()):get()
	local health_percent = health_part:getHealthRemaining() / health_part.health
	health_bar.scale = pool.get_v3f(health_percent, health_bar.scale.y, health_bar.scale.z)

	local x_loc = ((health_bar.size.x * health_bar.scale.x) / 2.0) + (health_bar.size.x / -2.0)
	health_bar.location = v3f(x_loc, health_bar.location.y, health_bar.location.z)
end

function showAndScaleHealthBarForObject(game_object)
	scaleHealthBarForObject(game_object)

	local health_bar_container = game_object.lua.m_health_bar_container

	if health_bar_container == nil then
		return
	end

	health_bar_container.hidden_recursive = false
	health_bar_container:clearActions()
	health_bar_container:addAction(LuaAction(1.5, function (inner)
		inner.hidden_recursive = true
	end))
end

function spawnCoin(game, type, location, value)
	value = value or 1
	local coin_index = game:createGameObject("item", type, value)

	local coin = coin_index:get()
	if not coin then
		return
	end

	coin.team = GameObject.TEAM_COIN
	coin.color = pool.get_colorf(1,1,1,1)

	location.y = Map.get():getYGivenX(GameObject.TEAM_ENEMY, location.x)
	location.z = engine.platform.constants.game_camera.scene.clipping_planes.x - 1
	coin.location = location

	return coin_index
end

function tryToSpawnCoin(owner)
	local chance = chunks.load.getStat("units", owner.label, "iron_chance", owner:getLevel()) or 0
	local far_right = Map.get():getLastPoint(GameObject.TEAM_ALLY).x
	
    if math.random() < chance then

		if math.random() < 0.50 then
			local sleep = 0.1
			local spacing = 22
			local count = chunks.load.getStat("units", owner.label, "iron_drop", owner:getLevel()) or 0
			for i=0,(count-1) do

				local game = owner:getMainGame()
				local new_location = owner:getWorldLocation()

				local coin_index = spawnCoin(game, "iron_bar", new_location)
				if not coin_index then
					print("Coin index is invalid. Most likely the maximum number of game objects has been reached.")
					return
				end
				local coin = coin_index:get()
				if not coin then
					return
				end

				coin.color = pool.get_colorf(1,1,1,0)

				local length = 0.5
				local height = 50.0

				local offset_y = 0
				local render = PartRenderIndex(owner:getObjectIndex()):get()
				if render then
					offset_y = math.max(-7, render.render_actor.location.y)
				end

				local offset_x = math.min(-60, far_right - new_location.x) + i*spacing

				local map_y = Map.get():getYGivenX(GameObject.TEAM_ENEMY, new_location.x+offset_x)
				if map_y == 0 then
					coin:destroy()
					return
				end
				new_location.y = map_y + offset_y + height
				new_location.z = engine.platform.constants.game_camera.scene.clipping_planes.x - 1
				coin.location = new_location

				local loop = FiniteLoopAction(10)
				loop:addAction(ScaleAction(0.25, InterpAction.Smooth, pool.get_v3f(1.1, 1.1, 1)))
				loop:addAction(ScaleAction(0.25, InterpAction.Smooth, pool.get_v3f(0.9, 0.9, 1)))

				local split = SplitAction()
				split:addAction(MoveOffsetAction(length, InterpAction.Ramp,  pool.get_v3f(offset_x, -height, 0.0)))
				split:addAction(AlphaAction(length, InterpAction.Ramp, 1.0))
				split:addAction(loop)

				coin:addAction(SleepAction(i*sleep))
				coin:addAction(split)
				coin:addAction(AlphaAction(length, InterpAction.Lerp, 0.0))
				coin:addAction(DestroyAction(0.0))
			end
		else

			local sleep = 0.1
			local spacing = 12
			local count = chunks.load.getStat("units", owner.label, "gold_drop", owner:getLevel()) or 0
			local value = 1
			if count ~= 0 then
				local current_wave = UserData.instance():getCurrentWave()
				local wave_name = string.format("wave_%02d", current_wave)
				value = chunks.waves[wave_name].gold_value or 1
			end
			for i=0,(count-1) do

				local game = owner:getMainGame()
				local new_location = owner:getWorldLocation()

				local coin_index = spawnCoin(game, "gold_coin", new_location, value)
				if not coin_index then
					print("Coin index is invalid. Most likely the maximum number of game objects has been reached.")
					return
				end
				local coin = coin_index:get()
				if not coin then
					return
				end

				coin.color = pool.get_colorf(1,1,1,0)

				local length = 0.5
				local height = 50.0
				local height_above_path = 0.0


				local offset_y = 0
				local render = PartRenderIndex(owner:getObjectIndex()):get()
				if render then
					offset_y = math.max(-7, render.render_actor.location.y)
				end

				local offset_x = math.min(-40, far_right - new_location.x) + i*spacing

				local map_y = Map.get():getYGivenX(GameObject.TEAM_ENEMY, new_location.x+offset_x)
				if map_y == 0 then
					coin:destroy()
					return
				end
				new_location.y = map_y + offset_y + height
				new_location.z = engine.platform.constants.game_camera.scene.clipping_planes.x - 1
				coin.location = new_location

				local loop = FiniteLoopAction(10)
				loop:addAction(ScaleAction(0.25, InterpAction.Smooth, pool.get_v3f(1.1, 1.1, 1)))
				loop:addAction(ScaleAction(0.25, InterpAction.Smooth, pool.get_v3f(0.9, 0.9, 1)))

				local split = SplitAction()
				split:addAction(MoveOffsetAction(length, InterpAction.Ramp,  pool.get_v3f(offset_x, -height+height_above_path, 0.0)))
				split:addAction(AlphaAction(length, InterpAction.Ramp, 1.0))
				split:addAction(loop)

				coin:addAction(SleepAction(i*sleep))
				coin:addAction(split)
				coin:addAction(AlphaAction(length, InterpAction.Lerp, 0.0))
				coin:addAction(DestroyAction(0.0))
			end -- for loop

		end -- if 50%
    end --if drop chance

end --function

function createUnitOnPath(game, unit_type, level, initial_x_location, dont_play_spawn_sound)
	local unit_index = game:createGameObject("unit", unit_type, level or 0)
	local unit = unit_index:get()
	if unit == nil then
		return unit_index
	end

	local team = unit.team
	if (team ~= GameObject.TEAM_ENEMY) and (team ~= GameObject.TEAM_ALLY) then
		return unit_index
	end

	level = unit:getLevel()
	local is_unique = chunks.load.getStat("units", unit.label, "unique", level) or false
	local fly_in_on_start = chunks.load.getStat("units", unit.label, "fly_in_on_start", level) or false

	local start_loc = Map.get():getFirstPoint(team)
	if initial_x_location then
		local new_location = pool.get_v3f(initial_x_location, Map.get():getYGivenX(team, initial_x_location), start_loc.z)
		if team == GameObject.TEAM_ALLY then
			if new_location.x < start_loc.x then
				new_location = start_loc
			end
		else
			if new_location.x > start_loc.x then
				new_location = start_loc
			end
		end

		unit.location = new_location
	else
		if team == GameObject.TEAM_ALLY then
			unit.location = start_loc - pool.get_v3f(100,0,0)
		else
			unit.location = start_loc + pool.get_v3f(75,0,0)
		end

		if fly_in_on_start == true and team == GameObject.TEAM_ENEMY then
			local orders = PartOrdersIndex(unit_index):get()
			if orders then
				unit.team = GameObject.TEAM_NEUTRAL
				local logic = PartLogicIndex(unit_index):get()
				if logic then
					logic:requestProcess(0)
				end
				unit.location = unit.location + pool.get_v3f(0, 225, 0)
				unit:addAction(MoveAction(3,  InterpAction.Lerp, start_loc + pool.get_v3f(-100, 0, 0)))
				unit:addAction(LuaAction(0.0, function(inner)
					inner.team = team
					local inner_logic = PartLogicIndex(inner:getObjectIndex()):get()
					if inner_logic then
						inner_logic:requestProcess(1.0 / 60.0)
					end
				end))
			end
		else
			local orders = PartOrdersIndex(unit_index):get()
			if orders then
				local move_order = MoveOrder(start_loc)
				move_order:setBlocking(true)
				orders:addOrder(move_order, PartOrders.QUEUE_END)

				local scale_offset_order = ScaleOffsetOrder(2.0, 3.0)
				orders:addOrder(scale_offset_order, PartOrders.QUEUE_END)
			end
		end
	end

	-- Try to offset the unit's render part
	local render = PartRenderIndex(unit_index):get()
	if render then
		--local max_offset = 18.75
		local offset -- = math.random(-max_offset, max_offset)
		local spawns_near_center = chunks.load.getStat("units", unit_type, "spawns_near_center") or false

		if spawns_near_center == true then
			offset = getNextSlot(center_spawn_slots)
		elseif is_unique == true then
			offset = getNextSlot(unique_spawn_slots)
		else
			offset = getNextSlot(spawn_slots)
		end

		render.render_actor.location = render.render_actor.location + pool.get_v3f(0, offset, -10.0*offset)
		render.render_actor.scale = render.render_actor.scale * pool.get_v3f(1.0, 1.0, 0.001)
		
		if( unit_type == "mini_ash" ) then
			render.render_actor:getOwned(0).rotation = pool.get_v3f( 0, 0, 180 )
		end

	end

	--Play Spawn Sound
	if not dont_play_spawn_sound then
		chunks.load.playNextSpawnSound(unit_type, game)

	end

	return unit_index
end

function createEffect(main_game, effect_name, location, scale, priority)
	scale = scale or 1.0
	priority = priority or 100
	local info = EffectInfo(URI("particles", effect_name), v3f(scale))
	local effect = main_game:createEffect(info, priority)

	if effect ~= nil then
		if location ~= nil then
			effect.location = location
		end
	end

	return effect
end

function createBloodEffect(game_object)
	local render = PartRenderIndex(game_object:getObjectIndex()):get()
	if render == nil then
		return nil
	end

	local owner = game_object:getMainGame():getScene()
	local lifetime = 0.2
	local color = colorf(0.7, 0.0, 0.0, 1.0)
	local size_multiplier = 0.5

	local shower = owner:spawn(ParticleSystem())
	shower.texture  = assets.getTexture(assets.textures.particle_circle)
	shower.location = render.render_actor:getWorldLocation() + pool.get_v3f(0.0, 50.0, 10.0)

	-- Set shower parameters
	shower.particle_size             = v2f(assets.textures.particle_circle.size) * size_multiplier
	shower.shape                     = ParticleSystem.Cone
	shower.explosion                 = true
	shower.lifetime                  = lifetime
	shower.lifetime_range            = pool.get_v2f(lifetime * 3, lifetime * 4)
	shower.particles_per_spawn_range = pool.get_v2i(6, 7)
	shower.spawn_frequency_range     = pool.get_v2f(0.04, 0.11)
	shower.speed_range               = pool.get_v2f(50, 60)
	shower.acceleration_range_y      = pool.get_v2f(-350, -450)
	shower.acceleration_range_x      = pool.get_v2f(-50, 50)
	shower.initial_rotation_range    = pool.get_v2f(0, 360.0)
	shower.rotation_speed_range      = pool.get_v2f(360, 720)
	shower.cone_angle_range			 = pool.get_v2f(60, 120)
	shower.position_range			 = pool.get_v2f(10, 30)
	shower.follow_system             = false

	shower:addColorRange(color)
	shower:addColorRange(color)
	shower:addColorRange(pool.get_colorf(color.r, color.g, color.b, 0.0))

	shower:addScaleRange(pool.get_v2f(1,1))
	shower:addScaleRange(pool.get_v2f(1,1))
	--shower:addScaleRange(pool.get_v2f(0,0))

	return shower
end

function createDustEffect(game_object)
	do return nil end

	local render = PartRenderIndex(game_object:getObjectIndex()):get()
	if render == nil then
		return nil
	end

	local location = render.render_actor:getWorldLocation()
	if game_object.team == GameObject.TEAM_ENEMY then
		location = location + v3f(0.0, 50.0, 10.0)
	else
		location = location + v3f(0.0, 50.0, 10.0)
	end

	--location.z = 400.0
	return createEffect(game_object:getMainGame(), "skeleton_dust", location, 0.35, 60)
end

function playSoundAtCamera(uri, options)

	local path
	if uri.path then
		path = "?://" .. uri.path
	else
		path = uri
	end

	chunks.sounds.play(path, false, options)

end

function playSoundLeftOfCamera(uri, game)

	game:playLeftOfCamera(URI(uri))
end

function playRandomSoundFromTable(table, options)
	if #table > 0 then
		local asset = table[ math.random(1, #table) ]
		if asset then
			if type(asset) == "string" then
				playSoundAtCamera("?://" .. asset, options)
			else
				playSoundAtCamera("?://" .. asset.path, options)
			end

		end
	end
end

function unloadMenuAssets()
	--[[
	local textures = chunks.assets.textures
	local menu_assets = {
		textures.castle_front_background,
		textures.castle_front_mask,
		textures.upgrade_background,
	}

	local manager = Textures.get()
	for i,v in ipairs(menu_assets) do
		manager:destroyTexture( URI( v.path ) )
	end
	]]
end

function deepcopy(object)
    local lookup_table = {}
    local function _copy(object)
        if type(object) ~= "table" then
            return object
        elseif lookup_table[object] then
            return lookup_table[object]
        end
        local new_table = {}
        lookup_table[object] = new_table
        for index, value in pairs(object) do
            new_table[_copy(index)] = _copy(value)
        end
        return setmetatable(new_table, getmetatable(object))
    end
    return _copy(object)
end

function hasAnyUnitsUnlocked()
	local current_wave = UserData.instance():getCurrentWave()

	for k,unit in pairs(globals.ally_units) do
		local unlock_wave = chunks.load.getStat("units", unit.name, "unlock_wave")
		if current_wave >= unlock_wave then
			return true
		end
	end

	return false
end

function saveStringList(save_id, list)
	local save_data = nil
	for i,v in ipairs(list) do
		if save_data then
			save_data = save_data .. ";" .. v
		else
			save_data = "" .. v
		end
	end

	if save_data then
		UserData.instance():setValue(save_id, save_data)
	else
		UserData.instance():setValue(save_id, "")
	end
end

function loadStringList(save_id)
	local save_data = UserData.instance():getValue(save_id)
	if #save_data ~= 0 then
		local list = splitString(save_data, ";")
		return list
	end

	return {}
end

function setStringValue(id, data)
	UserData.instance():setValue(id, data)
end

function getStringValue(id, default)
	local data = UserData.instance():getValue(id)

	if #data == 0 then
		setStringValue(id, default or "")
		data = UserData.instance():getValue(id)
	end

	return data
end

function setNumberValue(id, data)
	UserData.instance():setValue(id, tostring(data))
end

function getNumberValue(id, default)
	local data = UserData.instance():getValue(id)

	if #data == 0 then
		setNumberValue(id, default or 0)
		data = UserData.instance():getValue(id)
	end

	local number = tonumber(data)
	return number
end

function setBooleanValue(id, data)
	local value = 0
	if data == true then
		value = 1
	end

	setNumberValue(id, value)
end

function getBooleanValue(id, default)
	local actual_default = 0
	if default == true then
		actual_default = 1
	end

	return getNumberValue(id, actual_default) ~= 0
end

function getValidSpells()
	local spells = {}

	for k,v in pairs(chunks.spells) do
		if v.script then
			table.insert(spells, {name=k, unlock_wave=v.unlock_wave})
		end
	end

	table.sort(spells, function(left, right) return left.unlock_wave < right.unlock_wave end )

	return spells
end

function getValidUnits()
	local units = deepcopy(globals.ally_units)

	for k,v in pairs(units) do
		v.unlock_wave = chunks.load.getStat("units", v.name, "unlock_wave") or math.huge
	end

	table.sort(units, function(left, right) return left.unlock_wave < right.unlock_wave end )

	return units
end

local function autoSelectObjects(string_list_id, possible_objects, num_object_slots)
	local object_list = loadStringList(string_list_id)

	-- Load up list with default objects if we have less than the max selected.
	if #object_list < num_object_slots then
		local current_wave = UserData.instance():getCurrentWave()
		for i,v in ipairs(possible_objects) do
			local available = current_wave >= v.unlock_wave
			if available then
				local found = false
				for j,v2 in ipairs(object_list) do
					if v2 == v.name then
						found = true
						break
					end
				end

				if not found then
					table.insert(object_list, v.name)

					if #object_list >= num_object_slots then
						break
					end
				end
			end
		end
	end

	saveStringList(string_list_id, object_list)
end

function autoSelectUnits()
	local string_list_id = globals.selected_units_id
	local possible_units = getValidUnits()
	local num_unit_slots = globals.max_selected_units

	autoSelectObjects(string_list_id, possible_units, num_unit_slots)
end

function autoSelectSpells()
	local string_list_id = globals.selected_spells_id
	local possible_spells = getValidSpells()
	local num_spell_slots = getNumSpellSlotsUnlocked()

	autoSelectObjects(string_list_id, possible_spells, num_spell_slots)
end

function tryToAutoSelectObjects(force)
	force = force or false
	if force == false then
		if UserData.instance():getCurrentWave() > globals.auto_select_max_wave then
			return
		end
	end

	autoSelectUnits()
	autoSelectSpells()
end

function getCurrentWaveData()
	local zero_string = ""
    if UserData.instance():getCurrentWave() < 10 then
        zero_string = "0"
    end
	local wave_name = "wave_" .. zero_string .. UserData.instance():getCurrentWave()
	return chunks.waves[wave_name]
end

function getNumUnitsInCurrentWave()
	local wave_data = getCurrentWaveData()
	if wave_data == nil then
		return 0
	end

	local num_units = 0
	for k,v in pairs(wave_data.unit_stat) do
		if chunks.units[v.unit] ~= nil then
			num_units = num_units + 1
		end
	end

	return num_units
end

function getNumSpellSlotsUnlocked()
	local wave = UserData.instance():getCurrentWave()
	for i,v in ipairs(globals.spell_slot_unlock_levels) do
		if wave < v then
			return i - 1
		end
	end

	return #globals.spell_slot_unlock_levels
end

function createGrannyEffect(owner, model, texture, animation_name, options, onAnimationEnd_callback)
	local granny_actor = owner:spawn(BasicSkeletonActor())
	local new_options = options or { loop_count = 1 }
	granny_actor:setSkeleton("?://" .. model)
	granny_actor.texture = "?://" .. texture
	granny_actor:playAnimation(animation_name, new_options)
	granny_actor.onAnimationEnd = function (inner)
		if onAnimationEnd_callback then
			onAnimationEnd_callback(inner)
		end
		inner:destroy()
	end

	return granny_actor
end

function createMultiLineLabel(parent, text, format, wrap_width, maximum_width)

	local label = parent:spawn(Text())
	label.style = format[1]
	label.options = {height = format[2], align = TextAlign.Center, wrapping = wrap_width, max_width = maximum_width or 10000}
	label.color = format[3] or pool.get_colorf(1,1,1,1)

	if(type(text) == "string") then
		--print("Start Multilinelabel with", text, "with style", format[1])
		label.text = lstring(text)
	else
		--print("Start Multilinelabel with an lstring ", "with style", format[1])
		label.text = text
	end

	return label
end

function loadObjectRenderDataFromStats(owner, stat_list, name, level)
	
	level = level or 1
	local texture = chunks.load.getStat(stat_list, name, "texture", level)
	if texture == nil then
		return nil
	end
	texture = chunks.load.getRandomItemFromString(texture)

	local object = nil
	local model = chunks.load.getStat(stat_list, name, "model", level)
	if model ~= nil then
		model = chunks.load.getRandomItemFromString(model)

		if fileHasExtension(model, "gr2") then
			object = owner:spawn(BasicSkeletonActor())
			object:enableExtraZOffset()
			object:setSkeleton("?://" .. model)
			object.texture = "?://" .. texture
		elseif fileHasExtension(model, "obj") then
			object = owner:spawn(ActorBase())
			object.mesh = "?://" .. string.gsub(model, "%.obj", "")  -- Strip the extension
			object.texture = "?://" .. texture
		else
			return nil
		end
	else
		local texture_asset = chunks.assets.findTextureByPathName(texture)
		if texture_asset then
			object = owner:spawn(SpriteActor())
			chunks.assets.setTexture(object, texture_asset)
		else
			return nil
		end
	end

	local scale = chunks.load.getStat(stat_list, name, "scale", level)
	if scale then
		object.scale = pool.get_v3f(scale, scale, 1.0)
	end

	local weapon_name = chunks.load.getStat(stat_list, name, "weapon", level)
	if weapon_name ~= nil then
		weapon_name = chunks.load.getRandomItemFromString(weapon_name)
		local weapon = loadObjectRenderDataFromStats(owner, "weapons", weapon_name, level)
		if weapon ~= nil then
			object:attach(weapon, "bone_weapon")
			weapon.location = pool.get_v3f(0.0, 0.0, 4.0)
		end
	end

	local helmet_name = chunks.load.getStat(stat_list, name, "helmet", level)
	if helmet_name ~= nil then
		helmet_name = chunks.load.getRandomItemFromString(helmet_name)
		local helmet = loadObjectRenderDataFromStats(owner, "items", helmet_name, level)
		if helmet ~= nil then
			object:attach(helmet, "bone_head")
			helmet.location = pool.get_v3f(0.0, 0.0, 5.0)
		end
	end

	local shield_name = chunks.load.getStat(stat_list, name, "shield", level)
	if shield_name ~= nil then
		shield_name = chunks.load.getRandomItemFromString(shield_name)
		local shield = loadObjectRenderDataFromStats(owner, "items", shield_name, level)
		if shield ~= nil then
			object:attach(shield, "bone_shield")
			shield.location = pool.get_v3f(0.0, 0.0, 6.0)
		end
	end

	if( name == "mini_ash" ) then
		object.rotation = pool.get_v3f( 0, 0, 180 )
	end

	return object
end

function setEndlessModeUnlocked()
	setBooleanValue(globals.endless_mode_unlocked_id, true)
end

function isEndlessModeUnlocked()
	return getBooleanValue(globals.endless_mode_unlocked_id, false)
end

function tryToPopOffShieldsAndHelmets(object_index)
	local inventory_part = PartInventoryIndex(object_index):get()
	if inventory_part == nil then
		return
	end

	local health_part = PartHealthIndex(object_index):get()
	local damage_percent = health_part:getHealthRemaining() / health_part.health
	local render = PartRenderIndex(object_index):get()

	local pop_and_kill_item = function(item_index)
		if item_index:get() == nil then
			return
		end

		inventory_part:removeItem(item_index)

		local y_launch_offset = 25.0
		local x_launch_offset = 15.0
		local half_arc_time = 0.2
		local item_game_object = item_index:get():getObjectIndex():get()

		local rotation_speed = math.random(-90, -90)
		local rotation_split = SplitAction()
		rotation_split:addAction(RotateOffsetAction(half_arc_time * 2, InterpAction.Lerp, pool.get_v3f(0.0, 0.0, rotation_speed)))

		-- Fly up
		local fly_up_action = SplitAction()
		fly_up_action:addAction(MoveOffsetAction(half_arc_time, InterpAction.Fall, pool.get_v3f(0.0, y_launch_offset, 0.0)))
		fly_up_action:addAction(MoveOffsetAction(half_arc_time, InterpAction.Lerp, pool.get_v3f(x_launch_offset, 0.0, 0.0)))
		fly_up_action:addAction(MoveOffsetAction(half_arc_time, InterpAction.Lerp, pool.get_v3f(x_launch_offset, 0.0, 0.0)))

		-- Fly down
		local fly_down_action = SplitAction()
		local target_y =  render.render_actor:getWorldLocation().y - item_game_object:getWorldLocation().y - y_launch_offset
		fly_down_action:addAction(MoveOffsetAction(half_arc_time * 2, InterpAction.Ramp, pool.get_v3f(0.0, target_y, 0.0)))
		fly_down_action:addAction(MoveOffsetAction(half_arc_time * 2, InterpAction.Lerp, pool.get_v3f(x_launch_offset, 0.0, 0.0)))

		local flight_sequence = SequenceAction()
		rotation_split:addAction(fly_up_action)
		rotation_split:addAction(fly_down_action)
		flight_sequence:addAction(rotation_split)
		flight_sequence:addAction(AlphaAction(1.0, InterpAction.Lerp, 0.0))
		flight_sequence:addAction(DestroyAction(0.0))
		item_game_object:addAction(flight_sequence)
	end

	-- Pop off shield
	if damage_percent < 0.666667 then
		local shield_item = inventory_part:getEquippedItemAtSlot(PartInventory.SHIELD_EQUIP_SLOT)
		pop_and_kill_item(shield_item)
	end

	-- Pop off helmet
	if damage_percent < 0.333333 then
		local helmet_item = inventory_part:getEquippedItemAtSlot(PartInventory.HELMET_EQUIP_SLOT)
		pop_and_kill_item(helmet_item)
	end
end

function shuffleSpawnSlots()
	temp_table = {}
	local size = #spawn_slots
	while size > 0 do
		local temp = table.remove(spawn_slots, math.random(1, size))
		table.insert(temp_table, temp)
		size = size - 1
		--print(temp)
	end
	spawn_slots = temp_table

	temp_table = {}
	local size = #unique_spawn_slots
	while size > 0 do
		local temp = table.remove(unique_spawn_slots, math.random(1, size))
		table.insert(temp_table, temp)
		size = size - 1
		--print(temp)
	end
	unique_spawn_slots = temp_table
end

function getNextSlot(slots)
	local result = table.remove(slots, 1)
	table.insert(slots, result)
	return result
end

function getCurrentWaveString()
	local format = TextFormat("font://Georgia Bold", 21.0, pool.get_colorf(0.93, 0.72, 0.26, 1.0))
	format.shadow_color = pool.get_colorf(0.0, 0.0, 0.0, 1.0)
	format.shadow_offset = pool.get_v2f(5.0, 5.0)

	local current_wave = UserData.instance():getCurrentWave()
	if current_wave == globals.endless_wave then
		return "Wave ∞", format
	end

	return "Wave " .. current_wave, format
end

function getCurrentWave()

	local current_wave = UserData.instance():getCurrentWave()
	if current_wave == globals.endless_wave then
		return lstring("infinity_symbol")
	end

	return current_wave
end

function setReceiveInputRecursive(actor, receive_input)
	local window = Window.fast_cast(actor)
	if window then
		window.receive_input = receive_input
	end

	local num_children = actor:getNumOwned()
	if num_children > 0 then
		for i = num_children - 1, 0, -1 do
			local child = actor:getOwned(i)
			setReceiveInputRecursive(child, receive_input)
		end
	end
end

function widthConstraint(text, width)
	local x = text.size.x
	if x > width then
		local scale = text.scale
		local x_scale = width / x
		text.scale = pool.get_v3f(x_scale, scale.y, scale.z)
	end
end

function unifyTextHeight(text_objects)
    -- Get the minimum width
    min_height = 1000000  -- This should be okay as our sizes are world sizes are small
    for _,text_obj in ipairs(text_objects) do
        min_height = math.min(min_height, text_obj.size.y)
    end

    
    --TODO: NOTE: this will not work correctly with text controls with different number of lines.
    --            Need to test for that and let the developer know.
    for _,text_obj in ipairs(text_objects) do
        
        local options = text_obj.options

        options.height = min_height
        text_obj.options = options
    end
end
        

function snapBellowOther(sprite_actor, other_actor, offset)
	local new_location =  v3f(sprite_actor.location.x,  other_actor.location.y -(other_actor.size.y/2.0) - (sprite_actor.size.y/2.0) -offset, sprite_actor.location.z)
   	sprite_actor.location = new_location
end

function snapAboveOther(sprite_actor, other_actor, offset)
	local new_location = v3f(sprite_actor.location.x,  other_actor.location.y +(other_actor.size.y/2.0) + (sprite_actor.size.y/2.0) +offset, sprite_actor.location.z)
   	sprite_actor.location = new_location
end

function snapRightOfOther(sprite_actor, other_actor, offset)
	local new_location = v3f(other_actor.location.x+ (other_actor.size.x/2.0) +(sprite_actor.size.x/2.0) + offset, sprite_actor.location.y, sprite_actor.location.z)
  	sprite_actor.location = new_location
end

function snapLeftOfOther(sprite_actor, other_actor, offset)
	local new_location = v3f(other_actor.location.x - (other_actor.size.x/2.0) -(sprite_actor.size.x/2.0) - offset, sprite_actor.location.y, sprite_actor.location.z)
   	sprite_actor.location = new_location
end

function snapRight(sprite_actor, world_size, offset)
	local new_location = v3f((world_size.x/2.0)-(sprite_actor.size.x/2.0) - offset, sprite_actor.location.y, sprite_actor.location.z)
	sprite_actor.location = new_location;
end

function snapLeft(sprite_actor, world_size, offset)
	local new_location = v3f((-world_size.x/2.0) + (sprite_actor.size.x/2.0) + offset, sprite_actor.location.y , sprite_actor.location.z)
	sprite_actor.location = new_location
end

function snapTop(sprite_actor, world_size, offset)
	local new_location = v3f(sprite_actor.location.x, (world_size.y/2.0)-(sprite_actor.size.y/2.0) - offset, sprite_actor.location.z)
   	sprite_actor.location = new_location
end

function snapBottom(sprite_actor, world_size, offset)
	local new_location = v3f(sprite_actor.location.x,  (-world_size.y/2.0) + (sprite_actor.size.y/2.0) + offset,  sprite_actor.location.z)
	sprite_actor.location = new_location
end

function requireNetworkConnection( networkCallback )

    local platform_ui = PlatformUIService.get()
    local info     = DeviceInfo.get()
    local connection_code = info:query( "net.connection_type" )

    if( platform_ui and ui_handle == nil ) then

        if(connection_code == "lan" or connection_code == "wwan")then
            networkCallback()
        else
            local callback = function (handle, response)
                ui_handle = nil
            end

            local title = lstring("feature_unavailable")
            local text = lstring("feature_not_available_time")
            local ok = lstring("ok_text")
            local ok = lstring("ok")
            ui_handle = platform_ui:showDialog(title, text, {ok}, callback)
        end
    end
end

-- Used to keep the callback in scope.
local video_handle = nil

-- Play the aod intro video
function playIntroVideo( stopped_callback ) 
	local ui = PlatformUIService.get()
    local info = DeviceInfo.get()

    if ui and info then
    	chunks.music.pauseSong()
        Audio.get().resetEnabled = false;

        local video_uri = GameService.get():getRawResourcePath()
        video_uri = video_uri .. "retina.mp4"

        video_handle = ui:playLocalVideo( video_uri, UIOrientation.Landscape, function() 
        	video_handle = nil
        	chunks.music.playSong()
        	Audio.get().resetEnabled = true;
        	setBooleanValue("has_seen_intro", true)
        	stopped_callback()
        	end)
    end
end

local gc_login_handle = nil
function requireGCLogin( _success_callback, _failure_callback )

	local gc = GameCenterService.get()
	if (gc:isLoggedIn()) then
		_success_callback()
	else
	    local callback = function(handle, response)
	        gc_login_handle = nil

	        if(response.error ~= nil)then
	            _failure_callback()
	        else
	        	_success_callback()
	        end
	    end

	    gc_login_handle = gc:login(true, callback)
	end
end

